;; =====================================================================
;;     usenet: thorsley@ssd.csd.harris.com  USMail: Tom Horsley
;; compuserve: 76505,364                         511 Kingbird Circle
;;      genie: T.HORSLEY                         Delray Beach, FL  33444
;; ======================== Aging: Just say no! ========================
;;
;; This file provides an interactive buffer-list capability.
;; When the function select-buffer is invoked, the minibuffer
;; prompts you for another buffer to select.  The default is the second
;; buffer on the buffer-list.  Also, all the keys that are normally
;; bound to next-line and previous-line are bound to functions that
;; navigate through the buffer list.  Any keys bound to kill-buffer
;; are rebound to a function that will kill the buffer currently
;; named in the minibuffer, then move to the next buffer on the list.
;; This is a faster means of selecting another buffer than buffer-menu
;; is, but with most of the power.
;; 
;; Bob Weiner, Motorola, Inc., 4/5/89
;;   Added 'select-buffer-other-window' command bound to {C-x4b}
;;   usually.
;; 
;; Bob Weiner, Motorola, Inc., 10/3/91
;;   Eliminated inefficient recursion for computing buffer list.
;;   This eliminated error of passing max-lisp-eval-depth when working
;;     with many buffers.
;;   Added completion to 'select-buffer' so it works like standard
;;     'switch-to-buffer' function.
;;
;; We have gotten to where we use this technique for several of the
;; packages we have written where something prompts for input, each
;; command keeps its own history list so you can quickly cycle through
;; the previous input to just that command.
;; 
;; It is very handy to rebind the keys where next line is, so you can
;; continue to use any cursor keys.
;;
;; (autoload 'select-buffer "buff-sel" nil t)
;; (global-set-key "\C-xb" 'select-buffer)
;; (global-set-key "\C-x4b" 'select-buffer-other-window)
;;
;; Modifications:
;; --------------
;; Modified on: 2 April 1992
;; Modified by: Uffe Kjaerulff <uk@iesd.auc.dk>
;; Description: Prompt changed from "Switch to buffer: " to
;;		"Switch to buffer: (default <next buffer>) " and
;;		initial input changed from "<next buffer>" to "".
;;		In this way `select-buffer' differs from the
;;		standard `switch-to-buffer' only by providing the
;;		interactive buffer-list capability described above.
;;		The fix eliminates the need to delete the initial input
;;		before typing (the prefix of) the name of a buffer.
;;		(Typically, you don't want to use the `next-line'/
;;		`previous-line' keys to obtain a buffer name if it's
;;		to be found far down the buffer list.)
;;
;; Modified on: 29 April 1992
;; Modified by: Marc Munro marc@bloodnok.com
;; Description: Changed the call to completing-read so that it doesn't
;;              require the named buffer to exist.  This is far more in
;;              keeping with the conversion to make select-buffer look
;;              like switch-to-buffer, in that it allows select-buffer to
;;              be used to create new buffers.
;; Modified on: 31 Dec 2002
;; Modified by: Marc Munro marc@bloodnok.com
;; Description: Much simplified.  Removed kill-buffer functionality.  Also
;;              next and previous functions, instead better using 
;;              completing-read for mosty equivalent functionality.  This
;;              eliminates custome keymaps and all sorts of complexity.
;; 

(defvar buffer-select-list-index 0 "Index into buffer-list")

(defvar buffer-select-local-list nil "Local copy of buffer-list")

(defun make-buffer-list (buffer-list)
  "Returns names from BUFFER-LIST excluding those beginning with a space."
  (delq nil (mapcar '(lambda (b)
		       (if (= (aref (buffer-name b) 0) ? ) nil b))
		    buffer-list)))

(defun select-buffer (&optional other-window)
  "Interactively select or kill buffer using the minibuffer.
Optional argument OTHER-WINDOW non-nil means display buffer in another window.
The default buffer is the second one in the buffer-list. Other buffers can
selected either explicitly, or by using buffer-select-next and
buffer-select-prev.  Keys normally bound to next-line are bound to
buffer-select-next, those normally bound to previous-line are bound to
buffer-select-prev, and those normally bound to kill-buffer are bound to
buffer-select-kill-buf."
   (interactive)
   (setq buffer-select-local-list (make-buffer-list (buffer-list)))
   (let ((default-buffer (buffer-name (car (cdr buffer-select-local-list))))
	 (buffer-names (mapcar 'buffer-name buffer-select-local-list))
	 inpt)
      (setq inpt
	    (completing-read (concat
			      "Switch to buffer"
			      (if other-window " in other window")
			      "(" default-buffer "): ")
			     (mapcar 'list buffer-names)
			     nil nil nil
			     (let ((len (length buffer-names)))
			       (cond ((= 0 len)
				      'buffer-names)
				     ((= 1 len)
				      '(buffer-names . 0))
				     (t '(buffer-names . 1))))
			     default-buffer
			     ))
      (if other-window
	  (switch-to-buffer-other-window inpt)
	(switch-to-buffer inpt))))

(defun select-buffer-other-window ()
  "See documentation for 'select-buffer'."
  (interactive)
  (select-buffer t))

