;;**********************************************************************;;
;;									;;
;;		e l e c - s q l                         		;;
;;		---------------                           		;;
;;									;;
;;	Marc Munro					1-Oct-93	;;
;;									;;
;;	Electric mode stuff for writing SQL statements.			;;
;;									;;
;;                                                                      ;;
;;                                                                      ;;
;;**********************************************************************;;

;; Define some basic mode variables.

(defvar elec-sql-mode-abbrev-table nil
  "Abbrev table in use in elec-sql-mode buffers.")
(defvar elec-sql-mode-map nil
  "Keymap used in elec sql mode.")
(defvar elec-sql-verbatim nil
  "Should keyword abbrevs be expanded to give full command syntax?")

(defvar sql-mode-syntax-table nil
  "Syntax table in use in SQL-mode buffers.")

(if sql-mode-syntax-table
    ()
  (setq sql-mode-syntax-table (make-syntax-table))
  (modify-syntax-entry ?\\ "\\" sql-mode-syntax-table)
  (modify-syntax-entry ?+ "." sql-mode-syntax-table)
  (modify-syntax-entry ?- "." sql-mode-syntax-table)
  (modify-syntax-entry ?= "." sql-mode-syntax-table)
  (modify-syntax-entry ?% "." sql-mode-syntax-table)
  (modify-syntax-entry ?< "." sql-mode-syntax-table)
  (modify-syntax-entry ?> "." sql-mode-syntax-table)
  (modify-syntax-entry ?& "." sql-mode-syntax-table)
  (modify-syntax-entry ?| "." sql-mode-syntax-table)
  (modify-syntax-entry ?\' "\"" sql-mode-syntax-table))


;; Variables used for determining indentation and position within sql
;; statements.

(defvar sql-command-words
  (concat
   "\\(insert\\|update\\|select\\|delete\\|commit\\|rollback"
   "\\|create\\|alter\\|drop\\)")
  "The sql reserved words that introduce a command.")

(defvar sql-syntax-words
  (concat
   "\\(into\\|from\\|where\\|set\\|values\\|union\\|minus\\|intersect"
   "\\|and\\|or\\|group\\s-+by\\|order\\s-+by\\|having\\)")
  "The other sql reserved words that affect and control indentation.")

(defvar sql-terminators
  "\\(;\\|^/[ \t]*$\\)"
  "Regular expressions that complete or terminate a sql statement.")


;; Variables and functions used to identify whether we are in SQL statements
;; and comments.  Customisation of this mode will require the modification of
;; these.

(defvar elec-sql-in-code-p 'elec-sql-in-code-p-default
  "Identfifies a function used to determine whether we are currently in a 
code statement as opposed to a comment or string.")

(defun elec-sql-in-code-p-default ()
  "Function to determine whether we are currently in a code statement as
opposed to a comment or string."
  (not (elec-sql-in-string-p)))

(defun elec-sql-in-string-p ()
  "Function to determine whether we are in a SQL string or not."
  (nth 3 (parse-partial-sexp (point-min) (point))))



(defvar elec-sql-in-sql-p 'elec-sql-in-sql-p-default
  "Identfifies a function used to determine whether we are currently in a 
sql statement.")

(defun elec-sql-in-sql-p-default ()
  "Function to determine whether we are currently in a sql statement."
  (save-excursion)
   (if (re-search-backward (concat "\\(" sql-terminators 
				   "\\|" sql-command-words
				   "\\)") nil t)
       (looking-at sql-command-word)))


;; Functions and variables for determining appropriate indentation level.

(defun elec-sql-prev-control ()
  "Function to scan back to the previous indentation control keyword."
  (let* ((found)
	 (done)
	 (regexp (concat "\\(" sql-terminators 
			 "\\|" sql-command-words
			 "\\|[()]\\)")))
    (while (not done)
      (setq found (re-search-backward regexp nil t))
      (setq done 
	    (or (not found)
		(and found (funcall elec-sql-in-code-p)))))
    (looking-at regexp)        ;; Reset match-data that may have been modified 
                               ;; by elec-sql-in-code-p
    found))

(defvar sql-indent-in-paren 4
  "Amount of extra identation as result of parenthesis.")

(defvar sql-control-indent 0
  "Amount of indentation for reserved (control) words in sql (eg select).")

(defvar sql-body-indent 7
  "Amount of indentation for body of sql statements.")

(defun sql-indent-level (same-level)
  "Determine the level of indent necessary at this point.  same-level is 
boolean identifying whether we are looking to indent the next word to
the same level as the previous control word.  If this is nil, we 
indent to the next indent level (that for the body of the statement)."
  (if (looking-at sql-terminators)
      0
    (let* ((found)
	   (done)
	   (extra 0)
	   (result 0))
      (save-excursion
	(while (not done)
	  (setq found (elec-sql-prev-control))
	  (if (looking-at "(")
	      (if (looking-at "([ \t\n]*select")    ;; Is this a sub-query
		  (setq extra (+ sql-indent-in-paren extra))
		;; Just looking at a parenthesised expresssion.
		(setq result (1+ (current-column)))
		(setq found nil)
		(setq done t))
	    ;; Not at an open paren
	    (if (looking-at ")")
		(progn               ;; Skip to start of parenthesised expr.
		  (forward-char 1)
		  (backward-sexp))
	      ;; Not looking at any part of a parenthesised expression.
	      (setq done t))))
	(if found
	    (setq result (if (looking-at sql-terminators)
			     sql-control-indent
			   (+ extra 
			      (current-column)
			      (if same-level
				  0
				7))))))
      result)))

(defun sql-indent ()
  "Modify indentation of current sql line."
  (interactive)
  (save-excursion
    (beginning-of-line)
    (setq left-margin 
	  (sql-indent-level (looking-at 
			     (concat "[ \t]*\\(" sql-command-words
				     "\\|" sql-syntax-words "\\)"))))
    (indent-to-left-margin)
    (move-to-column left-margin)))

(defun non-expanded-tab ()
  "Tab operation that does not expand abbrevs."
  (let* ((cur-col (current-column))
	 (tabs tab-stop-list)
	 (tab-stop (car tabs)))
    (while tab-stop
      (if (> tab-stop cur-col)
	  (progn
	    (insert-char 32 (- tab-stop cur-col))
	    (setq tab-stop nil))
	(setq tabs (cdr tabs))
	(setq tab-stop (car tabs))))))

(defun sql-tab ()
  "Tab operation for SQL stuff."
  (interactive)
  (let ((target-col
	 (sql-indent-level (looking-at 
			    (concat "[ \t]*\\(" sql-command-words
				    "\\|" sql-syntax-words "\\)")))))
    (if (< (current-column) target-col)
	(insert-char 32 (- target-col (current-column)))
      (non-expanded-tab))))

(defun sql-indent-and-tab ()
  "Like its name implies."
  (interactive)
  (let ((wordpos)
	(tab-needed))
    (if (looking-at (concat "[ \t]*\\(" sql-command-words
			    "\\|" sql-syntax-words "\\)"))
	(move-to-column (sql-indent))
      (save-excursion
	    (forward-word -1)
	    (setq wordpos (point))
	    (beginning-of-line)
	    (setq tab-needed (not (string-match "[ \t]*[^ \t\n]"
					  (buffer-substring (point) wordpos))))
	    (if (looking-at "[ \t]*[^ \t\n]")
		(sql-indent)))
      (if tab-needed 
	(sql-tab)
	(insert " ")))))

(defun sql-expand-and-set ()
  "Check mode, indent and tab, and remove last command character"
  (if (funcall elec-sql-in-code-p)
      (progn
	(sql-indent-and-tab)
	(setq unread-command-char ?\^?))
    (unexpand-abbrev)))

(defun sql-expand-select ()
  "Generate the details of a select statement."
  (if (funcall elec-sql-in-code-p)
      (let ((position)
	    (marker (point-marker)))
	(sql-indent)
	(re-search-backward "\\b\\w" nil t)
	(if (not (looking-at "select"))
	    ;; We must have expanded a union or something, so let's put in
	    ;; the select on the next line.
	    (progn
	      (end-of-line)
	      (insert "\nselect")
	      (sql-indent-and-tab))
	  (goto-char marker)
	  (sql-tab))
	(if elec-sql-verbatim nil
	  (setq position (point))
	  (insert "\nfrom")
	  (sql-indent-and-tab)
	  (insert "\nwhere")
	  (sql-indent-and-tab)
	  (goto-char position))
	(setq unread-command-char ?\^?))
    (unexpand-abbrev)))

(defun sql-expand-delete ()
  "Check mode, indent and tab, and remove last command character"
  (if (funcall elec-sql-in-code-p)
      (let ((position)
	    (marker (point-marker)))
	(sql-indent)
	(goto-char marker)
	(if elec-sql-verbatim nil
	  (insert "\nfrom")
	  (sql-indent-and-tab)
	  (setq position (point))
	  (insert "\nwhere")
	  (sql-indent-and-tab)
	  (goto-char position))
	(setq unread-command-char ?\^?))
    (unexpand-abbrev)))

(defun sql-expand-update ()
  "Check mode, indent and tab, and remove last command character"
  (if (funcall elec-sql-in-code-p)
      (let ((position)
	    (marker (point-marker)))
	(sql-indent)
	(goto-char marker)
	(sql-tab)
	(if elec-sql-verbatim nil
	  (setq position (point))
	  (insert "\nset")
	  (sql-indent-and-tab)
	  (insert "\nwhere")
	  (sql-indent-and-tab)
	  (goto-char position))
	(setq unread-command-char ?\^?))
    (unexpand-abbrev)))

(defun sql-expand-insert ()
  "Check mode, indent and tab, and remove last command character"
  (if (funcall elec-sql-in-code-p)
      (let ((position)
	    (marker (point-marker)))
	(sql-indent)
	(goto-char marker)
	(if elec-sql-verbatim nil
	  (insert "\ninto")
	  (sql-indent-and-tab)
	  (setq position (point))
	  (insert "\n()")
	  (beginning-of-line)
	  (sql-tab)
	  (goto-char position)
	  (setq unread-command-char ?\^?)))
    (unexpand-abbrev)))

(defun sql-expand-values ()
  "Insert values clause for insert statement."
  (if (funcall elec-sql-in-code-p)
      (let ((marker (point-marker)))
	(sql-indent)
	(goto-char marker)
	(sql-tab)
	(if elec-sql-verbatim nil
	  (insert "()")
	  (forward-char -1))
	(setq unread-command-char ?\^?))
    (unexpand-abbrev)))

;; Now the code to implement the electric-sql major mode.

(defun elec-sql-switch-verbatim (arg)
  "Switch elec-sql verbatim mode.
This has three values:
1 - verbatim off, abbrevs on  This gives full syntax expansion
2 - verbatim on, abbrevs on   This just expands keywords
3 - verbatim on, abbrevs off  This gives no expansion at all

with an argument, if arg is positive turns off all expansion, 
otherwise turns on full expansion."

  (interactive "P")
  (or (assq 'elec-sql-verbatim minor-mode-alist)
      (setq minor-mode-alist (append minor-mode-alist
				     (list '(elec-sql-verbatim
					     " Verbatim")))))
  (if arg 
      (if (> (prefix-numeric-value arg) 0)
	  (progn
	    (setq elec-sql-verbatim t)
	    (abbrev-mode -1))
	(setq elec-sql-verbatim nil)
	(abbrev-mode 1))
    (let ((cur-verbatim 
	   (+ (if elec-sql-verbatim 1 0)
	      (if abbrev-mode 1 2))))
      (message (int-to-string cur-verbatim))
      (setq elec-sql-verbatim (< cur-verbatim 3))
      (abbrev-mode (if (= cur-verbatim 2) -1 1)))))

(defun elec-sql-mode ()
  "High powered SQL editing mode
Elec sql mode provides expansion of SQL DML statements, and automatic
indentation for them.

Ultimately it will allow statements to be tested in a special SQL shell, but
that is for the future."
  (interactive)
  (kill-all-local-variables)
  (use-local-map elec-sql-mode-map)
  (setq major-mode 'elec-sql-mode)
  (setq mode-name "elec SQL")
  (make-sql-electric)
  (set-syntax-table sql-mode-syntax-table)
  (make-local-variable 'paragraph-start)
  (setq paragraph-start (concat "^$\\|" page-delimiter))
  (make-local-variable 'paragraph-separate)
  (setq paragraph-separate paragraph-start)
  (make-local-variable 'indent-line-function)
  (setq indent-line-function 'sql-indent)
  (make-local-variable 'require-final-newline)
  (setq require-final-newline t)
  (make-local-variable 'comment-start)
  (setq comment-start "REM ")
  (make-local-variable 'comment-end)
  (setq comment-end "")
  (make-local-variable 'comment-column)
  (setq comment-column 32)
  (make-local-variable 'comment-start-skip)
  (setq comment-start-skip "REM *")
;;  (make-local-variable 'comment-indent-hook)
;;  (setq comment-indent-hook 'c-comment-indent)
  (make-local-variable 'parse-sexp-ignore-comments)
  (setq parse-sexp-ignore-comments t)
  (run-hooks 'elec-sql-mode-hook))

(defun make-sql-electric ()
  "Function to set up the electric parts of electric sql without going the
whole hog and turning on the electric mode.  This is useful for 
incorporating electric sql statements into other modes."
  (if (not elec-sql-mode-abbrev-table)
      (let ((prev-a-c abbrevs-changed))
	(define-abbrev-table 'elec-sql-mode-abbrev-table '(
		("sel" "select" sql-expand-select 0)
		("select" "select" sql-expand-select 0)
		("union" "union" sql-expand-select 0)
		("inter" "intersect" sql-expand-select)
		("intersect" "intersect" sql-expand-select 0)
		("minus" "minus" sql-expand-select 0)
		("and" "and" sql-expand-and-set 0)
		("into" "into" sql-expand-and-set 0)
		("from" "from" sql-expand-and-set 0)
		("wh" "where" sql-expand-and-set 0)
		("where" "where" sql-expand-and-set 0)
		("del" "delete" sql-expand-delete 0)
		("delete" "delete" sql-expand-delete 0)
		("upd" "update" sql-expand-update 0)
		("update" "update" sql-expand-update 0)
		("ins" "insert" sql-expand-insert 0)
		("insert" "insert" sql-expand-insert 0)
		("grp" "group by" sql-expand-and-set 0)
		("group" "group by" sql-expand-and-set 0)
		("hv" "having" sql-expand-and-set 0)
		("having" "having" sql-expand-and-set 0)
		("ord" "order by" sql-expand-and-set 0)
		("order" "order by" sql-expand-and-set 0)
		("val" "values" sql-expand-values 0)
		("values" "values" sql-expand-values 0)))
	(setq abbrevs-changed prev-a-c)))
  (setq local-abbrev-table elec-sql-mode-abbrev-table)
  (make-local-variable 'elec-sql-verbatim)
  (auto-fill-mode 1))
  (abbrev-mode 1)
  (set-syntax-table sql-mode-syntax-table)

(setq elec-sql-mode-map (make-sparse-keymap))
(define-key elec-sql-mode-map "\t"       'sql-tab)
(define-key elec-sql-mode-map "\C-cv"     'elec-sql-switch-verbatim)

(modify-syntax-entry ?\" "\'" sql-mode-syntax-table)
(modify-syntax-entry ?\" "\"" sql-mode-syntax-table)

(provide 'elec-sql)