;;**********************************************************************;;
;;									;;
;;		p x - c   Praxis electric c definitions			;;
;;		---------------------------------------			;;
;;									;;
;;	Marc Munro					1-Oct-93	;;
;;									;;
;;	Defines key-bindings and extra functions for electric-c mode.	;;
;;									;;
;;                                                                      ;;
;;                                                                      ;;
;;**********************************************************************;;

;; I invoke this from my .emacs file with the following incantation
;;
;; (defvar c-mode-hook 'load-elec-c)
;; (defvar c-mode-loaded nil)
;; 
;; (defun load-elec-c ()
;;   "Function to load electric-c mode if necessary."
;;   (if c-mode-loaded
;;       nil
;;     (load "px-c"))
;;   (elec-c-mode)
;;   (setq c-mode-loaded t))


;; First we load the basic electric c stuff.  This file adds functionality
;; and better conformance to Praxis project C standards.

(load "elec-c")

;; Now redefine some of the standard functions to give us nicer indentation..

(defun elec-if-while ()
  (if (elec-c-inside-comment-or-string-p)
      nil
    (insert-string " ()\n{")
    (c-indent-line)
    (insert-string "\n}")
    (c-indent-line)
    (search-backward ")")
    (setq unread-command-events '(?\^?))))

(defun elec-else ()
  (if (elec-c-inside-comment-or-string-p)
      nil
    (insert-string "\n{")
    (c-indent-line)
    (insert-string "\n\n}")
    (c-indent-line)
    (next-line -1)
    (c-indent-line)
    (setq unread-command-events '(?\^?))))

(defun elec-for ()
  (if (elec-c-inside-comment-or-string-p)
      nil
    (insert-string " (;;)\n{")
    (c-indent-line)
    (insert-string "\n}")
    (c-indent-line)
    (search-backward ";;)")
    (setq unread-command-events '(?\^?))))

(defun elec-do ()
  (if (elec-c-inside-comment-or-string-p)
      nil
    (insert-string "\n{")
    (c-indent-line)
    (insert-string "\n\n} while ();")
    (c-indent-line)
    (next-line -1)
    (c-indent-line)
    (setq unread-command-events '(?\^?))))

;; New functions follow:

(defun c-back-function ()
  "Place point at the line prior to the current or previous function."
  (interactive)
  (beginning-of-defun)
  (backward-paragraph))

(defun c-forward-function ()
  "Place point at the line after the current or next function."
  (interactive)
  (end-of-defun))

(defun c-kill-function ()
  "Kill the current or previous function (may be yanked)."
  (interactive)
  (mark-c-function)
  (kill-region (point) (mark)))

(defun c-my-ctrl-j ()
  "Perform newline and then c-my-indent."
  (interactive)
  (newline 1)
  (setq this-command 'c-my-ctrl-j) ;; prevent kills being appended to
  (setq last-command 'c-my-ctrl-j) ;; each other
  (c-my-indent))


(defun setup-elec-c ()
  "Function to start up electric-c mode.  This will be the defult function
for elec-c-mode-hook"
  (interactive)
  (auto-fill-mode 1)
  (setq fill-column (- (window-width) 5)))


;; Functions to deal with comment boxes, as used in some Project C standards.

(defun elec-box ()
  "Create or tidy up a comment box (depending upon whether we are already
within a comment or not)."
  (interactive)
  (if (elec-c-inside-comment-p)
    (tidy-box)
    (make-box)))

(defun make-box ()
  "Create a comment box, suitably indented, and of width determined by
the last-comment-char constant."
  (interactive)
  (let ((chars-to-insert))
    (insert-string "\n")
    (next-line -1)
    (c-indent-line)
    (insert "/**/\n\n")
    (backward-char 4)
    (setq chars-to-insert (- last-comment-char (current-column)))
    (insert-char ?- chars-to-insert)
    (next-line 2)
    (c-indent-line)
    (insert "/*")
    (insert-char ?- chars-to-insert)
    (insert "*/\n")
    (next-line -2)
    (c-indent-line)
    (insert-string "/* ")))


(defun tidy-box ()
  "Tidy a box comment.  This places closing comment tokens at the end of
each line in the last-comment-char (constant) position.  If the text of
any line goes beyond this position, that line will not be affected."
  (interactive)
  (let (gen-marker (make-marker))
    (search-forward "-*/")
    (setq gen-marker (point-marker))
    (backward-char 2)
    (search-backward "-*/")
    (forward-char 2)
    (search-forward "*/")
    (while (< (point) (marker-position gen-marker))
      (backward-delete-char 2)
      (search-forward "*/"))
    (backward-char 2)
    (search-backward "-*/")
    (next-line 1)
    (end-of-line)
    (while (< (point) (marker-position gen-marker))
      (insert-char 32 (- last-comment-char (current-column)))
      (insert-string "*/")
      (next-line 1)
      (end-of-line))
    (next-line 1)
    (end-of-line)))

(defun kill-box-edges ()
  "This function removes the comment start and end tokens from each line
within a comment box, and brings the content of the comment to the
leftmost column of the file.  It may then be reformatted for later
re-insertion into another comment."
  (interactive)
  (tidy-box)
  (search-backward "-*/")
  (search-backward "-*/")	;; Find top line of box comment.
  (next-line 1)
  (end-of-line)
  (backward-char 3)
  (while (not (string= (buffer-substring (point) (+ (point) 3)) "-*/"))
    (end-of-line)
    (delete-backward-char 2)
    (delete-horizontal-space)
    (beginning-of-line)
    (delete-horizontal-space)
    (delete-backward-char -3)
    (next-line 1)
    (end-of-line)
    (backward-char 3))
  (search-backward "-*/")
  (forward-char 4))


;; Functions for providing C-function listing and usage.  These assume that
;; functions are defined with the function name at the start of a line
;; in the form:
;;
;; typename
;; functionname (args)
;; {
;;     functionbody
;; }

(defvar func-str "[a-z][a-z0-9_]*[ 	]*("
  "Regular expression identifying syntax of function definition and usage.")

(defvar reserved-words '(if for while do switch return)
  "Words which look like function calls, but aren't!")

(defvar ignore-functions '(strcmp strcpy strlen strcat strncmp strncpy
			   strncat STREQ malloc MALLOC calloc CALLOC
                           free FREE printf fprinft sprintf scanf gets
			   puts getc putc getchar putchar)
  "Functions and macros whose usage we are not interested in.")

(defvar matched-functions nil
  "The list of functions so far identified as used in the current function.")

(defvar new-function)

(defun c-reserved-word (word)
  "word must be a symbol."
  (memq word reserved-words))

(defun c-boring-function (word)
  "word must be a symbol."
  (memq word ignore-functions))

(defun c-matched-function (word)
  "word must be a symbol."
  (memq word matched-functions))

(defun elec-c-goto-next-function ()
  "Look for and process the next function definition or usage."
  (let (found word start)
    (setq new-function nil)
    (while (not found)
      (if (re-search-forward func-str nil t)
	  (progn
	    (forward-word -1)
	    (setq new-function (eq 0 (current-column)))
	    (if (and (not (elec-c-inside-comment-p))
		     (not (elec-c-inside-string-p)))
		(progn
		  (setq start (point))
		  (forward-word 1)
		  (setq word (car (read-from-string 
				   (buffer-substring start (point)))))
		  (if (and (not (c-reserved-word word))
			   (not (c-boring-function word)))
		      (setq found t)))
	      (forward-word 1)))
	(setq word nil)
	(setq found t)))
    word))

(defun write-function-to-temp (word definition)
  "Write details of functions and usage to a temporary buffer from which they
can be copied into your program."
  (interactive)
  (let (tmp-buf)
    (save-excursion
      (if (not (setq tmp-buf (get-buffer "*c-functions*")))
	  (progn
	    (setq tmp-buf (generate-new-buffer "*c-functions*"))
	    (set-buffer tmp-buf)
	    (insert "Functions defined in this compilation unit:")
	    (set-mark (point))))
      (set-buffer tmp-buf)
      (if definition
	  (progn
	    (goto-char (mark))
	    (newline)
	    (insert "/*	          " (symbol-name word))
	    (set-mark (point))
	    (goto-char (point-max))
	    (newline 2)
	    (insert "/* Function: " (symbol-name word) " calls:"))
	(newline)
	(insert "/*           " (symbol-name word))))))
		

(defun elec-c-list-functions ()
  "This lists in another buffer all of the functions defined in the
current buffer, and describes which other functions each of these call"
  (interactive)
  (save-excursion
    (let (word)
      (goto-char (point-min))
      (if (get-buffer "*c-functions*")
	  (kill-buffer "*c-functions*"))
      (while (setq word (elec-c-goto-next-function))
	(if new-function
	    (progn
	      (setq matched-functions nil)
	      (message (concat "Scanning function: " (symbol-name word)))
	      (write-function-to-temp word t))
	  (if (not (c-matched-function word))
	      (progn
		(setq matched-functions (cons word matched-functions))
		(write-function-to-temp word nil)))))))
  (switch-to-buffer-other-window "*c-functions*")
  (goto-char (point-min))
  (message "Done"))

;; Finally some key-bindings and other customisable bits.

(setq elec-c-mode-map (make-sparse-keymap))
(define-key elec-c-mode-map "{"	       	 'elec-c-left-brace)
(define-key elec-c-mode-map "}"	       	 'electric-c-brace)
(define-key elec-c-mode-map "("	       	 'elec-c-opening-brac)
(define-key elec-c-mode-map "["	       	 'elec-c-opening-brac)
(define-key elec-c-mode-map ";"	       	 'elec-c-semi)
(define-key elec-c-mode-map ":"	       	 'electric-c-terminator)
(define-key elec-c-mode-map "\e\C-h"   	 'mark-c-function)
(define-key elec-c-mode-map "\e\C-q"   	 'indent-c-exp)
(define-key elec-c-mode-map "\177"     	 'backward-delete-char-untabify)
(define-key elec-c-mode-map "\C-c\C-c" 	 'elec-c-close-block)
(define-key elec-c-mode-map "\C-cv"    	 'toggle-verbatim)
(define-key elec-c-mode-map "\e{"      	 'elec-c-remove-braces)
(define-key elec-c-mode-map "\C-j"     	 'c-my-ctrl-j)
(define-key elec-c-mode-map "\t"       	 'c-indent-command)

(define-key elec-c-mode-map  [kp-f2]     'backward-sexp)
(define-key elec-c-mode-map  [kp-f3]     'forward-sexp)

(setq elec-c-gold-map  (make-sparse-keymap))

(define-prefix-command 'elec-c-gold-map)

(mapcar (lambda (key)
	  (define-key elec-c-mode-map key 'elec-c-gold-map))   
	(where-is-internal 'gold-map global-map))

(define-key elec-c-gold-map [kp-f2] 	  'c-back-function)
(define-key elec-c-gold-map [kp-f3] 	  'c-forward-function)
(define-key elec-c-gold-map [kp-f4] 	  'c-kill-function)   
(define-key elec-c-gold-map [kp-8]  	  'elec-box)
(define-key elec-c-gold-map [kp-0]  	  'elec-c-close-block)
(define-key elec-c-gold-map [kp-subtract] 'elec-c-remove-braces)
(define-key elec-c-gold-map [f6]          'kill-box-edges)

;; Add underscore ("_") and hash ("#") to syntax table so that they form
;; part of the syntax for a word.

(modify-syntax-entry ?# "w" c-mode-syntax-table)
(modify-syntax-entry ?_ "w" c-mode-syntax-table)

;; And modify the comment and indentation styles.

(defvar elec-c-mode-hook 'setup-elec-c)
(setq c-tab-always-indent nil)
(setq c-auto-newline t)
(setq c-indent-level 4)
(setq c-argdecl-indent 4)
(setq c-brace-offset -4)
(setq c-brace-imaginary-offset 0)
(setq c-continued-statement-offset 4)
(setq comment-multi-line nil)
(setq comment-edged nil)
(setq last-comment-char (- (window-width) 3))


