;; TODO:
;; next thing is to identify which frame to switch to when switching to
;; a buffer that has already been displayed.  Also ensure that frame is raised.



;; My frame management
;; provides smarter management of frames, more integration between files
;; and buffers, and elimination of most window splitting.

(defun mfm-special-buffer-p (buf)
  "Predicate identifying whether a buffer (or buffer name) represents a
special buffer.  Special buffers are buffers like *scratch*,
*Compilation*, etc."
  (and (string-match "\\*" (buffer-name buf)) t command-history))


(defun mfm-common-string-prefix (list &optional prefix)
  "Return the common string prefix from LIST, optionally matching PREFIX."
  (if prefix
      (if list
	  (mfm-common-string-prefix (cdr list)
				    (fill-common-string-prefix prefix 
							       (car list)))
	prefix)
    (if list
	(mfm-common-string-prefix (cdr list) (car list)))))


(defun mfm-all-buffers ()
  "Return a list of all buffer names."
  (mapcar 'buffer-name (buffer-list)))

(defun mfm-matching-files (dirname filename-part)
  "Return the set of filenames for within DIRNAME that match FILENAME-PART"
  (all-completions filename-part 
		   (directory-files dirname)))

(defun mfm-try-completion (completions dirname)
  "Helper function for mfm-completions, that handles the try-completion
  operation."
  (cond
   ((eq 0 (length completions))	nil)
   ((eq 1 (length completions))	(car completions))
   (t (concat dirname (mfm-common-string-prefix completions)))))


(defun mfm-completion-boundaries (dirname str)
  "Helper function for the 'boundaries completion interface."
  (cons (list 'boundaries 
	      (if (string= dirname (fill-common-string-prefix dirname str))
		  (length dirname) 0))
	(length str)))


(defun mfm-completions (str pred mode)
  "Completions function for completing based on list of buffers, or real
file names."
  (let* ((dirname (replace-regexp-in-string "[^/]+$""" str))
	 (filename (replace-regexp-in-string "^.*/""" str))
	 (buffers (and (string= dirname "") 
		       (all-completions filename (mfm-all-buffers))))
	 (files (mfm-matching-files dirname filename))
	 (completions (append buffers files)))

    (delete-dups completions)
    ;; return differently based on what mode we were called in
    (cond
     ((not mode) (mfm-try-completion completions dirname))
     ((eq mode 't) completions)
     ((eq mode 'lambda) (member str completions))
     ((eq (car mode) 'boundaries) (mfm-completion-boundaries dirname str))
     ((eq mode 'metadata)
      (error "Metadata not handled in mfm-completions"))
     (t
      (error "Unknown mode in mfm-completions %S" mode))
    )))


(defun mfm-completing-read (prompt &optional default)
  "Do a completing read that uses both buffer and file histories, and
which does completion on bothe buffer and file names.  This is part of
an attempt to better unify buffersi and files."
  (let* ((my-history (append (reverse buffer-name-history) file-name-history))
	 (my-pos (length file-name-history)))
  (completing-read prompt 'mfm-completions nil nil default
		   (cons 'my-history my-pos))))

(defun mfm-autocreate-p (filename)
  "Should FILENAME be automatically created as a file?  True if the name
contains a / character (directory separator)."
  (and (string-match "/" filename) t))

(string-match "/" "wibble")

(defun mfm-select-bfile-to-buffer (prompt &optional default)
  "Identify a buffer or file and switch to it, creating the buffer or
file as necessary."
  (let* ((buf-or-file (mfm-completing-read prompt default))
	 (buf (get-buffer buf-or-file)))
    (or buf
	(if (or (file-exists-p buf-or-file)
		(mfm-autocreate-p buf-or-file))
	    (find-file-noselect buf-or-file)
	  (prog1
	      (generate-new-buffer buf-or-file)
	    (message "Created non-file buffer: %s" buf-or-file))))))

(defun mfm-switch-to-frame (frame)
  "Select FRAME for interactive use, ensuring that the mouse pointer follows 
the frame selection if necessary"
  (let ((mouse-pos (cdr (mouse-position))))
    (select-frame frame)
    (if mouse-pos
	;; The mouse was inside the original frame, so it may have to be moved.
	(let* ((new-pos (mouse-position))
	       (mouse-frame (car new-pos)))
	  (if (eq frame mouse-frame)
	      ;; The new frame happens to be in the right place for the mouse
	      nil
	    ;; Warp the mouse to the equivalent position in the new frame
	    (set-mouse-position frame (car mouse-pos) (cdr mouse-pos)))))
    frame))

(defun mfm-frame-for-buffer (buf)
  "Return the most appropriate frame for displaying BUF.  The most
appropriate will be one that is currently showing BUF, or the one that
has most recently shown BUF.  If no such frame can be found, return
nil."
  nil)

(defun mfm-select-bfile-std (&optional default)
  "Select a buffer or file, creating a new frame for it"
  (let ((buf (mfm-select-bfile-to-buffer "(new) File/Buffer: " default)))
    (buffer-name
     (if mfm-retry-read
	 ;; Our operating mode was changed, so just return buf and let the
	 ;; caller change mode for us.
	 buf
       (let ((frame (make-frame)))
	 (mfm-switch-to-frame frame)
	 (switch-to-buffer buf))))))

(defun mfm-select-bfile-gold (&optional default)
  "Select a buffer or file, if possible switching to an existing frame
that contains the buffer, otherwise creating a new buffer."
  (let ((buf (mfm-select-bfile-to-buffer "(switch) File/Buffer: " default)))
    (buffer-name
     (if mfm-retry-read
	 ;; Our operating mode was changed, so just return buf and let the
	 ;; caller change mode for us.
	 buf
       (let ((frame (or (mfm-frame-for-buffer) (make-frame))))
	 (mfm-switch-to-frame frame)
	 (switch-to-buffer buf))))))

(defun mfm-select-bfile-blue (&optional default)
  "Select a buffer or file, creating a new frame for it"
  (let ((buf (mfm-select-bfile-to-buffer "(select) File/Buffer: " default)))
    (buffer-name
     (if mfm-retry-read
	 ;; Our operating mode was changed, so just return buf and let the
	 ;; caller change mode for us.
	 buf
       (let ((mouse-pos (mouse-position))
	     (frame (make-frame)))
	 (select-frame frame)
	 (switch-to-buffer buf))))))


(defun mfm-next-mode ()
  "Cycle to the next minibuffer-completion-mode."
  (setq mfm-completion-mode
	(cond
	 ((eq mfm-completion-mode 'blue) nil)
	 ((eq mfm-completion-mode 'gold) 'blue)
	 (t 'gold))))


(defun mfm-select-bfile (mfm-completion-mode)
  "Call one of the select-bfile functions, and allow us to switch to a
different function if required.  The result of this function is a string
containing the name of a buffer"
  (let ((mfm-retry-read t)
	(result default-directory))
    (while mfm-retry-read
      (setq mfm-retry-read nil)
      (setq result
	    (cond
	     ((eq mfm-completion-mode 'gold)
	      (mfm-select-bfile-gold result))
	     ((eq mfm-completion-mode 'blue)
	      (mfm-select-bfile-blue result))
	     (t
	      (mfm-select-bfile-std result)))))
    result))

(defun mfm-select-bfile-newframe ()
  (interactive)
  (mfm-select-bfile nil))

(defun mfm-select-bfile-selectframe ()
  (interactive)
  (mfm-select-bfile 'gold))

(defun mfm-select-bfile-curframe ()
  (interactive)
  (mfm-select-bfile 'blue))

(defun mfm-switch-mode ()
  (interactive)
  (mfm-next-mode)
  (setq mfm-retry-read t)
  (exit-minibuffer))

;;(define-prefix-command 'mfm-gold-map)
;;(define-key minibuffer-local-map [f1] 'mfm-gold-map)
;;(define-key mfm-gold-map "m" 'mfm-switch-mode)
;;(define-key global-map [C-n] 'mfm-select-bfile-newframe)
;;(define-key gold-map "n" 'mfm-select-bfile-selectframe)
;;(define-key dblgold-map "n" 'mfm-select-bfile-curframe)




