;; Notes
;; redo should normally be a stopper
;; switching buffers should be a stopper
;; stuff in minibuffers should be ignored
;; minibuffer completion should probably add a complex command
;; after switching buffers we should maybe reset the cursor in the prev buffer
;; TODO:
;; - allow minibuffer stuff to be captured?
;; - ensure mouse operations are stoppers (use last-input-event)
;; - optimise redo-list stuff (eliminate all redundant entries in lists)
;; - add customisation variables

(require 'ring)

(defvar redo-state 'none
  "The current state of redo capture.  The states are:
'none      - redo is not being captured and has not been captured
'argonly   - we have captured a universal argument
'capturing - we are currently capturing redo
'held      - we have captured redo and are holding it
'heldarg   - we have begun capturing a universal argument from the held state
'captarg   - we have begun capturing a universal argument from the capturing
             state
")

;; Transition from one state to another will be for the following reasons:
;; arg   - universal argument has been entered
;; edit  - an edit operation has commenced
;; stop  - a stop function has been encountered
;; redo  - the redo key has been hit (this is like edit but more complex)

;; Start      |                    Operation                                  |
;; State      |    arg    |    edit    |    stop    |    redo    |    quit    |
;;------------+-----------+------------+------------+------------+------------+
;;            |           |            |            |            |            |
;; none       | argonly   | capturing  |  none      |  none      |  none      |
;;            |           |            |            |            |            |
;; argonly    | argonly   | capturing  |  none      |  capturing |  none      |
;;            |           |            |            |            |            |
;; capturing  | captarg   | capturing  |  held      |  capturing |  none      |
;;            |           |            |            |            |            |
;; held       | heldarg   | capturing  |  held      |  none      |  none      |
;;            |           |            |            |            |            |
;; heldarg    | heldarg   | capturing  |  held      |  capturing |  none      |
;;            |           |            |            |            |            |
;; captarg    | captarg   | capturing  |  held      |  capturing |  none      |
;;            |           |            |            |            |            |
;;

(defvar redo-pre-buffer nil
  "The buffer we were in at the start of the command")

(defvar redo-pre-undo nil
  "The state of undo at the start of the command")

(defvar redo-pre-mark nil
  "The mark at the start of the command.")
	
(defvar redo-pre-mark-ring nil
  "The mark-ring at the start of the command.")
	
(defconst redo-allow-redo-recording t
  "This allows the recording of redo information.  It is overridden during the
redo-redo function to prevent recursion.")

(defvar redo-buffer nil
  "Variable into which our current redo is collected.")

(defvar redo-univarg-buffer nil
  "Buffer into which universal arguments are collected prior to collecting
real redo.")

(defvar redo-stoppers
  (list 'minibuffer-complete-and-exit 'save-buffer)
  "List of functions that stop the accumulation of redo.  These are known
as stopper or stop functions.")

(defvar redo-conditional-stoppers
  (list 'next-line 'previous-line 'end-of-buffer 'beginning-of-buffer
	'scroll-down 'scroll-up 'scroll-up-in-place 'scroll-down-in-place)
  "List of functions that may stop the accumulation of redo.  Whether
they do so will depend on whether we have marked something as part of the
redo-able operation.")

(defun redo-set-cursor (new-state)
  "Set cursor in accordance with state."
  (cond
   ((equal new-state 'capturing) 
    (setq cursor-type 'bar))
   ((member new-state '(none held))
    (setq cursor-type t))
   ((equal new-state '(argonly heldarg captarg))
    (setq cursor-type '(bar . 8)))))

(defun redo-active-mark-p ()
  "t if we are in transient-mark-mode and the mark is active"
  (and transient-mark-mode mark-active))

(defun redo-stopper-fn-p (fn)
  "Predicate identifying whether FN is a stopper function."
  (or (member fn redo-stoppers)
      (and (not (redo-active-mark-p))
	   (member fn redo-conditional-stoppers))))

(defun redo-equivalent-undo-p (undo1 undo2)
  "Compare UNDO1 and UNDO2, returning t if they are equivalent."
  ;; TODO: figure out a better way to do this.
  (while (and undo1 (not (car undo1)))
    ;; step past any undo boundaries
    (setq undo1 (cdr undo1)))
  (while (and undo2 (not (car undo2)))
    ;; step past any undo boundaries
    (setq undo2 (cdr undo2)))
  (eq (cdr undo1) (cdr undo2)))

(defun redo-undo-updated-p (cmd)
  "Did this command begin an editing sequence?  It is if it has caused 
undo to be modified and is not itself an undo command."
  ;; Comparing cdr of the undo lists appears to be the safe way to
  ;; tell whether or not we modified undo
  ;; TODO rewrite to make it cleaner.
  (let ((is-undo (member cmd (list 'undo 'advertised-undo))))
    (and (not is-undo) 
	 (not (redo-equivalent-undo-p redo-pre-undo buffer-undo-list)))))

(defun redo-mark-updated-p ()
  "Did this command modify the mark?"
  (not (and (equal redo-pre-mark (mark-marker))
	    (equal redo-pre-mark-ring mark-ring))))

(defvar redo-redo-entry nil
  "The entry to be recorded for the current operation.")

(defun redo-record-pre-cmd-state ()
  "Record the buffer state prior to command execution."
  (setq redo-pre-buffer (current-buffer)
	redo-pre-undo buffer-undo-list
	redo-pre-mark (copy-marker (mark-marker))
	redo-pre-mark-ring mark-ring
	redo-redo-entry (list "KEY" (this-command-keys) this-command)))

(defun redo-add (entry)
  "Add ENTRY to the current list of redo"
  (setq redo-buffer 
	(if redo-buffer
	    (append redo-buffer (list entry))
	  (list entry))))

(defun redo-add-arg (entry)
  "Add ENTRY to the current list of redo arguments"
  (setq redo-univarg-buffer
	(if redo-univarg-buffer
	    (append redo-univarg-buffer (list entry))
	  (list entry))))

(defun redo-append-args-to-capture ()
  "Append the contents of the universal arg buffer to the redo-buffer."
  (while redo-univarg-buffer
	 (setq redo-buffer (append redo-buffer 
				   (list (car redo-univarg-buffer)))
	       redo-univarg-buffer (cdr redo-univarg-buffer))))

(defun redo-enter-capture-state ()
  "Place the redo state machine into the capture state.  Note that the
result of this is assumed to be non-nil by redo-change-state."
  (setq redo-buffer redo-univarg-buffer
	redo-univarg-buffer nil
	redo-state 'capturing))

(defun redo-change-state (operation)
  "Modify the current redo state according to OPERATION.  This is the heart
of the Finite State Machine that is redo.  OPERATION will be one of 'edit,
'arg, 'stop, 'redo or nil (meaning none of the above)."
  (and operation
       (if (eq operation `quit)
	   (redo-set-cursor (setq redo-buffer nil
				  redo-state 'none))
	 (cond
	  ((eq redo-state 'none)
	   (cond
	    ((eq operation 'arg)
	     (setq redo-univarg-buffer nil
		   redo-state 'argonly))
	    ((eq operation 'edit)
	     (redo-enter-capture-state))
	    ((eq operation 'stop)
	     nil)
	    ((eq operation 'redo)
	     nil)
	    ))
	  ((eq redo-state 'argonly)
	   (cond
	    ((eq operation 'arg)
	     nil)
	    ((eq operation 'edit)
	     (redo-enter-capture-state))
	    ((eq operation 'stop)
	     (setq redo-univarg-buffer nil
		   redo-state 'stop))
	    ((eq operation 'redo)
	     (redo-enter-capture-state))))
	  ((eq redo-state 'capturing)
	   (cond
	    ((eq operation 'arg)
	     (setq redo-state 'captarg))
	    ((eq operation 'edit)
	     nil)
	    ((eq operation 'stop)
	     (setq redo-state 'held))
	    ((eq operation 'redo)
	     nil)
	    ))
	  ((eq redo-state 'held)
	   (cond
	    ((eq operation 'arg)
	     (setq redo-univarg-buffer nil
		   redo-state 'heldarg))
	    ((eq operation 'edit)
	     (redo-enter-capture-state))
	    ((eq operation 'stop)
	     nil)
	    ((eq operation 'redo)
	     (setq redo-buffer nil
		   redo-state 'none))
	    ))
	  ((eq redo-state 'heldarg)
	   (cond
	    ((eq operation 'arg)
	     nil)
	    ((eq operation 'edit)
	     (redo-enter-capture-state))
	    ((eq operation 'stop)
	     (setq redo-state 'held))
	    ((eq operation 'redo)
	     (redo-enter-capture-state))
	    ))
	  ((eq redo-state 'captarg)
	   (cond
	    ((eq operation 'arg)
	     nil)
	    ((eq operation 'edit)
	     (redo-append-args-to-capture)
	     (setq redo-state 'capturing))
	    ((eq operation 'stop)
	     (setq redo-state 'held))
	    ((eq operation 'redo)
	     (redo-append-args-to-capture)
	     (setq redo-state 'capturing))
	    )))
	 (redo-set-cursor redo-state))))

(defun redo-pre-cmd-fn ()
  "Hook function to record buffer state before running a command."
  (if redo-allow-redo-recording  ;; prevent recursive recording during
                                 ;; playback of redo
      (redo-record-pre-cmd-state)))

(defun redo-operation-type (cmd)
  "Identify the type of operation that has just been performed."
  (if (eq cmd 'keyboard-quit)
      'quit
    (if (redo-stopper-fn-p cmd)
	'stop
      (if (member cmd '(universal-argument
			universal-argument-other-key
			universal-argument-minus
			universal-argument-more
			negative-argument
			digit-argument))
	  'arg
	(if (member cmd '(redo-redo redo-redo-pop))
	    'redo
	  (if (and (eq (current-buffer) redo-pre-buffer)
		   (or (redo-undo-updated-p cmd) 
		       (redo-mark-updated-p)))
	      'edit
	    nil))))))

(defun redo-capture (cmd)
  "Main function for capturing redo."
  (redo-change-state (redo-operation-type cmd))
  ;; What we do depends on our state.
  (if (eq redo-state 'capturing)
      ;; Add current stuff to capture
      (redo-add redo-redo-entry)
    ;; Not capturing - maybe reading args?
    (if (member redo-state '(captarg heldarg argonly))
	;; we must be in one of the arg reading states
	(redo-add-arg redo-redo-entry))))

(defun redo-post-cmd-fn ()
  "Hook function to possibly record the current command"
  (if redo-allow-redo-recording
      (redo-capture this-command)))

(defconst REDO-RING-SIZE 10
  "Size of redo-ring")

(defvar redo-ring (make-ring REDO-RING-SIZE)
  "Ring of the last N redo entries")

(defvar redo-ring-idx 0
  "Index into redo-ring")

(defun redo-redo-one-cmd (cmd)
  "This function redoes the single command in CMD."
  (cond
   ; This is the case when set by our commands commands
   ((equal (car cmd) "CMPL") (eval (cdr cmd)))

   ;; If the command was minibuffer-complete-and-exit then
   ;; the command was stored in the command-history
   ((equal (car cmd) "HIST") (eval (car command-history)))

   ;; Edit key.
   ((equal (car cmd) "KEY") 
    (execute-kbd-macro (nth 1 cmd)) nil)
   ((equal (car cmd) "RPT")
    (redo-redo-cmd-list (nth 2 cmd) (nth 1 cmd)))))

(defun redo-redo-cmd-list (list count)
  "Redo the given LIST of commands COUNT times."
  (while (> count 0)
    (mapcar 'redo-redo-one-cmd list)
    (setq count (- count 1))))

(defvar redo-undo-point nil
  "Records the position of point prior to redo.  This is to enable point
to be restored for redo-pop.")

(defun redo-undo ()
  "Undo the last redo."
  (interactive)
  (advertised-undo)
  (goto-char redo-undo-point))

(defun redo-close-undo-boundaries (undo-list stop-at)
  "Remove any undo-boundaries created during redo."
  (let ((prev undo-list)
	(list (cdr undo-list)))
    (while (and list (not (eq list stop-at)))
      (if (car list)
	  (setq prev list)
	(setcdr prev (cdr list)))
      (setq list (cdr list)))))

(defun redo-redo (arg)
  "Redo all the stuff from the current element of the redo-ring"
  (interactive "p")
  (if (eq redo-state 'held)
      (progn 
	;; Copy redo-state to the head of the redo-ring
	(ring-insert redo-ring redo-buffer)
	(setq redo-ring-idx 0)))

  (if (ring-empty-p redo-ring)
      (error "Nothing-to-redo"))

  (undo-boundary)

  ;; Perform the redo
  (setq redo-undo-point (point)
	redo-undo-list buffer-undo-list)
  (let ((buf (ring-ref redo-ring redo-ring-idx))
	(current-prefix-arg 1)
	(prefix-arg 1)
	(redo-allow-redo-recording nil)
	(count arg))
    (while (> count 0)
      (mapcar 'redo-redo-one-cmd buf)
      (setq count (- count 1))))

  ;; Eliminate any undo boundaries within the redo.
  (redo-close-undo-boundaries buffer-undo-list redo-undo-list)

  ;; Finally, we modify redo-redo-entry to contain the current redo.  This
  ;; allows redo to be captured, if appropriate.
  (setq redo-redo-entry (list "RPT" arg (ring-ref redo-ring redo-ring-idx))))


(defun redo-redo-pop (arg)
  "Pop the previous redo entry from the redo ring"
  (interactive "p")
  (if (member real-last-command (list 'redo-redo 'redo-redo-pop))
      (progn
	(redo-undo)
	(setq redo-ring-idx (ring-plus1 redo-ring-idx REDO-RING-SIZE))
	(redo-redo arg))
    ;; Previous command was not redo.  We take this to mean that redo
    ;; should not be considered a stopper this time.
    (redo-redo arg)))

(defun redo-enable ()
  "Enable redo."
  (redo-pre-cmd-fn)
  (add-hook 'pre-command-hook 'redo-pre-cmd-fn)
  (add-hook 'post-command-hook 'redo-post-cmd-fn))  

(defun redo-disable ()
  "Disable redo."
  (remove-hook 'pre-command-hook 'redo-pre-cmd-fn)
  (remove-hook 'post-command-hook 'redo-post-cmd-fn))

(defun redo-reset ()
  "Reset all redo state."
  (setq redo-ring (make-ring REDO-RING-SIZE))
  (setq redo-buffer nil))

(provide 'redo)

(define-key global-map [f12] 'redo-redo)
(define-key global-map [S-f12] 'redo-redo-pop)

;;;; debug stuff
(progn
  (redo-disable)
  (redo-reset)
  (redo-enable)
)



;;
;;(redo-post-cmd-fn)
;;redo-buffer    xx
;;redo-ring          xxxxx  x
;;redo-pre-buffer
 
;;;;
;;;;(defvar previous-command nil
;;;;  "This function is supplied to allow other functions to set 
;;;;the previous command. The previous command takes the form:-
;;;;(\"CMPL\" <func> <args1> <arg2> ...) -- Complex command
;;;;(\"KEY\" <command-keys> <func> )     -- Keyboard command
;;;;(\"HIST\" )                          -- History command
;;;;In the COMPLEX command the function is executed with the arguments. With
;;;;the KEY command the function is ignored and the function replicated by
;;;;`execute-kbd-macro' using the command-keys. The command keys are from the 
;;;;function `this-command-keys'")
;;;;
;;;;
;;;;
;;;;(defun prev-command-hook ()
;;;;  "This is the hook in to the previous command so that normal editing 
;;;;keys can be put into the redo last command"
;;;;  (interactive)
;;;;  (cond
;;;;   ;; Don't save the keyboard press if these were the last commands
;;;;   ((member this-command (list 'rerun-prev-command
;;;;			       'keyboard-quit
;;;;			       'exit-minibuffer)) t)
;;;;    
;;;;   ;; If the command was `minibuffer-complete-and-exit' then the
;;;;   ;; last command must be stored in the command history. So tell 
;;;;   ;; the rerun command to use this
;;;;   ((equal this-command 'minibuffer-complete-and-exit)
;;;;    (set-prev-cmd (list "HIST"))
;;;;   )
;;;;
;;;;   ;; Otherwise it must be just a normal key
;;;;   (t
;;;;    (set-prev-cmd (list "KEY" (this-command-keys) this-command))
;;;;   )
;;;;  )
;;;;)
;;;;
;;;;
;;;;(defun rerun-prev-command ()
;;;;  "This function reruns the last command. The command is defined in 
;;;;the variable previous-command"
;;;;  (interactive)
;;;;  (cond
;;;;
;;;;   ; This is the case when set by our commands commands
;;;;   ((equal (car previous-command) "CMPL") (eval (cdr previous-command)))
;;;;
;;;;   ;; If the command was minibuffer-complete-and-exit then
;;;;   ;; the command was stored in the command-history
;;;;   ((equal (car previous-command) "HIST") (eval (car command-history)))
;;;;
;;;;   ;; Otherwise it must have been a keyboard macro
;;;;   ((equal (car previous-command) "KEY") 
;;;;    (execute-kbd-macro (nth 1 previous-command))
;;;;   )
;;;;  )
;;;;)
;;;;
;;;;(defun set-prev-cmd ( lst )
;;;;  "This function is supplied to allow other functions to set 
;;;;the previous command. The previous command takes the form:-
;;;;(\"CMPL\" <func> <args1> <arg2> ...) -- Complex command
;;;;(\"KEY\" <command-keys> <func> )     -- Keyboard command
;;;;(\"HIST\" )                          -- History command
;;;;In the COMPLEX command the function is executed with the arguments. With
;;;;the KEY command the function is ignored and the function replicated by
;;;;`execute-kbd-macro' using the command-keys. The command keys are from the 
;;;;function `this-command-keys'"
;;;; (interactive)
;;;; (setq previous-command lst)
;;;; (print previous-command)
;;;;)
;;;;
;;;;(defun griffith-advertised-undo ()
;;;;  "This is a roof over the undo command `advertised-undo' that 
;;;;allows access so that the command can be trapped and fed into the 
;;;;previous command"
;;;;  (interactive)
;;;;  (advertised-undo)
;;;;  (set-prev-cmd (list "CMPL" 'griffith-advertised-undo))
;;;;)
;;;;
;;;;(defun griffith-call-last-kbd-macro ( &optional PREFIX )
;;;;  "This is a roof over the `call-last-kbd-macro' command to 
;;;;allows access so that the command can be trapped and fed into the 
;;;;previous command"
;;;;  (interactive "p")
;;;;  (call-last-kbd-macro PREFIX)
;;;;  (set-prev-cmd (list "CMPL" 'griffith-call-last-kbd-macro PREFIX))
;;;;)
;;;;
;;;;
;;;;
;;;;(define-key global-map [f12] 'rerun-prev-command)
;;;;
;;
