;; Provide cua mode but with better integration with menus and functions

(require 'cua-base)
(cua-mode t)

;; Some cua customisations.
(setq cua-prefix-override-inhibit-delay 0.01)

;; Eliminate some of the less useful key-bindings that would otherwise
;; be shown in the menus in preference to our improved bindings
(define-key global-map [?\C-w] nil)

;; Modify some of the cua keys
(define-key global-map [insert] 'cua-paste)
(define-key global-map [S-insert] 'cua-paste-pop)
(define-key global-map [C-insert] 'overwrite-mode)

  
;; Ensure that an interactive kill-region does not do anything when the
;; mark is not active 
(defadvice kill-region (before my-cua first activate)
  "Prevent kill-region from operating when mark is not active"
  (interactive (list (point) (mark)))
  (if (and (interactive-p) (not mark-active))
      (error "The Mark is not active now")))

;; Some more active region hacks
(defadvice upcase-word (around my-cua first activate)
  "If in transient mark-mode and mark is active, upcase the whole region."
  (if (and transient-mark-mode mark-active)
      (upcase-region (point) (mark))
    ad-do-it))

(defadvice downcase-word (around my-cua first activate)
  "If in transient mark-mode and mark is active, upcase the whole region."
  (if (and transient-mark-mode mark-active)
      (downcase-region (point) (mark))
    ad-do-it))

(defadvice capitalize-word (around my-cua first activate)
  "If in transient mark-mode and mark is active, upcase the whole region."
  (if (and transient-mark-mode mark-active)
      (capitalize-region (point) (mark))
    ad-do-it))

;; Better kill functions
(defadvice kill-line (around my-cua first activate)
  "Make kill-line kill entire lines when sensible"
  (let ((arg (ad-get-arg 0)))
    (if (and (null arg)
	     (not (eq (point) (line-beginning-position))))
	ad-do-it
      (ad-set-arg 0 1)
      ad-do-it)))

(defadvice kill-word (around my-cua first activate)
  "Make kill-word more sensible about deleting trailing spaces."
  ad-do-it
  (if (looking-at "[[:space:]]")
      (let ((my-mark (point)))
	(re-search-forward "[[:space:]]+")
	(kill-region my-mark (point)))))

;;  (let* ((cur-point (point))
;;	 (at-word-start
;;	  (save-excursion
;;	    (forward-char 1)
;;	    (backward-word 1)
;;	    (eq (point) cur-point))))
;;    ad-do-it
;;    (if at-word-start
;;	(progn
;;	  ;; Close-up the gap between point and the next word
;;	  (set-mark (point))
;;	  (forward-word 1)
;;	  (backward-word 1)
;;	  (kill-region (mark) (point))))))


;; Hacked-version of cua-delete-region which uses kill-region instead of
;; delete-region.  This allows more natural usage of yank and yank-pop
;; Since we have redo, the loss of the slightly odd cua-mode repeat operations
;; is not a big deal
(setq cua-delete-copy-to-register-0 nil)
(defun cua-delete-region ()
  "Delete the active region.
Save a copy in register 0 if `cua-delete-copy-to-register-0' is non-nil."
  (interactive)
  (let ((start (mark)) (end (point)))
    (or (<= start end)
	(setq start (prog1 end (setq end start))))
    (setq cua--last-deleted-region-text (filter-buffer-substring start end))
    (if cua-delete-copy-to-register-0
	(set-register ?0 cua--last-deleted-region-text))
    (kill-region start end)
    (setq cua--last-deleted-region-pos
	  (cons (current-buffer)
		(and (consp buffer-undo-list)
		     (car buffer-undo-list))))
    (cua--deactivate)
    (/= start end)))




;; Menu-bar hacks for new key bindings
(let*
    ((menu-bar-map (lookup-key global-map [menu-bar]))
     (edit-menu-map (lookup-key menu-bar-map [edit])))
     
  (define-key-after edit-menu-map [undo]
    '("Undo" . undo) 'copy)
  (define-key-after edit-menu-map [cut]
    '("Cut" . kill-region) 'undo)
  (define-key-after edit-menu-map [copy]
    '("Copy" . copy-region-as-kill) 'cut)
  (define-key-after edit-menu-map [paste]
    '("Paste" . yank) 'copy)
  
  ;; Ensure cut and copy are disabled in menu when region is not
  ;; active
  (put 'kill-region 'menu-enable 'mark-active)
  (put 'copy-region-as-kill 'menu-enable 'mark-active)
  )

(provide 'my-cua)