;;**********************************************************************;;
;;									;;
;;		E m a c s   P o w e r   S e a r c h			;;
;;		-----------------------------------   			;;
;;									;;
;;	Marc Munro					      		;;
;;									;;
;;	Emacs functions to provide all of emacs search and replace	;;
;;	commands on a few keys and to standardise the interface to      ;;
;;      each search and replace function so that they share a search    ;;
;;      history.                                                        ;;
;;									;;
;;	These functions are provided as the power-search feature.	;;
;;									;;
;;**********************************************************************;;

(defvar power-search-reverse-char (string-to-char "\C-R")
  "*Character to repeat incremental search backwards.  This appears not to
exist in version 19, so is defined in power-search.el")

(defvar power-search-yank-word-char (string-to-char "\C-W")
  "*Character to pull next word from buffer into search string.  This 
appears not to exist in version 19, so is defined in power-search.el")

(defvar power-search-types
  '(re-search-forward
    search-forward
    word-search-forward
    isearch-forward
    isearch-forward-regexp
    list-matching-lines)   
  "This identifies the set of search-types available to power-search, with
the default search-type at the head of the list.")

(defconst power-search-reverse-map
  '((re-search-forward      . re-search-backward)
    (search-forward         . search-backward)
    (word-search-forward    . word-search-backward)
    (isearch-forward        . isearch-backward)
    (isearch-forward-regexp . isearch-backward-regexp)
    (list-matching-lines    . list-matching-lines))
  "Mapping of forward search function to backwards one.")

(defconst power-search-prompt-map
  '((re-search-forward      . "Re-search for ")
    (search-forward         . "Search for ")
    (word-search-forward    . "Word-search for ")
    (isearch-forward        . "")
    (isearch-forward-regexp . "")
    (list-matching-lines    . "Lines matching: "))
  "Mapping of forward search functions to prompt strings.")

(defvar power-replace-types
  '(query-replace-regexp 
    replace-regexp
    query-replace
    replace-string)
  "This identifies the set of replace functions to be used by the 
power-replace function with the default value at the head of the list.")

(defconst power-replace-prompt-map
  '((query-replace-regexp . "Regexp replace: ")
    (replace-regexp       . "Regexp replace: ")
    (query-replace        . "Replace: ")
    (replace-string       . "Replace: "))
  "Mapping of replace functions to search prompt strings.")

(defvar power-search-cur-prompt "Search for "
  "Prompt to be used when reading the search string.  This variable should
not normally be used, instead an overriding instance should be declared in
a local stack-frame.")

(defvar power-search-is-forward t
  "Boolean describing whether searches are in a forward or backward direction.
t means forward.")

(defvar power-search-history nil
  "History to be used by search functions.")

(defvar power-replace-history nil
  "History to be used by replace functions.")

(defvar power-search-match-point nil
  "The value of point at the last matching line found by next-match.")

(defvar power-search-string-minibuffer-map
  (copy-keymap minibuffer-local-completion-map)
  "This is a copy of the minibuffer completion keymap with small modifications
to allow toggle direction and word yanking to be performed.")

(defvar power-search-cur-buffer nil
  "The buffer from which the search function was invoked.  Used by
power-search-yank functions.  This variable should not actually be used as
such, but rather should be defined as a local variable in each search
function.  It exists as a named variable solely to provide this documentation
string.")

; Modify keymap to allow yanking of words from current buffer and changing
; the query direction just like incremental search, and to allow the 
; power-search function key to terminate minbuffer input so that pressing
; it twice will repeat the last search.

(mapcar
 (function
  (lambda (keyseq)
    (define-key power-search-string-minibuffer-map keyseq 
      'power-search-toggle-direction)))
 (list (char-to-string power-search-reverse-char)))

(mapcar
 (function
  (lambda (keyseq)
    (define-key power-search-string-minibuffer-map keyseq 
                                            'power-search-yank-word)))
 (list (char-to-string power-search-yank-word-char)))

(mapcar
 (function
  (lambda (keyseq)
    (define-key power-search-string-minibuffer-map keyseq 'exit-minibuffer)))
 (where-is-internal 'power-search))

(define-key power-search-string-minibuffer-map " " nil
  ;; remove space as a completion command key
)


(defun power-search-yank-word ()
  "Get the next word from the default buffer and copy it into the 
search string."
  (interactive)
  (let ((start)
	(end)
	(word)
	(cur-buf (current-buffer)))
    (set-buffer power-search-cur-buffer)
    (setq start (point))
    (forward-word 1)
    (setq end (point))
    (setq word (buffer-substring start end))
    (set-buffer cur-buf)
    (insert word)))

(defun power-search-display-case ()
  "This function describes whether search and replace operations are 
case-sensitive."
  (interactive)
  (if case-fold-search
      (princ " (case independent)")
    (princ " (case dependent)")))

(defun power-search-display-state ()
  "This function describes which search mechanism is currently set up as the
power-search function."
  (interactive)
  (let* ((search-mode (car power-search-types))
	 (reverse-mode (cdr (assoc search-mode power-search-reverse-map))))
    (if power-search-is-forward
	(princ search-mode)
      (princ reverse-mode))
    (power-search-display-case)))

(defun power-search-switch-mode ()
  "This function switches the current search mode for the power-search 
function."
  (interactive)
  (setq power-search-types (append (cdr power-search-types)
				   (cons (car power-search-types) nil)))
  (power-search-display-state))

(defun power-search-toggle-case-dependence ()
  "This function inverts the current case-sensitivity of search and replace
operations."
  (interactive)
  (setq case-fold-search (not case-fold-search))
  (power-search-display-case))

(defun power-search-toggle-direction ()
  "This function toggles the direction in which most of the power-search 
functions operate."
  (interactive)
  (setq power-search-is-forward (not power-search-is-forward))
  (power-search-display-state))

(defvar power-search-string-last ""
  "Value of last string returned from search.")

(defvar power-replace-string-last ""
  "Value of last string returned from replace.")

(defun power-search-read (prompt history default)
  "Perform a completing read, prompting with PROMPT, and providing options
from HISTORY.  DEFAULT does what you might expect."
  (let ((save-minibuffer-map minibuffer-local-completion-map))
    (unwind-protect
	(progn
	  (setq minibuffer-local-completion-map
		power-search-string-minibuffer-map)
	  (cons
	   (completing-read (concat prompt "["
				    default "]: ")
			    (mapcar 'list history)
			    nil nil 
			    nil
			    (let ((len (length history)))
			      (cond ((= 0 len)
				     'history)
				    ((= 1 len)
				     '(history . 0))
				    (t '(history . 1))))
			    default)
	   history))
      (setq minibuffer-local-completion-map save-minibuffer-map))))

(defun power-search-get-string ()
  "Function to get a search string using the history mechanism."
  (let ((result 
	 (power-search-read power-search-cur-prompt 
			    power-search-history power-search-string-last)))
    (setq power-search-history 
	  (power-search-tidy-history (cdr result)))
    (setq power-search-string-last (car result))))


(defadvice re-search-forward (before power-search first activate)
  "Use power-search get-search-string function to interact with the user."
  (interactive 
   (let ((power-search-cur-buffer (current-buffer))
	 (power-search-cur-prompt (cdr (assoc 're-search-forward
					      power-search-prompt-map))))
     (list (power-search-get-string)))))

(defadvice re-search-backward (before power-search first activate)
  "Use power-search get-search-string function to interact with the user."
  (interactive 
   (let ((power-search-cur-buffer (current-buffer))
	 (power-search-cur-prompt (cdr (assoc 're-search-forward
					      power-search-prompt-map))))
     (list (power-search-get-string)))))

(defadvice word-search-forward (before power-search first activate)
  "Use power-search get-search-string function to interact with the user."
  (interactive 
   (let ((power-search-cur-buffer (current-buffer))
	 (power-search-cur-prompt (cdr (assoc 'word-search-forward
					      power-search-prompt-map))))
     (list (power-search-get-string)))))

(defadvice word-search-backward (before power-search first activate)
  "Use power-search get-search-string function to interact with the user."
  (interactive 
   (let ((power-search-cur-buffer (current-buffer))
	 (power-search-cur-prompt (cdr (assoc 'word-search-forward
					      power-search-prompt-map))))
     (list (power-search-get-string)))))

(defadvice search-forward (before power-search first activate)
  "Use power-search get-search-string function to interact with the user."
  (interactive 
   (let ((power-search-cur-buffer (current-buffer))
	 (power-search-cur-prompt (cdr (assoc 'search-forward
					      power-search-prompt-map))))
     (list (power-search-get-string)))))

(defadvice search-backward (before power-search first activate)
  "Use power-search get-search-string function to interact with the user."
  (interactive 
   (let ((power-search-cur-buffer (current-buffer))
	 (power-search-cur-prompt (cdr (assoc 'search-forward
					      power-search-prompt-map))))
     (list (power-search-get-string)))))

(defadvice list-matching-lines (around power-search first activate)
  "Use power-search get-search-string function to interact with the user."
  (interactive 
   (let ((power-search-cur-buffer (current-buffer))
	 (power-search-cur-prompt (cdr (assoc 'list-matching-lines
					      power-search-prompt-map))))
     (list (power-search-get-string))))
  (let ((power-search-cur-buffer (current-buffer)))
    (setq power-search-match-point nil)
    ad-do-it))

(defadvice query-replace-regexp (before power-search first activate)
  "Use power-search get-search-string and get-replace-string functions
to interact with the user."
  (interactive 
   (let ((power-search-cur-buffer (current-buffer))
	 (power-search-cur-prompt (cdr (assoc 'query-replace-regexp
					      power-replace-prompt-map))))
     (list (power-search-get-string)
	   (power-replace-get-string)))))

(defadvice query-replace (before power-search first activate)
  "Use power-search get-search-string and get-replace-string functions
to interact with the user."
  (interactive 
   (let ((power-search-cur-buffer (current-buffer))
	 (power-search-cur-prompt (cdr (assoc 'query-replace
					      power-replace-prompt-map))))
     (list (power-search-get-string)
	   (power-replace-get-string)))))

(defadvice replace-regexp (before power-search first activate)
  "Use power-search get-search-string and get-replace-string functions
to interact with the user."
  (interactive 
   (let ((power-search-cur-buffer (current-buffer))
	 (power-search-cur-prompt (cdr (assoc 'replace-regexp
					      power-replace-prompt-map))))
     (list (power-search-get-string)
	   (power-replace-get-string)))))

(defadvice replace-string (before power-search first activate)
  "Use power-search get-search-string and get-replace-string functions
to interact with the user."
  (interactive 
   (let ((power-search-cur-buffer (current-buffer))
	 (power-search-cur-prompt (cdr (assoc 'replace-string
					      power-replace-prompt-map))))
     (list (power-search-get-string)
	   (power-replace-get-string)))))

(defun power-replace-get-string ()
  "Function to get a replace string using the history mechanism."
  (let ((save-minibuffer-map minibuffer-local-completion-map))
    (unwind-protect
	(progn
	  (setq minibuffer-local-completion-map
		power-search-string-minibuffer-map)
	  (setq power-replace-string-last
		(completing-read "Replace with "
				 (mapcar 'list power-replace-history)
				 nil nil power-replace-string-last
				 (let ((len (length power-replace-history)))
				   (cond ((= 0 len)
					  nil)
					 ((= 1 len)
					  '(power-replace-history . 0))
					 (t '(power-replace-history . 1)))))))
      (setq minibuffer-local-completion-map save-minibuffer-map)
      (setq power-replace-history 
	    (power-search-tidy-history power-replace-history))
      "")))

(defun power-search-tidy-history (history)
  "Remove duplicates and empty elements from the history.  Duplicates will
be left as the latest element with previous versions removed."
  (if (string= "" (car history))
      (cdr history)
    (let* ((first (car history))
	   (cur-elem history)
	   (next-elem (cdr history)))
      (while next-elem
	(if (string= first (car next-elem))
	    (setcdr cur-elem (cdr next-elem)))
	(setq cur-elem (cdr cur-elem))
	(setq next-elem (cdr cur-elem))))
    history))
	  

(defun power-search ()
  "This function provides access to all emacs search functions, based on the 
value of search-type.  Non-incremental searches may yank words from the
current-buffer using C-w."
  (interactive)
  (let* ((search-mode (car power-search-types))
	 (reverse-mode (cdr (assoc search-mode power-search-reverse-map))))
    (if power-search-is-forward
	(call-interactively search-mode)
      (call-interactively reverse-mode))))

(defun power-replace-display-state ()
  "This function displays the current replace operation to be used by the
power-replace function based on the value of replace-type."
  (interactive)
  (princ (car power-replace-types))
  (power-search-display-case))

(defun power-replace-switch-mode ()
  "This function selects the current replace mode for the power-replace 
function."
  (interactive)
  (setq power-replace-types (append (cdr power-replace-types)
				    (cons (car power-replace-types) nil)))
  (power-replace-display-state))

(defun power-replace ()
  "This function provides access to all emacs replace functions, based on the 
value of replace-type."
  (interactive)
  (call-interactively (car power-replace-types)))

(defun power-search-next-match ()
  "This function finds the next match from the *Occur* buffer generated by 
the list-matching-lines function.  The function returns nil, if there are no 
more matches or the *Occur* buffer doesn't exist anyway.  The user can select
a specific occurrence by moving point to that line."
  (interactive)
  (if (get-buffer "*Occur*")
      (progn
	(if (not (string= (buffer-name) "*Occur*"))
	    (set-buffer "*Occur*"))
	(if (null power-search-match-point)
	    (forward-line 1)) ; Move to first match
	(if (> (count-lines (point) (point-max)) 0)
	    (occur-mode-goto-occurrence))
	(save-excursion
	  (set-buffer "*Occur*")
	  (forward-line 1)      ; Move to the next line.
	  (setq power-search-match-point (point))))
    nil))

(defun power-search-next-error ()
  "Attempts to use next-match to find next matching line in *Occur* buffer
before attempting to perform a normal next-error function."
  (interactive)
  (if (not (power-search-next-match))
      (next-error)))

(provide 'power-search)

