;; redo.el
;; Provide an automated redo capture system.  This is a bit like having 
;; keyboard macros being recorded at all times and having immediate access
;; to them.  It is also quite a bit different from keyboard macros.
;;
;; Redo is recorded in redo groups.  A redo group begins with an
;; interactive edit (changes to the buffer by minibuffer-invoked
;; commands do not begin a redo group).  The redo group is captured when
;; a non-edit operation moves point to a different line.  The last
;; element of a redo-group will always be an edit operation, never a
;; cursor movement.  Redo-groups are terminated (and not captured) by
;; quit, switching buffers or frames, opening or closing frames or
;; windows, or by any undo that undoes changes made outside of the scope
;; of the redo group.
;;
;; Redo can be replayed by the redo-redo command.  Previous redo can be
;; selected after redo-redo using redo-pop, and the next can be selected
;; using redo-depop.
;;
;; TODO: 
;; Funny keybindings for dot and minus
;; customisation stuff
;; Documentation

(require 'ringset)
(require 'rollback)
(require 'watched-undo)

(define-key global-map "\C-r" 'redo-redo)
(define-key global-map [f12] 'redo-pop)

(defvar redo-state 'quiescent
  "The current state of redo capture.  The states are:
'quiescent - redo is not being captured
'argonly   - we have captured only a universal argument
'capturing - we are currently capturing redo
'captarg   - we have begun capturing a universal argument from the capturing
             state
'minibuf   - we are doing something in a minibuffer (from a quiescent state)
")

;; Transition from one state to another will be for the following reasons:
;; arg     - universal argument has been entered
;; edit    - an edit operation has commenced
;; stop    - a stop function has been encountered
;; quit    - ctrl-g has been pressed
;; minibuf - we have entered a minibuffer

;; Note that any function that switches from the initial buffer to 
;; another buffer, other than a minibuffer, is a stop function.  Also,
;; generally, any non-edit operation that moves point to another line
;; is also a stop function.

(defvar redo-key-buffer nil
  "Variable into which our current redo is collected.  This is of type keyseq,
which is a list of type (n elems) where n is an integer describing how many 
times to repeat elems list, and elems is a list of elements each of which may
be a string, vector or keyseq.")

(defvar redo-dullkeys-buffer nil
  "Records key sequences that do not update the current buffer.  These may be 
small cursor movements, for instance.  We do not record such keys unless they
are followed by a true edit operation.")

(defvar redo-minibufkeys-buffer nil
  "Records key sequences within the minibuffer.  On exit from the minibuffer,
these will be appended to redo-dullkeys-buffer.")

(defvar redo-command-keys nil
  "Variable into which the keys that invoked the current command are recorded.
This is a keyseq element type (ie string, vector or keyseq)")

(defvar redo-active-buffer nil
  "The buffer in which we are collecting redo.")

(defvar redo-prev-mark nil
  "The mark at the start of the command.")
	
(defvar redo-prev-point 1
  "The value of point at the start of the command.")
	
(defvar redo-prev-tick 1
  "The buffer-modified-tick at the start of the command.")
	
(defvar redo-prev-window-config (current-window-configuration)
  "Variable used to record the window configuration of the current frame
when we start capturing redo.")

(defvar redo-prev-frame nil
  "Variable used to identify the frame at the start of a command.")

(defconst redo-allow-redo-recording t
  "This allows the recording of redo information.  It is overridden locally 
during redo functions to prevent recursive re-recording of redo.")

(defvar redo-stoppers
  (list 'minibuffer-complete-and-exit 'save-buffer)
  "List of functions that stop the accumulation of redo.  These are known
as stopper or stop functions.")

(defvar redo-conditional-stoppers
  (list 'next-line 'previous-line 'end-of-buffer 'beginning-of-buffer
	'scroll-down 'scroll-up 'scroll-up-in-place 'scroll-down-in-place)
  "List of functions that may stop the accumulation of redo.  Whether
they do so will depend on whether we have marked something as part of the
redo-able operation.")

(defvar redo-ring (make-ring 10 'equal)
  "Ringset of the last N distinct redo entries")

(defvar redo-last-edit-was-redo nil
  "Tells us whether the last edit operation was a redo.  It is set to
true when a redo function is performed, and is reset to nil by quit,
window changes, etc, or whenever an edit operation occurs that is not
a redo.  The purpose of this variable is to allow a wildkey to be
implemented.")
 


;; Edit cursor objects  These manage cursor type changes when redo is 
;; being captured.
(defvar redo-edit-cursor-buffer nil
  "The buffer in which the edit cursor was set")

(defun redo-reset-edit-cursor ()
  (save-excursion
    (if redo-edit-cursor-buffer
	(set-buffer redo-edit-cursor-buffer))
    (setq cursor-type t
	  redo-edit-cursor-buffer nil)))

(defun redo-set-edit-cursor ()
  (redo-reset-edit-cursor)
  (setq cursor-type 'bar
	redo-edit-cursor-buffer (current-buffer)))
  
(defun redo-set-cursor ()
  "Set cursor in accordance with state."
  (cond
   ((member redo-state '(capturing captarg)) 
    (redo-set-edit-cursor))
   ((member redo-state '(quiescent argonly minibuf))
    (redo-reset-edit-cursor))))


;; Miscelaneous functions
(defun redo-empty-keyseq ()
  "Return an empty redo keyseq.  See description in redo-key-buffer."
  (list 1 ""))

(defun redo-keyseq-empty-p (keyseq)
  "Predicate to determine if KEYSEQ is empty."
  (let ((elem (cadr keyseq)))
    (and (stringp elem)
	 (string= elem ""))))

(defun redo-fetch-from-ring ()
  "Return the currently selected redo"
  (ring-ref redo-ring redo-ring-idx))

(defun redo-record-window-config ()
  "Record the window configuration of the current frame."
  (setq redo-prev-window-config (current-window-configuration)))

;; Functions for monitoring undo
(defvar redo-excess-undo nil
  "Variable will be set true when undo exceeds captured edits")

(defun redo-set-excess-undo ()
  "Set redo-excess-undo."
  (setq redo-excess-undo t))

(defun redo-set-undo-watch ()
  "Set an undo watchpoint."
  (setq redo-excess-undo nil)
  (set-undo-watchpoint 'redo-set-excess-undo))


;; Functions for determining the type of operation just performed
(defun redo-active-mark-p ()
  "t if we are in transient-mark-mode and the mark is active"
  (and transient-mark-mode mark-active))

(defun redo-mark-updated-p ()
  "Did this command modify the mark?"
  (not (equal redo-prev-mark (mark-marker))))

(defun redo-buffer-modified-p ()
  "Predicate to determine whether the current buffer has been modified."
  (not (eq redo-prev-tick (buffer-modified-tick))))

(defun redo-redo-cmd-p (fn)
  "Predicate to identify whether FN is one of the redo functions."
  (member fn '(redo-redo redo-pop redo-depop)))

(defun redo-quit-if-excess-undo (fn)
  "If we have more undo than we have captured edits, or if we are quiescent
and FN is undo, return 'quit"
  (or (and (eq redo-state 'quiescent)
	   (eq fn 'undo))
      (and redo-excess-undo 'quit)))

(defun redo-quit-if-quit-fn (fn)
  "If FN is keyboard-quit, return 'quit."
  (and (member fn '(keyboard-quit abort-recursive-edit 
		    exit-recursive-edit mkacro-keyboard-quit))
       'quit))

(defun redo-minibuf-if-in-minibuf ()
  "If we are in the minibuffer, return 'minibuf"
  (and (string-match " \*Minibuf.*" (buffer-name (current-buffer)))
       'minibuf))

(defun redo-stop-if-minibuf-exit ()
  "If we have just exited the minibuffer, return 'stop"
  (and (eq redo-state 'minibuf)
       (not (string-match " \*Minibuf.*" (buffer-name (current-buffer))))
       'stop))

(defun redo-stop-if-stopper-fn (fn)
  "If FN is a stopper function, return 'stop"
  (and (or (member fn redo-stoppers)
	   (and (not (redo-active-mark-p))
		(member fn redo-conditional-stoppers)))
       'stop))

(defun redo-quit-if-buffer-changed ()
  "Return 'quit if we changed buffers"
  (and (not (eq redo-active-buffer (current-buffer)))
       'quit))

(defun redo-quit-if-windows-changed ()
  "If the current window configuration has changed since the start of
capturing, and we are not in a minibuffer, return 'quit."
  (if (string-match " \*Minibuf.*" (buffer-name (current-buffer)))
      nil
    (if (compare-window-configurations 
	 redo-prev-window-config (current-window-configuration))
	nil
      'quit)))

(defun redo-quit-if-frame-changed ()
  "Return 'quit if we changed frames."
  (and (not (eq redo-prev-frame (selected-frame)))
       'quit))

(defun redo-arg-if-arg-fn (fn)
  "If FN is a universal argument function, return 'arg"
  (and (member fn '(universal-argument
		    universal-argument-other-key
		    universal-argument-minus
		    universal-argument-more
		    negative-argument
		    digit-argument))
       'arg))

(defun redo-edit-if-edited ()
  "If we updated the buffer or set a mark, return 'edit"
  (and (or (redo-buffer-modified-p) 
	   (redo-mark-updated-p))
       'edit))

(defun redo-stop-if-moved-line ()
  "If we have moved lines within the buffer without an active mark, return 
'stop"
  (and (not (redo-active-mark-p))
       (> (count-lines (point) redo-prev-point) 1)
       'stop))

(defun redo-operation-type (fn)
  "Identify the type of operation that has just been performed by FN."
  (or (redo-quit-if-quit-fn fn)
      (redo-stop-if-stopper-fn fn)
      (redo-quit-if-excess-undo fn)
      (redo-minibuf-if-in-minibuf)
      (redo-quit-if-buffer-changed)
      (redo-quit-if-frame-changed)
      (redo-quit-if-windows-changed)
      (redo-arg-if-arg-fn fn)
      (redo-edit-if-edited)
      (redo-stop-if-moved-line)))


  
;; Functions for recording redo
(defun redo-save-redo-in-ring ()
  "Record the current redo-key-buffer"
  (ring-insert redo-ring redo-key-buffer)
  (setq redo-ring-idx 0))

(defun redo-append (keyseq elem)
  "Concatentate keyseq KEYSEQ with ELEM, returning KEYSEQ."
  (let* ((last2 (last keyseq 2))
	 (last (cdr last2))
	 (lastelem (car last)))
    (if (and (stringp lastelem)
	     (string= lastelem ""))
	;; The last element is an empty string.  Replace it with copy of elem.
	(if (and (listp elem)
		 (eq 1 (car elem)))
	    ;; If elem is a keyseq with count 1, copy its elements.  This
	    ;; enables us to have an element consisting solely of redo
	    ;; match the redo from which it was derived.
	    (setcdr last2 (append (cdr elem) nil))
	  ;; Simply replace the empty string with elem
	  (setcar last elem))
      ;; Last element was not empty string
      (if (and (arrayp elem) 
	       (arrayp (car last))
	       (eq (type-of elem) (type-of (car last))))
	  ;; If last elem and new elem are compatible array types,
	  ;; concatenate them
	  (setcar last 
		  (if (and (stringp lastelem) (stringp elem))
		      (concat lastelem elem)
		    (vconcat lastelem elem)))
	;; Otherwise, append the new seq to the end of our list
	(setcdr last (list elem)))))
  keyseq)

(defvar redo-keybuffer-savepoint nil
  "Savepoint of redo-key-buffer recorded for use by redo-pop.  This is 
a list of the form (<last element> <next to last cons cell>")

(defun redo-get-savepoint (keyseq)
  "Record current state of KEYSEQ so that it can be restored later by 
redo-restore-to-savepoint"
  (let ((last (last keyseq 2)))
    (cons (cadr last) last)))

(defun redo-restore-to-savepoint (savepoint)
  "Restore the keyseq from which SAVEPOINT was taken, to its saved state."
  (let ((elem (car savepoint))
	(cell (cdr savepoint)))
    (setcdr cell (cons elem nil))))



;; State transition functions
(defun redo-enter-quiescent-state (redo-complete)
  "Place redo into the quiescent state.  If REDO-COMPLETE we record the
current redo into the redo ring."
  (if redo-complete
      (redo-save-redo-in-ring))
  (setq redo-key-buffer (redo-empty-keyseq)
	redo-dullkeys-buffer (redo-empty-keyseq)
	redo-minibufkeys-buffer (redo-empty-keyseq)
	redo-state 'quiescent)
  (redo-set-cursor))

(defun redo-enter-argonly-state ()
  "Place redo into the capture state."
  (setq redo-state 'argonly)
  (redo-set-cursor))

(defun redo-enter-capturing-state ()
  "Place redo into the capturing state."
  (setq redo-state 'capturing)
  (redo-set-cursor))

(defun redo-enter-minibuf-state ()
  "Place redo into the minibuf state."
  (setq redo-minibufkeys-buffer nil
	redo-state 'minibuf)
  (redo-set-cursor))

(defun redo-enter-captarg-state ()
  "Place redo into the captarg state."
  (setq redo-state 'captarg)
  (redo-set-cursor))

(defun redo-change-state (operation)
  "Based on the current state and the operation, change state, recording
appropriate data as we go"
  (cond 
   ((eq redo-state 'minibuf)
    (cond
     ((eq operation 'edit) 
      (setq redo-minibufkeys-buffer (redo-empty-keyseq))
      (redo-enter-capturing-state))
     ((eq operation 'arg) 
      (setq redo-minibufkeys-buffer (redo-empty-keyseq))
      (redo-enter-argonly-state))
     ((eq operation 'stop) (redo-enter-quiescent-state nil))
     ((eq operation 'quit) (redo-enter-quiescent-state nil))))
   ((eq redo-state 'quiescent)
    (cond
     ((eq operation 'arg) (redo-enter-argonly-state))
     ((eq operation 'edit) (redo-enter-capturing-state))
     ((eq operation 'minibuf) (redo-enter-minibuf-state))))
   ((eq redo-state 'argonly)
    (cond
     ((member operation '(stop quit)) (redo-enter-quiescent-state nil))
     ((eq operation 'minibuf) (redo-enter-minibuf-state))
     ((eq operation 'edit) (redo-enter-capturing-state))
     (;; Must be nil operation
      t (redo-enter-quiescent-state nil))))
   ((eq redo-state 'capturing)
    (cond
     ((eq operation 'stop) (redo-enter-quiescent-state t))
     ((eq operation 'quit) (redo-enter-quiescent-state nil))
     ((eq operation 'arg) (redo-enter-captarg-state))))
   ((eq redo-state 'captarg)
    (cond
     ((member operation '(edit minibuf)) (redo-enter-capturing-state))
     ((eq operation 'stop) (redo-enter-quiescent-state t))
     ((eq operation 'quit) (redo-enter-quiescent-state nil)))))
)

;; Pre and post command state recording/checking functions
(defun redo-record-pre-cmd-state ()
  "Record the buffer state prior to command execution."
  (if (eq redo-state 'quiescent) 
      (redo-set-undo-watch))
  (setq redo-command-keys (this-command-keys)
	redo-prev-tick (buffer-modified-tick)
	redo-prev-tick (buffer-modified-tick)
	redo-prev-point (point)
	redo-prev-mark (copy-marker (mark-marker))
	redo-prev-frame (selected-frame))
  (unless (redo-minibuf-if-in-minibuf)
    (setq redo-active-buffer (current-buffer))))

(defvar redo-univarg-count 1
  "Contains value of universal argument given to redo-redo.  This will be
set by redo-redo.")

(defun redo-post-cmd-process (cmd)
  "Main function for capturing redo."
  (let ((prev-state redo-state)
	(operation (redo-operation-type cmd)))
     (redo-change-state operation)
    ;; Update redo-last-edit-was-redo
    (if (eq operation 'edit)
	(setq redo-last-edit-was-redo (redo-redo-cmd-p cmd))
      (if (eq operation 'quit)
	  (setq redo-last-edit-was-redo nil)))

    ;; What we do depends on our state.
    (if (eq redo-state 'capturing)
	(cond
	 ((null operation)
	  ;; Record the latest happenings in redo-dullkeys-buffer
	  (if (redo-keyseq-empty-p redo-minibufkeys-buffer)
	      (redo-append redo-dullkeys-buffer redo-command-keys)
	    (redo-append redo-dullkeys-buffer redo-minibufkeys-buffer)
	    (setq redo-minibufkeys-buffer (redo-empty-keyseq))))
	 ((eq operation 'minibuf)
	  (redo-append redo-minibufkeys-buffer redo-command-keys))
	 (t
	  ;; We are performing an edit in the capturing state. 
	  (unless (redo-keyseq-empty-p redo-dullkeys-buffer)
	    ;; Record any non-edit operations recorded in redo-dullkeys-buffer
	    (redo-append redo-key-buffer redo-dullkeys-buffer)
	    (setq redo-dullkeys-buffer (redo-empty-keyseq)))
	  
	  (if (redo-keyseq-empty-p redo-minibufkeys-buffer)
	      ;; Not a minibuffer action, so just record the keys
	      (redo-append redo-key-buffer redo-command-keys)
	    ;; Was a minibuffer action so record the minibuffer action.
	    ;; Also set the edit cursor as it may have been reset somehow
	    (redo-set-cursor)
	    (redo-append redo-key-buffer redo-minibufkeys-buffer)
	    (setq redo-minibufkeys-buffer (redo-empty-keyseq))))))
      (redo-record-window-config)))

;; Hook functions
(defun redo-pre-cmd-fn ()
  "Hook function to record buffer state before running a command."
  (if redo-allow-redo-recording  ;; prevent recursive recording during
                                 ;; playback of redo
      (redo-record-pre-cmd-state)))

(defun redo-post-cmd-fn ()
  "Hook function to possibly record the current command"
  (if redo-allow-redo-recording
      (redo-post-cmd-process this-command)))



;; Redo management functions
(defun redo-enable ()
  "Enable redo."
  (redo-enter-quiescent-state nil)
  (redo-pre-cmd-fn)
  (add-hook 'pre-command-hook 'redo-pre-cmd-fn)
  (add-hook 'post-command-hook 'redo-post-cmd-fn)
  (redo-enter-quiescent-state nil))

(defun redo-disable ()
  "Disable redo."
  (redo-reset-edit-cursor)
  (remove-hook 'pre-command-hook 'redo-pre-cmd-fn)
  (remove-hook 'post-command-hook 'redo-post-cmd-fn))

;; Functions that apply redo
(defun redo-redo-one-seq (seq)
  "This function redoes the single sequence in SEQ."
  (if (listp seq)
      (redo-redo-keyseq seq)
    (execute-kbd-macro seq nil)))

(defun redo-redo-keyseq (keyseq)
  "Redo the key sequence KEYSEQ"
  (let ((count (car keyseq))
	(seq (cdr keyseq)))
  (while (> count 0)
    (mapcar 'redo-redo-one-seq seq)
    (setq count (1- count)))))

(defvar redo-undo-savepoint nil
  "Savepoint of undo recorded for use by redo-pop")

(defvar redo-univarg 1
  "Value of universal argument from redo-redo for use by redo-pop")

(defun redo-set-this-command-as-redo()
  "Hook function to allow execute-kbd-macro to terminate abnormally while
still setting this-command appropriately"
  (setq this-command this-redo-command))

(defun redo-redo (count)
  "Redo all the stuff from the current element of the redo-ring COUNT times"
  (interactive "p")
  (if (ring-empty-p redo-ring)
      (error "Nothing-to-redo"))
  (setq redo-univarg-count count
	redo-ring-idx 0)
  (let* ((keyseq (redo-fetch-from-ring))
	 (current-prefix-arg 1)
	 (prefix-arg 1)
	 (redo-allow-redo-recording nil)
	 (this-redo-command 'redo-redo)
	 (kbd-macro-termination-hook nil))
    (add-hook 'kbd-macro-termination-hook 'redo-set-this-command-as-redo)
    (if (> count 1)
	;; Make a copy of the current redo, with the repetition arg
	;; set appropriately
	(setq keyseq (cons (* count (car keyseq)) (cdr keyseq))))
    (setq redo-undo-savepoint (savepoint)
	  redo-keybuffer-savepoint (redo-get-savepoint redo-key-buffer)
	  redo-command-keys keyseq)
    (undo-boundary)
    (redo-redo-keyseq keyseq)
    (coalesce-to-savepoint redo-undo-savepoint))
  (setq this-command 'redo-redo
	redo-univarg-count count)) ;; Reset this-command 

(defun redo-replace (this-redo-command)
  "Replace the last redo wit the current entry from the redo-ring,
with this-command set to THIS-REDO-COMMAND."
  (let* ((keyseq (redo-fetch-from-ring))
	 (current-prefix-arg 1)
	 (prefix-arg 1)
	 (redo-allow-redo-recording nil)
	 (kbd-macro-termination-hook nil))
    (add-hook 'kbd-macro-termination-hook 'redo-set-this-command-as-redo)
    ;; Restore buffer to pre-redo state
    (rollback redo-undo-savepoint)
    (setq redo-undo-savepoint (savepoint))
    
    ;; Restore redo-key-buffer to pre-redo-state
    (redo-restore-to-savepoint redo-keybuffer-savepoint)
    
    (undo-boundary)
    (if (> redo-univarg-count 1)
	;; Make a copy of the current redo, with the repetition arg
	;; set appropriately
	(setq keyseq (cons (* redo-univarg-count (car keyseq)) 
			   (cdr keyseq))))
    (redo-redo-keyseq keyseq)
    (coalesce-to-savepoint redo-undo-savepoint)
    (setq redo-command-keys keyseq)
    (setq this-command this-redo-command))) ;; Reset this-command 

(defun redo-pop ()
  "Replace the last redo, with the previous entry from the redo ring."
  (interactive)
  (if (not (redo-redo-cmd-p last-command))
      (error "Last command was not redo"))

  ;; Select to next element in ringset
  (setq redo-ring-idx (1+ redo-ring-idx))
  (redo-replace 'redo-pop))
	
(defun redo-depop ()
  "Replace the last redo, with the next entry from the redo ring."
  (interactive)
  (if (not (redo-redo-cmd-p last-command))
      (error "Last command was not redo"))

  ;; Select to next element in ringset
  (setq redo-ring-idx (1- redo-ring-idx))
  (redo-replace 'redo-depop))
	



;;
;;
;;(defun redo-equivalent-undo-p (undo1 undo2)
;;  "Compare UNDO1 and UNDO2, returning t if they are equivalent."
;;  ;; TODO: figure out a better way to do this.
;;  (while (and undo1 (not (car undo1)))
;;    ;; step past any undo boundaries
;;    (setq undo1 (cdr undo1)))
;;  (while (and undo2 (not (car undo2)))
;;    ;; step past any undo boundaries
;;    (setq undo2 (cdr undo2)))
;;  (eq (cdr undo1) (cdr undo2)))
;;
;;(defun redo-undo-updated-p (cmd)
;;  "Did this command begin an editing sequence?  It is if it has caused 
;;undo to be modified and is not itself an undo command."
;;  ;; Comparing cdr of the undo lists appears to be the safe way to
;;  ;; tell whether or not we modified undo
;;  ;; TODO rewrite to make it cleaner.
;;  (let ((is-undo (member cmd '(undo advertised-undo))))
;;    (and (not is-undo) 
;;	 (not (redo-equivalent-undo-p redo-pre-undo buffer-undo-list)))))
;;
;;(defun redo-mark-updated-p ()
;;  "Did this command modify the mark?"
;;  (not (and (equal redo-pre-mark (mark-marker))
;;	    (equal redo-pre-mark-ring mark-ring))))
;;
;;(defun redo-record-pre-cmd-state ()
;;  "Record the buffer state prior to command execution."
;;  (setq redo-pre-buffer (current-buffer)
;;	redo-pre-undo buffer-undo-list
;;	redo-pre-mark (copy-marker (mark-marker))
;;	redo-pre-mark-ring mark-ring
;;	redo-redo-entry (list "KEY" (this-command-keys) this-command)))
;;
;;(defun redo-add (entry)
;;  "Add ENTRY to the current list of redo"
;;  (setq redo-key-buffer 
;;	(if redo-key-buffer
;;	    (append redo-key-buffer (list entry))
;;	  (list entry))))
;;
;;(defun redo-add-arg (entry)
;;  "Add ENTRY to the current list of redo arguments"
;;  (setq redo-univarg-buffer
;;	(if redo-univarg-buffer
;;	    (append redo-univarg-buffer (list entry))
;;	  (list entry))))
;;
;;(defun redo-enter-none-state ()
;;  "Place the redo state machine into the none state."
;;  (setq redo-key-buffer nil
;;	redo-univarg-buffer nil
;;	redo-state 'none))
;;
;;(defun redo-enter-argonly-state ()
;;  "Place the redo state machine into the capture state."
;;  (setq redo-univarg-buffer nil
;;	redo-state 'argonly))
;;
;;(defun redo-enter-capturing-state ()
;;  "Place the redo state machine into the capturing state."
;;  (setq redo-key-buffer redo-univarg-buffer
;;	redo-univarg-buffer nil
;;	redo-state 'capturing))
;;
;;(defun redo-enter-held-state ()
;;  "Place the redo state machine into the held state.  In this state, redo can
;;be added to the redo ring.  This will happen automatically if redo-auto-capture
;;is t, otherwise it will only happen on redo."
;;  (setq redo-univarg-buffer nil
;;	redo-state 'held)
;;  (if redo-auto-capture
;;      (progn
;;	(redo-capture-to-ring)
;;	(redo-enter-none-state))))
;;
;;(defun redo-enter-heldarg-state ()
;;  "Place the redo state machine into the heldarg state."
;;  (setq redo-univarg-buffer nil
;;	redo-state 'heldarg))
;;
;;(defun redo-enter-captarg-state ()
;;  "Place the redo state machine into the captarg state."
;;  (setq redo-univarg-buffer nil
;;	redo-state 'captarg))
;;
;;(defvar redo-state-map
;;  '((none . ((arg . redo-enter-argonly-state)
;;	     (edit . redo-enter-capturing-state)
;;	     (stop . nil)
;;	     (redo . nil)
;;	     (quit . nil)))
;;    (argonly . ((arg . nil)
;;		(edit . redo-enter-capturing-state)
;;		(stop . redo-enter-held-state)
;;		(redo . redo-enter-none-state)
;;		(quit . redo-enter-none-state)))
;;    (capturing . ((arg . redo-enter-captarg-state)
;;		  (edit . nil)
;;		  (stop . redo-enter-held-state)
;;		  (redo . redo-enter-none-state)
;;		  (quit . redo-enter-none-state)))
;;    (held . ((arg . redo-enter-heldarg-state)
;;	     (edit . redo-enter-capturing-state)
;;	     (stop . nil)
;;	     (redo . redo-enter-none-state)
;;	     (quit . redo-enter-none-state)))
;;    (heldarg . ((arg . nil)
;;		(edit . redo-enter-capturing-state)
;;		(stop . redo-enter-held-state)
;;		(redo . redo-enter-capturing-state)
;;		(quit . redo-enter-none-state)))
;;    (captarg . ((arg . nil)
;;		(edit . redo-enter-capturing-state)
;;		(stop . redo-enter-held-state)
;;		(redo . redo-enter-capturing-state)
;;		(quit . redo-enter-none-state))))
;;  "Nested alist identifing the function required to change state from a 
;;given start state when a given operation type is performed.")
;;
;;(defun redo-change-state (operation)
;;  "Change the current state of redo capture"
;;  (if operation
;;      (let* ((alist (cdr (assoc redo-state redo-state-map)))
;;	     (op (assoc operation alist))
;;	     (delta-fn (cdr op)))
;;	(if delta-fn
;;	    (progn
;;	      (funcall delta-fn)
;;	      (redo-set-cursor redo-state)))
;;	;; final exp is for debug info only - lose it later
;;	(list delta-fn  redo-state))))
;;
;;(defun redo-stopper-fn-p (fn)
;;  "Predicate identifying whether FN is a stopper function."
;;  (or (member fn redo-stoppers)
;;      (and (not (redo-active-mark-p))
;;	   (member fn redo-conditional-stoppers))))
;;
;;(defun redo-redo-cmd-p (fn)
;;  "Predicate identifying whether FN is a redo function."
;;  (member fn '(redo-redo redo-redo-pop)))
;;
;;(defun redo-redo-arg-p (fn)
;;  "Predicate identifying whether FN is a universal argument function."
;;  (member cmd '(universal-argument
;;		universal-argument-other-key
;;		universal-argument-minus
;;		universal-argument-more
;;		negative-argument
;;		digit-argument)))
;;
;;(defun redo-operation-type (cmd)
;;  "Identify the type of operation that has just been performed."
;;  (if (eq cmd 'keyboard-quit)
;;      'quit
;;    (if (redo-stopper-fn-p cmd)
;;	'stop
;;      (if (redo-redo-arg-p cmd)
;;	  'arg
;;	(if (redo-redo-cmd-p cmd)
;;	    'redo
;;	  (if (and (eq (current-buffer) redo-pre-buffer)
;;		   (or (redo-undo-updated-p cmd) 
;;		       (redo-mark-updated-p)))
;;	      'edit
;;	    nil))))))
;;
;;(defun redo-post-cmd-process (cmd)
;;  "Main function for capturing redo."
;;  (redo-change-state (redo-operation-type cmd))
;;  ;; What we do depends on our state.
;;  (if (eq redo-state 'capturing)
;;      ;; Add current stuff to capture
;;      (redo-add redo-redo-entry)
;;    ;; Not capturing - maybe reading args?
;;    (if (member redo-state '(captarg heldarg argonly))
;;	;; we must be in one of the arg reading states
;;	(redo-add-arg redo-redo-entry))))
;;
;;(defun redo-pre-cmd-fn ()
;;  "Hook function to record buffer state before running a command."
;;  (if redo-allow-redo-recording  ;; prevent recursive recording during
;;                                 ;; playback of redo
;;      (redo-record-pre-cmd-state)))
;;
;;(defun redo-post-cmd-fn ()
;;  "Hook function to possibly record the current command"
;;  (if redo-allow-redo-recording
;;      (redo-post-cmd-process this-command)))
;;
;;(defun redo-enable ()
;;  "Enable redo."
;;  (redo-pre-cmd-fn)
;;  (add-hook 'pre-command-hook 'redo-pre-cmd-fn)
;;  (add-hook 'post-command-hook 'redo-post-cmd-fn))  
;;
;;(defun redo-disable ()
;;  "Disable redo."
;;  (remove-hook 'pre-command-hook 'redo-pre-cmd-fn)
;;  (remove-hook 'post-command-hook 'redo-post-cmd-fn)
;;  (setq redo-state 'none))
;;
;;(defun redo-redo-cmd-list (list count)
;;  "Redo the given LIST of commands COUNT times."
;;  (while (> count 0)
;;    (mapcar 'redo-redo-one-cmd list)
;;    (setq count (- count 1))))
;;
;;(defun redo-redo-one-cmd (cmd)
;;  "This function redoes the single command in CMD."
;;  (cond
;;   ; This is the case when set by our commands commands
;;   ((equal (car cmd) "CMPL") (eval (cdr cmd)))
;;
;;   ;; If the command was minibuffer-complete-and-exit then
;;   ;; the command was stored in the command-history
;;   ((equal (car cmd) "HIST") (eval (car command-history)))
;;
;;   ;; Edit key.
;;   ((equal (car cmd) "KEY") 
;;    (execute-kbd-macro (nth 1 cmd)) nil)
;;   ((equal (car cmd) "RPT")
;;    (redo-redo-cmd-list (nth 2 cmd) (nth 1 cmd)))))
;;
;;(defvar redo-undo-point nil
;;  "Records the position of point prior to redo.  This is to enable point
;;to be restored for redo-pop.")
;;
;;(defvar redo-undo-list nil
;;  "Records the state of undo prior to executing redo."
;;)
;;
;;(defun redo-set-undo ()
;;  "Record information to allow redo to be undone."
;;  (setq redo-undo-point (point)
;;	redo-undo-list buffer-undo-list))
;;
;;(defun redo-close-undo-boundaries ()
;;  "Remove any undo-boundaries created during redo."
;;  (let ((prev buffer-undo-list)
;;	(list (cdr buffer-undo-list)))
;;    (while (and list (not (eq list redo-undo-list)))
;;      (if (car list)
;;	  (setq prev list)
;;	(setcdr prev (cdr list)))
;;      (setq list (cdr list))))
;;  buffer-undo-list)
;;
;;(defun redo-undo ()
;;  "Undo the last redo.  Set undo back to the point is was prior to redo."
;;  (interactive)
;;  (advertised-undo)
;;  (goto-char redo-undo-point)
;;  (setq buffer-undo-list redo-undo-list))
;;
;;(defun redo-capture-to-ring ()
;;  "Move the currently held item into the redo-ring."
;;  (ring-insert redo-ring redo-key-buffer)
;;  (setq redo-ring-idx 0))
;;
;;(defun redo-redo (arg)
;;  "Redo all the stuff from the current element of the redo-ring"
;;  (interactive "p")
;;  (if (member redo-state '(held heldarg))
;;      (redo-capture-to-ring))
;;  (if (ring-empty-p redo-ring)
;;      (error "Nothing-to-redo"))
;;  (undo-boundary)
;;  ;; Perform the redo
;;  (redo-set-undo)
;;  (let ((buf (ring-ref redo-ring redo-ring-idx))
;;	(current-prefix-arg 1)
;;	(prefix-arg 1)
;;	(redo-allow-redo-recording nil)
;;	(count arg))
;;    (while (> count 0)
;;      (mapcar 'redo-redo-one-cmd buf)
;;      (setq count (- count 1))))
;;  (redo-close-undo-boundaries) ;; Eliminate undo boundaries within the redo.
;;  ;; Finally we modify redo-redo-entry to contain the current redo.  This
;;  ;; allows redo to be captured if appropriate.
;;  (setq redo-univarg-buffer nil
;;	redo-redo-entry (list "RPT" arg (ring-ref redo-ring redo-ring-idx))))
;;
;;(defun redo-stop-and-redo (arg)
;;  "Call redo-redo after first acting as a stop function."
;;  (interactive "p")
;;  (redo-change-state 'stop)
;;  (redo-redo arg))
;;
;;(defun redo-redo-pop (arg)
;;  "Pop the previous redo entry from the redo ring"
;;  (interactive "p")
;;  (if (redo-redo-cmd-p real-last-command)
;;      (progn
;;	(redo-undo)
;;	(setq redo-ring-idx (ring-plus1 redo-ring-idx REDO-RING-SIZE))
;;	(redo-redo arg))
;;    ;; Previous command was not redo.
;;   (redo-redo arg)))
;;
;;(provide 'redo)
;;
;;(define-key global-map [f12] 'redo-redo)
;;(define-key global-map [S-f12] 'redo-redo-pop)
;;(define-key global-map [C-f12] 'redo-stop-and-redo)
;;
;;(redo-disable)
;;(redo-reset)
;;(redo-enable)
;;
;;
;;;;(defun redo-debug-test (cmd key)
;;;;  "Test function for developing redo"
;;;;  (redo-this-command cmd key)
;;;;  (redo-pre-cmd-fn)
;;;;  (insert "a")  ;; simple edits to cause undo to be recorded.
;;;;  (backward-delete-char-untabify 1)
;;;;  (redo-post-cmd-fn))
;;  
;;
;;
;;(defun redo-debug-test-blah ()
;;  "Function to test redo operations.  Run these expressions interactively"
;;  (progn
;;    ;; 1 - simple capture of edits
;;    (redo-reset)
;;    (redo-debug-test 'self-insert-command "a")
;;    (redo-debug-test 'self-insert-command "a")
;;
;;    ;; XXXX Why does this test not change the cursor?
;;    (redo-debug-test 'next-line "<up>")
;;
;;    ;; 2 - arg then stop
;;    (redo-capture (redo-this-command 'universal-argument "C-u"))
;;    (redo-capture (redo-this-command 'next-line "<up>"))
;;    ;; 3 - edit, arg then stop
;;    (redo-capture (redo-this-command 'self-insert-command "a"))
;;    (redo-capture (redo-this-command 'universal-argument "C-u"))
;;    (redo-capture (redo-this-command 'next-line "<up>"))
;;    ;; 4 - edit, arg, edit, arg then stop
;;    (redo-capture (redo-this-command 'self-insert-command "a"))
;;    (redo-capture (redo-this-command 'universal-argument "C-u"))
;;    (redo-capture (redo-this-command 'self-insert-command "a"))
;;    (redo-capture (redo-this-command 'universal-argument "C-u"))
;;    (redo-capture (redo-this-command 'next-line "<up>"))
;;    
;;    ;; important state variables
;;    redo-state
;;    redo-key-buffer
;;    redo-ring
;;)    
;;  (redo-change-state 'arg)
;;  (redo-change-state 'edit)
;;  (redo-change-state 'stop)
;;  (redo-change-state 'redo)
;;  (redo-change-state 'quit)
;;  (redo-reset)
;;)
;;
;;;;
;;;;
;;;;
;;;;
;;;;
;;;;
;;;;
;;;;;;;; debug stuff
;;;;(progn
;;;;  (redo-disable)
;;;;  (redo-reset)
;;;;  (redo-enable)
;;;;) 
;;;;
;;;;
;;;;redo-state 
;;;;pre-command-hook
;;;;post-command-hook
;;;;(redo-post-cmd-fn)
;;;;
;;;;
;;;;
;;;;;;
;;;;;;(redo-post-cmd-fn)
;;;;;;redo-key-buffer    xx
;;;;;;redo-ring          xxxxx  x
;;;;;;redo-pre-buffer
;;;; 
;;;;;;;;
;;;;;;;;(defvar previous-command nil
;;;;;;;;  "This function is supplied to allow other functions to set 
;;;;;;;;the previous command. The previous command takes the form:-
;;;;;;;;(\"CMPL\" <func> <args1> <arg2> ...) -- Complex command
;;;;;;;;(\"KEY\" <command-keys> <func> )     -- Keyboard command
;;;;;;;;(\"HIST\" )                          -- History command
;;;;;;;;In the COMPLEX command the function is executed with the arguments. With
;;;;;;;;the KEY command the function is ignored and the function replicated by
;;;;;;;;`execute-kbd-macro' using the command-keys. The command keys are from the 
;;;;;;;;function `this-command-keys'")
;;;;;;;;
;;;;;;;;
;;;;;;;;
;;;;;;;;(defun prev-command-hook ()
;;;;;;;;  "This is the hook in to the previous command so that normal editing 
;;;;;;;;keys can be put into the redo last command"
;;;;;;;;  (interactive)
;;;;;;;;  (cond
;;;;;;;;   ;; Don't save the keyboard press if these were the last commands
;;;;;;;;   ((member this-command (list 'rerun-prev-command
;;;;;;;;			       'keyboard-quit
;;;;;;;;			       'exit-minibuffer)) t)
;;;;;;;;    
;;;;;;;;   ;; If the command was `minibuffer-complete-and-exit' then the
;;;;;;;;   ;; last command must be stored in the command history. So tell 
;;;;;;;;   ;; the rerun command to use this
;;;;;;;;   ((equal this-command 'minibuffer-complete-and-exit)
;;;;;;;;    (set-prev-cmd (list "HIST"))
;;;;;;;;   )
;;;;;;;;
;;;;;;;;   ;; Otherwise it must be just a normal key
;;;;;;;;   (t
;;;;;;;;    (set-prev-cmd (list "KEY" (this-command-keys) this-command))
;;;;;;;;   )
;;;;;;;;  )
;;;;;;;;)
;;;;;;;;
;;;;;;;;
;;;;;;;;(defun rerun-prev-command ()
;;;;;;;;  "This function reruns the last command. The command is defined in 
;;;;;;;;the variable previous-command"
;;;;;;;;  (interactive)
;;;;;;;;  (cond
;;;;;;;;
;;;;;;;;   ; This is the case when set by our commands commands
;;;;;;;;   ((equal (car previous-command) "CMPL") (eval (cdr previous-command)))
;;;;;;;;
;;;;;;;;   ;; If the command was minibuffer-complete-and-exit then
;;;;;;;;   ;; the command was stored in the command-history
;;;;;;;;   ((equal (car previous-command) "HIST") (eval (car command-history)))
;;;;;;;;
;;;;;;;;   ;; Otherwise it must have been a keyboard macro
;;;;;;;;   ((equal (car previous-command) "KEY") 
;;;;;;;;    (execute-kbd-macro (nth 1 previous-command))
;;;;;;;;   )
;;;;;;;;  )
;;;;;;;;)
;;;;;;;;
;;;;;;;;(defun set-prev-cmd ( lst )
;;;;;;;;  "This function is supplied to allow other functions to set 
;;;;;;;;the previous command. The previous command takes the form:-
;;;;;;;;(\"CMPL\" <func> <args1> <arg2> ...) -- Complex command
;;;;;;;;(\"KEY\" <command-keys> <func> )     -- Keyboard command
;;;;;;;;(\"HIST\" )                          -- History command
;;;;;;;;In the COMPLEX command the function is executed with the arguments. With
;;;;;;;;the KEY command the function is ignored and the function replicated by
;;;;;;;;`execute-kbd-macro' using the command-keys. The command keys are from the 
;;;;;;;;function `this-command-keys'"
;;;;;;;; (interactive)
;;;;;;;; (setq previous-command lst)
;;;;;;;; (print previous-command)
;;;;;;;;)
;;;;;;;;
;;;;;;;;(defun griffith-advertised-undo ()
;;;;;;;;  "This is a roof over the undo command `advertised-undo' that 
;;;;;;;;allows access so that the command can be trapped and fed into the 
;;;;;;;;previous command"
;;;;;;;;  (interactive)
;;;;;;;;  (advertised-undo)
;;;;;;;;  (set-prev-cmd (list "CMPL" 'griffith-advertised-undo))
;;;;;;;;)
;;;;;;;;
;;;;;;;;(defun griffith-call-last-kbd-macro ( &optional PREFIX )
;;;;;;;;  "This is a roof over the `call-last-kbd-macro' command to 
;;;;;;;;allows access so that the command can be trapped and fed into the 
;;;;;;;;previous command"
;;;;;;;;  (interactive "p")
;;;;;;;;  (call-last-kbd-macro PREFIX)
;;;;;;;;  (set-prev-cmd (list "CMPL" 'griffith-call-last-kbd-macro PREFIX))
;;;;;;;;)
;;;;;;;;
;;;;;;;;
;;;;;;;;
;;;;;;;;(define-key global-map [f12] 'rerun-prev-command)
;;;;;;;;
;;;;;;
;;;;
