;;; ringset.el --- Provide ring operations that only allow unique
;;;                entries in the ring.  Adding a new entry that 
;;;                matches an existing entry results in the old entry
;;;                being removed from the ring and placed at its head.
;;;
;;; This extends the standard ring functionality, by adding an optional
;;; argument to make-ring which describes the equality test function
;;; to be used.
;;;
;;; The standard ring structure:
;;;    (hd-index length . vector)
;;; is slightly modified to
;;;    (hd-index (length . equality-fn) . vector)
;;; The usual equality-fn for creating a ringset would be 'equal.
;;;

(require 'ring)


(defadvise ring-p (around ringset (x))
  "Return t if X is a ring; nil otherwise."
  ;; Note that x may be an original (pre-advise) form of ring, or
  ;; it may be a new form.
  (and (consp x) 
       (integerp (car x))
       (consp (cdr x)) 
       (or (integerp (car (cdr x)))
	   (and (consp (car (cdr x)))
		(integerp (car (car (cdr x))))))
       (vectorp (cdr (cdr x)))))


(defadvise make-ring (around ringset (size &optional equality-fn))
  "Make a ring that can contain SIZE elements, with duplicates detected
by EQUALITY-FN."
  (cons 0 (cons (0 equality-fn) (make-vector size nil))))



;;
;;(defun ring-insert-at-beginning (ring item  &optional despace)
;;  "Add to RING the item ITEM.  Add it at the front (the early end)."
;;  (if despace
;;      (setq item (ring-despace-input item)))
;;  (ring-remove-duplicate ring item)
;;  (let* ((vec (cdr (cdr ring))) 
;;	 (veclen (length vec))
;;	 (hd (car ring))
;;	 (ln (car (cdr ring))))
;;    (setq ln (min veclen (1+ ln))
;;	  hd (ring-minus1 hd veclen))
;;    (aset vec hd item)
;;    (setcar ring hd)
;;    (setcar (cdr ring) ln)))
;;
;;(defun ring-plus1 (index veclen)
;;  "INDEX+1, with wraparound"
;;  (let ((new-index (+ index 1)))
;;    (if (= new-index veclen) 0 new-index)))
;;
;;(defun ring-minus1 (index veclen)
;;  "INDEX-1, with wraparound"
;;  (- (if (= 0 index) veclen index) 1))
;;
;;(defun ring-length (ring)
;;  "Number of elements in the ring."
;;  (car (cdr ring)))
;;
;;(defun ring-empty-p (ring)
;;  (= 0 (car (cdr ring))))
;;
;;(defun ring-index (index head ringlen veclen)
;;  (setq index (mod index ringlen))
;;  (mod (1- (+ head (- ringlen index))) veclen))
;;
;;(defun ring-insert (ring item  &optional despace)
;;  "Insert onto ring RING the item ITEM, as the newest (last) item.
;;If the ring is full, dump the oldest item to make room."       
;;  (if despace
;;      (setq item (ring-despace-input item)))
;;  (ring-remove-duplicate ring item)
;;  (let* ((vec (cdr (cdr ring))) 
;;	 (veclen (length vec))
;;	 (hd (car ring))
;;	 (ln (car (cdr ring))))
;;    (prog1
;;	(aset vec (mod (+ hd ln) veclen) item)
;;      (if (= ln veclen)
;;	  (setcar ring (ring-plus1 hd veclen))
;;	(setcar (cdr ring) (1+ ln))))))
;;
;;(defun ring-remove (ring &optional index)
;;  "Remove an item from the RING.  Return the removed item.
;;If optional INDEX is nil, remove the oldest item.  If it's
;;numeric, remove the element indexed."
;;  (if (ring-empty-p ring)
;;      (error "Ring empty")
;;    (let* ((hd (car ring))
;;	  (ln (car (cdr ring)))
;;	  (vec (cdr (cdr ring)))
;;	  (veclen (length vec))
;;	  (tl (mod (1- (+ hd ln)) veclen))
;;	  oldelt)
;;      (if (null index)
;;	  (setq index (1- ln)))
;;      (setq index (ring-index index hd ln veclen))
;;      (setq oldelt (aref vec index))
;;      (while (/= index tl)
;;	(aset vec index (aref vec (ring-plus1 index veclen)))
;;	(setq index (ring-plus1 index veclen)))
;;      (aset vec tl nil)
;;      (setcar (cdr ring) (1- ln))
;;      oldelt)))
;;
;;(defun ring-ref (ring index)
;;  "Returns RING's INDEX element.
;;INDEX need not be <= the ring length, the appropriate modulo operation
;;will be performed.  Element 0 is the most recently inserted; higher indices
;;correspond to older elements until they wrap."
;;  (if (ring-empty-p ring)
;;      (error "indexed empty ring")
;;    (let* ((hd (car ring))  (ln (car (cdr ring)))  (vec (cdr (cdr ring))))
;;      (aref vec (ring-index index hd ln (length vec))))))
;;
;;(provide 'ring)
;;
;;;; Marc's hacks (also the calls added above to ring-remove-duplicate)
;;
;;(defun ring-item-in-ring (ring item)
;;  "Return index into RING for ITEM.  If item does not exist, return nil."
;;  (let ((found)
;;	(index (car ring))
;;	(count 0)
;;	(ln (car (cdr ring)))
;;	(vec (cdr (cdr ring))))
;;    (while (and (not found) (< count ln))
;;      (if (string-equal item (aref vec index))
;;	  (setq found (1- count))
;;	(setq index (ring-minus1 index ln))
;;	(setq count (1+ count))))
;;    found))
;;
;;(defun ring-remove-duplicate (ring item)
;;  "Remove from RING the first occurrence of ITEM."
;;  (let ((index (ring-item-in-ring ring item)))
;;    (if index
;;	(ring-remove ring index))))
;;
;;
;;(defvar ring-despace-regexp 
;;  (concat
;;   "\\([ 	]*\\(\n\\)[ 	]*\\|\\([ 	]+\\)\\)\\|"
;;   "\\(\\([\"\'\`]\\)\\|\\([^ 	
;;\'\"\`]+\\)\\)")
;;  "Regular expresssion identifying:
;;\2 A linefeed (possibly embedded in spaces);
;;\3 A sequence of spaces and/or tabs;
;;\1 \2 or \3;
;;\5 A quote character;
;;\6 A word (a string not containing spaces or quotes);
;;\4 \5 or \6;
;;\0 \1 or \4.")
;;
;;(defun ring-despace-input (input)
;;  "Function to standardise format of input to history rings.  This changes
;;whitespace that is not quoted into single spaces or newlines only.
;;It returns a string that is the modified input string."
;;  (let ((modified-input "")
;;	(quote-char)
;;	(limit (length input))
;;	(start)
;;	(end (string-match "[^ 	]" input))
;;	(word)
;;	(done))
;;    (while (not done)
;;      (if (setq start (string-match ring-despace-regexp input end))
;;	  (progn
;;	    (setq end (match-end 0))
;;	    (cond ((match-beginning 2)  ;; Line feed
;;		   (if (< end limit)
;;		       (setq word "\n")
;;		     (setq word "")))
;;		  ((match-beginning 3)  ;; Spaces
;;		   (if (< end limit)
;;		       (setq word " ")
;;		     (setq word "")))
;;		  ((match-beginning 5)  ;; One of the quote characters
;;		   (setq quote-char (substring input start end))
;;		   (if (string-match quote-char input (+ start 1))
;;		       (progn
;;			 (setq end (match-end 0))
;;			 (setq word (substring input start end)))
;;		     (setq word (substring input start))
;;		     (setq done t)))
;;		  ((match-beginning 6)  ;; A word
;;		   (setq word (substring input start end)))
;;		  (t
;;		   (setq done t)))
;;	    (setq modified-input (concat modified-input word)))
;;	(setq done t)))
;;    modified-input))
;;