;;; ringset.el --- Provide ring operations that only allow unique
;;;                entries in the ring.  Adding a new entry that 
;;;                matches an existing entry results in the old entry
;;;                being removed from the ring and placed at its head.
;;;
;;; This extends the standard ring functionality by adding an optional
;;; argument to make-ring which describes the equality test predicate
;;; to be used to check for duplicates.  Without the extra argument
;;; to make-ring, rings worlk exactly as they used to.  With the extra 
;;; argumnent, rings become ringsets in which only a single copy of an 
;;; element will ever be stored in the ring.
;;;
;;; The standard ring structure:
;;;    (hd-index length . vector)
;;; is slightly modified to
;;;    (hd-index (length . equality-fn) . vector)
;;; The usual equality-fn for creating a ringset would be 'equal.
;;;

(require 'ring)
(require 'advice)

(defun ring-equality-fn (x)
  "Provide the equality-fn for x, if one exists"
  (let ((fn-cell (car (cdr x))))
    (and (consp fn-cell)
	 (cdr fn-cell))))

(defun ring-matching-elem (ring elem)
  "Return the index from RING of an element matching ELEM"
  (let ((fn (ring-equality-fn ring))
	(i))
    (and fn
	 (progn
	   (setq i (1- (ring-length ring)))
	   (while (and (>= i 0)
		       (not (funcall fn elem (ring-ref ring i))))
	     (setq i (1- i)))
	   (and (>= i 0) i)))))

(defun ring-remove-dupe (ring elem)
  (let ((match (ring-matching-elem ring elem)))
    (if match
	(ring-remove ring match))))

(defadvice ring-p (around ringset (x) activate)
  "Note that ring may be an original (pre-advice) form of ring, or it may be a 
ringset form." 
  (setq ad-return-value
	(and (consp x) 
	     (integerp (car x))
	     (consp (cdr x)) 
	     (or (integerp (car (cdr x)))
		 (and (consp (car (cdr x)))
		      (integerp (car (car (cdr x))))))
	     (vectorp (cdr (cdr x))))))

(defadvice make-ring (around ringset (size &optional equality-fn) activate)
  "Add optional parameter EQUALITY-FN which is a predicate for comparing
elements so that duplicates need not be stored."
  (setq ad-return-value 
	(cons 0 (cons (cons 0 equality-fn) (make-vector size nil)))))

(defadvice ring-insert-at-beginning (around ringset (ring item) activate)
  "Note that ring may be an original (pre-advice) form of ring, or it may be a 
ringset form." 
  (ring-remove-dupe ring item)
  (setq ad-return-value
	(let* ((hd (car ring))
	       (lncell (cdr ring))
	       (ln (car lncell))
	       (vec (cdr lncell))
	       (veclen (length vec))))
	(if (consp ln) ;; Handle new form of ring length storage
	    (setq lncell ln
		  ln (car ln)))
	(setq ln (min veclen (1+ ln))
	      hd (ring-minus1 hd veclen))
	(aset vec hd item)
	(setcar ring hd)
	(setcar lncell ln)))

(defadvice ring-length (after ringset activate)
  "Note that ring may be an original (pre-advice) form of ring, or it may be a 
ringset form." 
  (if (consp ad-return-value)
      (setq ad-return-value (car ad-return-value))))

(defadvice ring-empty-p (around ringset (ring) activate)
  "Note that ring may be an original (pre-advice) form of ring, or it may be a 
ringset form." 
  (setq ad-return-value
	(let* ((ln (car (cdr ring))))
	  (if (consp ln)
	      (zerop (car ln))
	    (zerop ln)))))

(defadvice ring-copy (around ringset (ring) activate) 
  "Note that ring may be an original (pre-advice) form of ring, or it may be a 
ringset form." 
  (let* ((vec (cdr (cdr ring)))
	 (hd  (car ring))
	 (ln  (car (cdr ring))))
    (if (consp ln)
	(setq ln (cons (car ln) (cdr ln))))
    (setq ad-return-value
	  (cons hd (cons ln (copy-sequence vec))))))

(defadvice ring-insert (around ringset (ring item) activate)
  "Note that ring may be an original (pre-advice) form of ring, or it may be a 
ringset form." 
  (ring-remove-dupe ring item)
  (setq ad-return-value
	(let* ((hd (car ring))
	       (lncell (cdr ring))
	       (ln (car lncell))
	       (vec (cdr lncell))
	       (veclen (length vec)))
	  (if (consp ln) ;; Handle new form of ring length storage
	      (setq lncell ln
		    ln (car ln)))
	  (prog1
	      (aset vec (mod (+ hd ln) veclen) item)
	    (if (= ln veclen)
		(setcar ring (ring-plus1 hd veclen))
	      (setcar lncell (1+ ln)))))))

(defadvice ring-remove (around ringset (ring &optional index) activate)
  "Note that ring may be an original (pre-advice) form of ring, or it may be a 
ringset form." 
  (if (ring-empty-p ring)
      (error "Ring empty")
    (setq ad-return-value
	  (let* ((hd (car ring))
		 (lncell (cdr ring))
		 (ln (car lncell))
		 (vec (cdr lncell))
		 (veclen (length vec))
		 tl
		 oldelt)
	    (if (consp ln) ;; Handle new form of ring length storage
		(setq lncell ln
		      ln (car ln)))
	    (setq tl (mod (1- (+ hd ln)) veclen))
	    (if (null index)
		(setq index (1- ln)))
	    (setq index (ring-index index hd ln veclen))
	    (setq oldelt (aref vec index))
	    (while (/= index tl)
	      (aset vec index (aref vec (ring-plus1 index veclen)))
	      (setq index (ring-plus1 index veclen)))
	    (aset vec tl nil)
	    (setcar lncell (1- ln))
	    oldelt))))

(defadvice ring-ref (around ringset (ring index) activate)
  "Note that ring may be an original (pre-advice) form of ring, or it may be a 
ringset form." 
  (if (ring-empty-p ring)
      (error "Accessing an empty ring")
    (let* ((hd (car ring))  
	   (lncell (cdr ring))
	   (ln (car lncell))  
	   (vec (cdr lncell)))
      (if (consp ln) ;; Handle new form of ring length storage
	  (setq ln (car ln)))
      (setq ad-return-value
	    (aref vec (ring-index index hd ln (length vec)))))))


(provide 'ringset)

