;; rollback.el
;; provide the means to record a savepoint, and the means to rollback to
;; that savepoint.

(defun first-undo-element (&optional undo-list)
  "Return the first non-nil entry in optional UNDO-LIST or `buffer-undo-list'."
  (let ((undo (or undo-list buffer-undo-list)))
    (if (consp undo)
	(progn
	  (while (not (car undo))
	    (setq undo (cdr undo)))
	  (car undo)))))

(defun savepoint ()
  "Return a savepoint to which we can later rollback"
  (cons (point) (first-undo-element)))

(defun fetch-undo-group (undo-list)
  "Scan UNDO-LIST to find the boundaries of the current undo group.
Return a list containing the first and last cons cells of the undo
group."
  ;; Step over boundaries
  (while (and undo-list (null (car undo-list)))
    (setq undo-list (cdr undo-list)))
  (let ((start undo-list))
    (while (cadr undo-list)
      (setq undo-list (cdr undo-list)))
    (and undo-list
	 (cons start undo-list))))
      

(defun primitive-coalesce (undo-list elem)
   "If, within UNDO-LIST, ELEM is found, coalesce undo into a single
undo group (ie remove intervening nil entries).  The result of each
level of recursion will be:
t if elem is found;
nil if elem was not found anywhere;
otherwise  a list of the form (<start of undo group> . <end of undo group>)."
   (let ((group (fetch-undo-group undo-list))
	 (next-group))
     (if group
	 (if (eq (caar group) elem)
	     t
	   (if (setq next-group (primitive-coalesce (cddr group) elem))
	       (progn
		 (if (consp next-group)
		     ;; Remove the boundary after group
		     (setcdr (cdr group) (car next-group)))
		 group))))))

(defun coalesce-to-savepoint (savepoint)
  "Coalesce undo up to SAVEPOINT into a single undo-group.  After using this
You may want to add another undo boundary"
  (primitive-coalesce buffer-undo-list (cdr savepoint)))

(defun undo-to-savepoint (savepoint)
  "Perform undo until we have returned the buffer state to that of
SAVEPOINT.  Unlike rollback, this generates more undo."
  (if (not (coalesce-to-savepoint buffer-undo-list ))
      (error "Insufficient undo"))
  (undo)
  (goto-char (car savepoint)))

(defun undo-contains-savepoint-p (savepoint)
  "Predicate to determine whether undo still contains SAVEPOINT."
  (member (cdr savepoint) buffer-undop-list))

(defun rollback (savepoint)
  "Rollback to SAVEPOINT.  Apply undo until savepoint is reached and then
restore point.  This does not generate further redo.  To generate redo
while restoring to a savepoint, use undo-to-savepoint."
  (let ((undo-elem (cdr savepoint))
	(point-elem (car savepoint)))
    ;; First we check that savepoint can be found in the undo-list
    (if (not (member undo-elem buffer-undo-list))
	(error "Insufficient undo for rollback"))
    (while (not (eq undo-elem (first-undo-element)))
      (setq buffer-undo-list (primitive-undo 1 buffer-undo-list)))
    (goto-char point-elem)))

(provide 'rollback)

;;; Useful function for debugging
(defun to-nth (list n)
  "Descend LIST to Nth item, returning all items to that point"
  (if (= n 1)
      (cons (car list) nil)
    (cons (car list) (to-nth (cdr list) (1- n)))))

;;; Useful function for debugging
(defun undo-to-string(&optional undo-list)
  "Return a string showing the top elements of undo"
  (prin1-to-string (to-nth (or undo-list buffer-undo-list) 8)))

