;;; sqlplus.el --- specialized comint.el for running sqlplus.
;;; Copyright (C) 1993 

;; Author: Marc Munro <marc@praxis.co.uk> 
;; Keywords: processes

;;; This file is part of GNU Emacs.

;;; GNU Emacs is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2, or (at your option)
;;; any later version.

;;; GNU Emacs is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.

;;; You should have received a copy of the GNU General Public License
;;; along with GNU Emacs; see the file COPYING.  If not, write to
;;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.


;;=============================================================================
;; Some suggestions for your .emacs file.
;;(autoload 'sqlplus "elec-sqlplus"
;;  "Load electric sqlplus mode when sqlplus is invoked." t)
;;
;;(defvar sqlplus-shell-name
;;  "/u/username/bin/startsql"
;;  "Name of command to run to start SQL*PLUS.  This command may interactively
;;set up the ORACLE_SID if necessary.")
;;

;;; Brief Command Documentation:
;;;============================================================================
;;; Comint Mode Commands: (common to shell and all comint-derived modes)
;;;
;;; m-p	    comint-previous-input    	    Cycle backwards in input history
;;; m-n	    comint-next-input  	    	    Cycle forwards
;;; m-c-r   comint-previous-input-matching  Search backwards in input history
;;; return  comint-send-input
;;; c-a     comint-bol                      Beginning of line; skip prompt.
;;; c-d	    comint-delchar-or-maybe-eof	    Delete char unless at end of buff.
;;; c-c c-u comint-kill-input	    	    ^u
;;; c-c c-w backward-kill-word    	    ^w
;;; c-c c-c comint-interrupt-subjob 	    ^c
;;; c-c c-z comint-stop-subjob	    	    ^z
;;; c-c c-\ comint-quit-subjob	    	    ^\
;;; c-c c-o comint-kill-output		    Delete last batch of process output
;;; c-c c-r comint-show-output		    Show last batch of process output
;;;         send-invisible                  Read line w/o echo & send to proc
;;;         comint-continue-subjob	    Useful if you accidentally suspend
;;;					        top-level job.
;;; comint-mode-hook is the comint mode hook.


;;; Customization and Buffer Variables
;;; ===========================================================================
;;; 
;;; sqlplus-shell-name

;;; Code:

(require 'comint)


;;;###autoload
(defvar sqlplus-prompt-pattern "^\\(SQL> \\)+"
  "Regexp to match prompts in the inferior shell.
allows SQL> SQL> ...
This variable is used to initialise `comint-prompt-regexp' in the 
shell buffer.")

(defvar sqlplus-mode-map '())
(cond ((not sqlplus-mode-map)
       (setq sqlplus-mode-map (full-copy-sparse-keymap comint-mode-map))
       (define-key sqlplus-mode-map "\C-m" 'sqlplus-handle-return)))

(defvar sqlplus-mode-hook '()
  "*Hook for customising sqlplus mode.")

(defvar sqlplus-newline '(lambda () (insert "\n"))
  "Function for doing simple newlines in sqplus buffer.")

(defvar sqlplus-shell-name
  nil
  "Name of command to run to start SQL*PLUS.  This command may interactively
set up the ORACLE_SID if necessary.")

(defvar sql-mode-syntax-table nil
  "Set up syntax table.")

(if sql-mode-syntax-table
    ()
  (setq sql-mode-syntax-table (make-syntax-table))
  (modify-syntax-entry ?\\ "\\" sql-mode-syntax-table)
  (modify-syntax-entry ?+ "." sql-mode-syntax-table)
  (modify-syntax-entry ?- "." sql-mode-syntax-table)
  (modify-syntax-entry ?= "." sql-mode-syntax-table)
  (modify-syntax-entry ?% "." sql-mode-syntax-table)
  (modify-syntax-entry ?< "." sql-mode-syntax-table)
  (modify-syntax-entry ?> "." sql-mode-syntax-table)
  (modify-syntax-entry ?& "." sql-mode-syntax-table)
  (modify-syntax-entry ?| "." sql-mode-syntax-table)
  (modify-syntax-entry ?\' "\"" sql-mode-syntax-table))


;;; Basic Procedures
;;; ===========================================================================
;;;

(defun sqlplus-mode ()
  "Major mode for interacting with Oracle SQL*Plus.
Return at end of buffer sends line as input.
Return not at end copies SQL statement to end and executes it.  
\\{sqlplus-mode-map}
This mode is normally invoked by 'M-x sqlplus' (not 'M-x sqlplus-mode').

If you accidentally suspend your process, use \\[comint-continue-subjob]
to continue it.

\\{sqlplus-mode-map}
Customization: Entry to this mode runs the hooks on `comint-mode-hook' and
`sqlplus-mode-hook' (in that order)."
  (interactive)
  (comint-mode)
  (setq comint-prompt-regexp sqlplus-prompt-pattern)
  (setq major-mode 'sqlplus-mode)
  (setq mode-name "Sqlplus")
  (use-local-map sqlplus-mode-map)
  (set-syntax-table sql-mode-syntax-table)
  (run-hooks 'sqlplus-mode-hook))


;;;###autoload
(defun sqlplus ()
  "Start up an interactive SQL*Plus session in a new buffer.
If an active SQL*Plus process already exists, will switch to that
buffer.

\(Type \\[describe-mode] in the shell buffer for a list of commands.)"
  (interactive)
  (cond ((not (comint-check-proc "*sqlplus*"))
	 (let* ((prog (or sqlplus-shell-name
			  "sqlplus"))
		(name (file-name-nondirectory prog))
		(startfile (concat "~/.emacs_" name))
		(xargs-name (intern-soft (concat "explicit-" name "-args"))))
	   (set-buffer (apply 'make-comint "sqlplus" prog
			      (if (file-exists-p startfile) startfile)
			      (if (and xargs-name (boundp xargs-name))
				  (symbol-value xargs-name))))
				
           (set-process-filter (get-buffer-process  "*sqlplus*")
			       'sqlplus-filter)
	   (sqlplus-mode))))
  (switch-to-buffer "*sqlplus*"))

;;; Other functions and variables.  These are used to control whether 
;;; statements are sent to the interpreter or not.

(defvar sqlplus-command-words
  (concat
   "\\(insert\\|update\\|select\\|delete\\|commit\\|rollback"
   "\\|create\\|alter\\|drop\\|explain\\|grant\\\)")
  "The sql reserved words that introduce a command.")

(defvar sqlplus-terminators
  (concat "\\(;\\|^[ \t]*/[ \t]*$\\|" sqlplus-prompt-pattern "\\)")
  "Regular expressions that start, complete or terminate a statement.")

(defun sqlplus-in-sql-p ()
  "Predicate to determine whether we are currently writing a sql statement."
  (let ((pos (point)))
    (save-excursion
      ;; Find start of statement.
      (goto-char (process-mark (get-buffer-process (buffer-name))))
      (if (not (re-search-forward "[^ \t\n]" nil t))
	  nil ;; if no word found, we're not in sql.
	(forward-char -1)
	(and (looking-at sqlplus-command-words)
	     (<= (point) pos))))))

(defun sqlplus-in-plsql-p ()
  "Function to determine whether we are currently writing a pl/sql block.
If so it returns the position of the start of the block, else nil."
  (let ((pos (point)))
    (save-excursion
      ;; Find start of statement.
      (goto-char (process-mark (get-buffer-process (buffer-name))))
      (if (not (re-search-forward "[^ \t\n]" nil t))
	  nil ;; if no word found, we're not in sql.
	(forward-char -1)
	(and (looking-at plsql-block-start-words)
	     (<= (point) pos))))))

(defun sqlplus-statement-complete ()
  "Predictate to determine whether the statement we are in is complete and
may therefore be sent to the sqlplus interpreter.  It is simplistic and
just looks for a sql-terminator to finish the current statement.  It returns
nil if the statement is not complete, otherwise it returns an alist of the 
start and end buffer positions."
  (if (sqlplus-in-sql-p)
      (save-excursion
	(if (not (re-search-forward sqlplus-terminators nil t))
	    nil  ;; Not in a complete statement
	  (cons (process-mark (get-buffer-process (buffer-name))) (point))))))

(defun sqlplus-handle-return ()
  "Do not pass incomplete bits of sql to the interpreter, but otherwise 
handle newline as comint normally does."
  (interactive)
  (let ((sql-start (process-mark (get-buffer-process (current-buffer))))
	(sql-region))
    (if (< (point) sql-start)
	(progn
	  (setq sql-region (sqlplus-statement-complete))
	  (if sql-region
	      (progn
		(goto-char (point-max))
		(insert (buffer-substring (car sql-region) (cdr sql-region)))))
	  (comint-send-input))
      (if (not (sqlplus-in-sql-p))
	  (comint-send-input)
	(save-excursion
	  (goto-char sql-start)
	  (forward-char 1)
	  (setq sql-region (sqlplus-statement-complete)))
	(if sql-region
	    (progn
	      (goto-char (cdr sql-region))
	      (comint-send-input))
	  (funcall sqlplus-newline))))))


(defun sqlplus-filter (proc string)
  "Filter for SQL*Plus to allow passwords to be read without echo."
  (comint-filter proc string)
  (if (string-match "Enter password" string)
      (let ((echo-keystrokes 0))
	(comint-simple-send proc (read-password)))))

(provide 'sqlplus)
