;; TODO:
;; - check mark handling for first mark of session - no need to check the
;; mark ring in this case?
;; - allow minibuffer stuff to be captured
;; - allow universal argument to be captured at the start of redo
;; - allow redo itself to be captured?
;; - allow minibuffer operations into redo
;; - add stop functions for narrow-widen
;; - ensure mouse movements are not collected
;; - ensure mouse operations are stoppers
;; - optimise redo-list stuff

(defvar redo-stoppers
  (list 'minibuffer-complete-and-exit 'keyboard-quit 'save-buffer)
  "List of functions that stop the accumulation of redo.  These are known
as stopper or stop functions.")

(defvar redo-conditional-stoppers
  (list 'next-line 'previous-line 'end-of-buffer 'beginning-of-buffer
	'scroll-down 'scroll-up 'scroll-up-in-place 'scroll-down-in-place)
  "List of functions that may stop the accumulation of redo.  Whether
they do so will depend on whether we have marked something.")

(defun redo-active-mark-p ()
  "t if we are in transient-mark-mode and the mark is active"
  (and transient-mark-mode mark-active))

(defun redo-stopper-fn-p (fn)
  "Predicate identifying whether FN is a stopper function."
  (or (member fn redo-stoppers)
      (and (not (redo-active-mark-p))
	   (member fn redo-conditional-stoppers))))
  
(defconst redo-RING-SIZE 10
  "Size of redo-ring")

(defvar redo-ring (make-ring redo-RING-SIZE)
  "Ring of the last N redo entries")

(defvar redo-ring-idx 0
  "Index into redo-ring")

(defvar redo-buffer nil
  "Variable into which our current redo is accumulated.")

(defvar redo-recording nil
  "Whether we are currently recording redo.")
  
(defvar redo-pre-mark-ring nil
  "The mark-ring at the start of the command.")
	
(defvar redo-pre-buffer nil
  "The buffer we were in at the start of the command")

(defvar redo-pre-undo nil
  "The state of undo at the start of the command")

(defun redo-complete (add)
  "Complete the current redo entry and maybe add it to the head of the ring."
  (if redo-buffer
      (progn
	(if add
	    (progn 
	      (ring-insert redo-ring redo-buffer)
	      (setq redo-ring-idx 0
		    redo-buffer nil)))
	(setq redo-recording nil
	      cursor-type t))))

(defun redo-add (cmd)
  "Add CMD to the current list of redo"
  (or redo-recording
      (setq redo-recording t
	    redo-buffer nil
	    cursor-type 'bar))
  (setq redo-buffer (if redo-buffer
			 (append redo-buffer (list cmd))
		       (list cmd))))

(defun redo-equivalent-undo-p (undo1 undo2)
  "Compare undo1 and undo2, returning t if they are equivalent."
  ;; TODO: figure out a better way to do this.
  (while (and undo1 (not (car undo1)))
    ;; step past any undo boundaries
    (setq undo1 (cdr undo1)))
  (while (and undo2 (not (car undo2)))
    ;; step past any undo boundaries
    (setq undo2 (cdr undo2)))
  (eq (cdr undo1) (cdr undo2)))

(defun redo-undo-updated-p ()
  "Did this command begin an editing sequence?  It is if it has caused 
undo to be modified and is not itself an undo command."
  ;; Comparing cdr of the undo lists appears to be the safe way to
  ;; tell whether or not we modified undo
  ;; TODO rewrite to make it cleaner.
  (let ((is-undo (member this-command (list 'undo 'advertised-undo))))
    (and (not is-undo) 
	 (not (redo-equivalent-undo-p redo-pre-undo buffer-undo-list)))))

(defun redo-mark-updated-p ()
  "Did this command modify the mark-ring?"
  (not (equal redo-pre-mark-ring mark-ring)))

(defun redo-addable-p (fn)
  "Figure out whether FN is addable to the current redo."
  (and fn
       (not (redo-stopper-fn-p fn))
       (eq (current-buffer) redo-pre-buffer)
       (or redo-recording (redo-undo-updated-p) (redo-mark-updated-p))))

    
(defun redo-completes-p (fn)
  "Figure out whether the current redo is complete."
  (and fn (not (redo-addable-p fn))))

(defconst redo-allow-redo-recording t
  "This allows the recording of redo information.  It is overridden during the
redo-redo function to prevent recursion.")

(defun redo-pre-cmd-fn ()
  "Hook function to possibly record the current command"
  (if redo-allow-redo-recording  ;; condition to prevent recursive recording
      (setq redo-pre-buffer (current-buffer)
	    redo-pre-undo buffer-undo-list
	    redo-pre-mark-ring mark-ring)))


(defun redo-post-cmd-fn ()
  "Hook function to possibly record the current command"
  (if redo-allow-redo-recording
      (progn
	(if (redo-addable-p this-command)
	    (redo-add (list "KEY" (this-command-keys) this-command)))
	(if (redo-completes-p this-command)
	    (redo-complete nil)))))


(defun redo-redo-one-cmd (cmd)
  "This function redoes a single command."
  (cond
   ; This is the case when set by our commands commands
   ((equal (car cmd) "CMPL") (eval (cdr cmd)))

   ;; If the command was minibuffer-complete-and-exit then
   ;; the command was stored in the command-history
   ((equal (car cmd) "HIST") (eval (car command-history)))

   ;; Edit key.
   ((equal (car cmd) "KEY") 
    (execute-kbd-macro (nth 1 cmd)) nil)))

(defvar redo-undo-list nil
  "Records just the undo associated associated with the last redo.
This allows us to easily undo the redo for the redo-pop function.")

(defvar redo-undo-point nil
  "Records the position of point prior to redo.  This is to enable point
to be restored for redo-pop.")

(defun redo-undo ()
  "Undo the last redo by undoing all in redo-undo-list."
  (interactive)
  (while (setq redo-undo-list (primitive-undo 1 redo-undo-list)))
  (goto-char redo-undo-point))

(defun redo-redo (arg)
  "Redo all the stuff from the current element of the redo-ring"
  (interactive "p")
  (redo-complete t)
  (if (ring-empty-p redo-ring)
      (princ "Nothing-to-redo." t)
    (setq redo-undo-point (point)
	  redo-undo-list
	  (let ((buffer-undo-list nil)
		(buf (ring-ref redo-ring redo-ring-idx))
		(current-prefix-arg 1)
		(prefix-arg 1)
		(redo-allow-redo-recording nil))
	    ;; Note that undo is being recorded in a local variable
	    ;; for the duration of this function.  This may be needed by
	    ;; redo-undo
	    (while (> arg 0)
	      (mapcar 'redo-redo-one-cmd buf)
	      (setq arg (- arg 1)))
	    buffer-undo-list))
    ;; Add redo-undo-list back onto the real undo list
    (undo-boundary)
    (setq buffer-undo-list (append redo-undo-list buffer-undo-list))))

(defun redo-redo-pop ()
  "Pop the previous redo entry from the redo ring"
  (interactive)
  (let ((last real-last-command))
    (if (member last (list 'redo-redo 'redo-redo-pop))
	(progn
	  (redo-undo)
	  (setq redo-ring-idx (ring-plus1 redo-ring-idx redo-RING-SIZE))
	  (redo-redo 1))
      (error "Previous command was not redo"))))

(defun redo-enable ()
  "Enable redo."
  (redo-pre-cmd-fn)
  (add-hook 'pre-command-hook 'redo-pre-cmd-fn)
  (add-hook 'post-command-hook 'redo-post-cmd-fn))  

(defun redo-disable ()
  "Disable redo."
  (remove-hook 'pre-command-hook 'redo-pre-cmd-fn)
  (remove-hook 'post-command-hook 'redo-post-cmd-fn))

(defun redo-reset ()
  "Reset all redo state."
  (setq redo-ring (make-ring redo-RING-SIZE))
  (setq redo-buffer nil))


(define-key global-map [f12] 'redo-redo)
(define-key global-map [S-f12] 'redo-redo-pop)



;;;; debug stuff
(progn
  (redo-disable)
  (redo-reset)
  (redo-enable)
)


;;
;;(redo-post-cmd-fn)
;;redo-buffer    xx
;;redo-ring          xxxxx  x
;;redo-pre-buffer
 
;;;;
;;;;(defvar previous-command nil
;;;;  "This function is supplied to allow other functions to set 
;;;;the previous command. The previous command takes the form:-
;;;;(\"CMPL\" <func> <args1> <arg2> ...) -- Complex command
;;;;(\"KEY\" <command-keys> <func> )     -- Keyboard command
;;;;(\"HIST\" )                          -- History command
;;;;In the COMPLEX command the function is executed with the arguments. With
;;;;the KEY command the function is ignored and the function replicated by
;;;;`execute-kbd-macro' using the command-keys. The command keys are from the 
;;;;function `this-command-keys'")
;;;;
;;;;
;;;;
;;;;(defun prev-command-hook ()
;;;;  "This is the hook in to the previous command so that normal editing 
;;;;keys can be put into the redo last command"
;;;;  (interactive)
;;;;  (cond
;;;;   ;; Don't save the keyboard press if these were the last commands
;;;;   ((member this-command (list 'rerun-prev-command
;;;;			       'keyboard-quit
;;;;			       'exit-minibuffer)) t)
;;;;    
;;;;   ;; If the command was `minibuffer-complete-and-exit' then the
;;;;   ;; last command must be stored in the command history. So tell 
;;;;   ;; the rerun command to use this
;;;;   ((equal this-command 'minibuffer-complete-and-exit)
;;;;    (set-prev-cmd (list "HIST"))
;;;;   )
;;;;
;;;;   ;; Otherwise it must be just a normal key
;;;;   (t
;;;;    (set-prev-cmd (list "KEY" (this-command-keys) this-command))
;;;;   )
;;;;  )
;;;;)
;;;;
;;;;
;;;;(defun rerun-prev-command ()
;;;;  "This function reruns the last command. The command is defined in 
;;;;the variable previous-command"
;;;;  (interactive)
;;;;  (cond
;;;;
;;;;   ; This is the case when set by our commands commands
;;;;   ((equal (car previous-command) "CMPL") (eval (cdr previous-command)))
;;;;
;;;;   ;; If the command was minibuffer-complete-and-exit then
;;;;   ;; the command was stored in the command-history
;;;;   ((equal (car previous-command) "HIST") (eval (car command-history)))
;;;;
;;;;   ;; Otherwise it must have been a keyboard macro
;;;;   ((equal (car previous-command) "KEY") 
;;;;    (execute-kbd-macro (nth 1 previous-command))
;;;;   )
;;;;  )
;;;;)
;;;;
;;;;(defun set-prev-cmd ( lst )
;;;;  "This function is supplied to allow other functions to set 
;;;;the previous command. The previous command takes the form:-
;;;;(\"CMPL\" <func> <args1> <arg2> ...) -- Complex command
;;;;(\"KEY\" <command-keys> <func> )     -- Keyboard command
;;;;(\"HIST\" )                          -- History command
;;;;In the COMPLEX command the function is executed with the arguments. With
;;;;the KEY command the function is ignored and the function replicated by
;;;;`execute-kbd-macro' using the command-keys. The command keys are from the 
;;;;function `this-command-keys'"
;;;; (interactive)
;;;; (setq previous-command lst)
;;;; (print previous-command)
;;;;)
;;;;
;;;;(defun griffith-advertised-undo ()
;;;;  "This is a roof over the undo command `advertised-undo' that 
;;;;allows access so that the command can be trapped and fed into the 
;;;;previous command"
;;;;  (interactive)
;;;;  (advertised-undo)
;;;;  (set-prev-cmd (list "CMPL" 'griffith-advertised-undo))
;;;;)
;;;;
;;;;(defun griffith-call-last-kbd-macro ( &optional PREFIX )
;;;;  "This is a roof over the `call-last-kbd-macro' command to 
;;;;allows access so that the command can be trapped and fed into the 
;;;;previous command"
;;;;  (interactive "p")
;;;;  (call-last-kbd-macro PREFIX)
;;;;  (set-prev-cmd (list "CMPL" 'griffith-call-last-kbd-macro PREFIX))
;;;;)
;;;;
;;;;
;;;;
;;;;(define-key global-map [f12] 'rerun-prev-command)
;;;;
;;
