;;; frame-assoc.el --- associate buffers and applications with their frames

;; Copyright (C) 1996 Marc Munro

;; Author: Marc Munro (marc@bloodnok.demon.co.uk)
;; Version:  0.5
;; Keywords: frames

;; This file is not part of GNU Emacs but is distributed according to the
;; terms of the GNU General Public License as described below.

;; GNU Emacs is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

;;; Commentary:
;;; This file defines a general mechanism whereby buffers become
;;; associated with the frames in which they appear, and frames become
;;; associated with the buffers that they have displayed.  It also
;;; emulates frames and frame-associations for non-windows systems.
;;; A particularly useful feature is the ability to create named frames.
;;; Such frames will typically be associated with a particular task (such
;;; as handling email).  The function fa-frame-select-buf allows 
;;; selection of these frames as though they contain only a single buffer.
;;; Only when the named frame is the currently selected frame, will this
;;; function offer the user the ability to select the individual buffers
;;; associated with the frame.  This enables fa-frame-select-buf to offer
;;; a much smaller list of buffers to switch to, than would otherwise 
;;; be the case.
;;; 
;;; The behaviour of a number of standard lisp functions and commands is
;;; altered by this package, using Hans Chalupsky's advice mechanism now
;;; shipped as part of GNU Emacs.  These include:
;;;
;;;   - find-file   now may use a different frame by default according
;;;     to the values of fa-ff-use-new-frame and the universal argument.
;;;   - switch-to-buffer  now switches to the frame associated with the
;;;     specified buffer if one exists unless an argument is provided.
;;;   - kill-buffer  now removes frames if the buffer being killed is the
;;;     last buffer associated with that frame (unless an argument is
;;;     provided).
;;;   - delete-frame  now also deletes any buffers that are only
;;;     associated with this frame (unless an argument is provided).
;;;   - bury-buffer  now by default acts only on the current frame.
;;;     if an argument is provided it buries the current buffer and then
;;;     switches to the most apropriate frame to display the newly selected 
;;;     buffer.
;;;
;;; This package can be used simply by loading it and running fa-frame-assoc
;;; There is no need to modify any key-bindings or load anything else, 
;;; however there are a number of customisations you may wish to perform.
;;;
;;; Functions you may wish to bind to keys:
;;; - fa-frame-select-buf
;;; - fa-name-frame
;;; - fa-create-named-frame
;;; - fa-frame-select-next
;;; - fa-frame-select-prev
;;;
;;; Variables you may wish to modify:
;;; - fa-ff-use-new-frame
;;; - fa-activate-hook
;;; - fa-deactivate-hook
;;; - fa-default-frame-parameters
;;; - fa-use-default

;;; Bugs:
;;;   None that I know of at the moment, but email them to me.

;;; Features:
;;;   As it stands, frame-assoc.el will probably not peacefully coexist with
;;;   any packages that perform their own frame handling (such as VM).  DO
;;;   NOT DESPAIR.  The awk script fa-adint.awk which should have been 
;;;   provided along with this file will deal with this problem for you.
;;;   Run this script against all files that contain functions that do
;;;   their own frame handling, and it will create a file of advice that
;;;   enables these functions to coexist with frame-assoc.el.  Load this
;;;   file after loading frame-assoc.el and all will be well.  Your .emacs
;;;   file will contain a few lines looking like this:
;;;
;;; ;; Load frame-assoc and the related advice for vm (created by fa-adint.awk)
;;; ;; and start frame-assoc.
;;; (load "~marc/emacs/frame-assoc.el")
;;; (load "~marc/emacs/fa-vm-hooks.el")
;;; (fa-frame-assoc 1)

;;; To do:
;;; - implement a frame-assoc version of the buffer menu (any volunteers?)
;;; - allow warping of mouse to selected frame
;;; - automatically take focus to raised frames.
;;; I invite any and all comments on the above (especially how to do it!),
;;; as well as any further ideas for functionality.


(require 'advice)

;; FUNDAMENTAL VARIABLES

(defvar fa-defer-operations nil
  "Boolean defining whether or not delete-frame will immediately delete the 
frame it has been asked to and whether frames are associated with buffers
during a make-frame or not.  If non-nil, the frame associations and deletions
are deferred until fa-perform-deferred is run.
It exists primarily to allow packages such as vm to handle buffers and 
frames for themselves.

The usual usage of this will be something like:

(defadvice vm-quit (around vm-frame-assoc first nil activate)
  \"vm-quit is advised to set fa-nokill-buffers to prevent frame-assoc code
from automatically killing-buffers associated with frames that vm is
deleting.\"
  (let ((fa-defer-operations t))
    ad-do-it)
  (fa-perform-deferred))
")

(defvar fa-select-buffer-fn 'switch-to-buffer
  "This is the function that will be called when fa-frame-select-buf is
invoked when frame-association is inactive.")

(defvar fa-ff-use-new-frame 1
  "Whether or not find-file will automatically use a new frame by default.
A value of t means create a new frame unless an argument is provided, nil
means don't create a new frame, and any other value means create a new
frame if an argument is given.")

(defvar fa-first-frame (selected-frame)
  "This is the first frame of the session.  It is most important for 
sessions in non-windows environments.")

(defvar fa-buffer-alist nil
  "Alist of known buffers to their frames.")

(defvar fa-frame-alist nil
  "Alist of the known frame-entries.")

(defvar fa-frame-names nil
  "Alist of frames to frame names.")

(defvar fa-win-configs nil
  "Alist of screens to window configurations.")

(defvar fa-cur-screen 0
  "Integer used for allocating screen definitions.")

(defvar fa-window-height nil
  "Alist of frames to frame heights.  Only used in non-windows environment.")

(defvar fa-window-width nil
  "Alist of frames to frame width.  Only used in non-windows environment.")

(defvar fa-cur-height (frame-height)
  "Current frame height.  Used when creating new frames in non-windows 
environment.")

(defvar fa-cur-width (frame-width)
  "Current frame width.  Used when creating new frames in non-windows 
environment.")

(defvar fa-deferred-list nil
  "List of frames to possibly be removed by fa-perform-deferred.")

;; DEFAULT FRAME PARAMETERS

(defvar fa-use-default t
  "Variable dscribing whether or not to use the default frame configuration
parameters or not.  If set to nil, then the default frame parameters set up
in fa-default-frame-parameters will be used instead.")

(defvar fa-default-frame-parameters
  '((width . 80) 
    (height . 40) 
    (cursor-type . box)
    (icon-type t)                  ;; Use the default Gnu icon.
    (border-color . "white") 
    (cursor-color . "RoyalBlue") 
    (mouse-color . "black") 
    (background-color . "PapayaWhip")
    (foreground-color . "black") 
    (vertical-scroll-bars . t)
    (internal-border-width . 2)
    (border-width . 2)
    (font . "-Adobe-Courier-Medium-R-Normal--12-120-75-75-M-70-ISO8859-1"))
  "Set of sensible defaults for emacs windows.  This should be used if 
there are no default settings (if the cursor and foreground colours are
 black and the  background colour is white we assume no default has been
defined), or if fa-use-default is nil.")


;; IDENTIFY FLOATING, DISPLAYED, AND INVISIBLE BUFFERS
;;
;; Floating buffers are those that can just turn up from time to time.  
;; They are not associated with any particlar frame (hence the term
;; floating). Invisible ones are those that begin with a space, and 
;; are not being displayed.  Invisible buffers are also classed as floating.

(defvar fa-floating-buffers
  '("\*Buffer List\*" "\*scratch\*" "\*compilation\*" "\*Help\*" 
    "\*Occur\*" "\*Shell Command Output\*" "\*Completions\*"
    "\*Backtrace\*" "\*Messages\*" "\*Compil")
  "Regexps for buffers which will not belong to specific frames.")

(defun fa-frame-displayed-buffers (frame)
  "Return a list of buffers displayed in FRAME."
  (if (frame-live-p frame)
      (let (win first-win cur-buf buffers)
	(setq first-win (setq win (frame-selected-window frame)))
	(while (progn
		 (setq cur-buf (window-buffer win))
		 (if (not (memq cur-buf buffers))
		     (setq buffers (cons cur-buf buffers)))
		 (setq win (next-window win 1 nil))
		 (not (eq win first-win))))
	buffers)
    nil))

(defun fa-buffer-invisible-p (buffer)
  "Predicate identifying whether or not BUFFER should be listed in buffer
and frame lists.  Note that a buffer is not classed as invisible if it is
being displayed (we only need to check the current frame)."
  (let (result)
    (if (memq buffer (fa-frame-displayed-buffers (car (car fa-frame-alist))))
	nil
      (save-match-data
	;; Check for buffer name beginning with " *".
	(setq result (eq 0 (string-match " \\*" (buffer-name buffer))))))
    result))

(defun fa-buffer-floats-p (buffer)
  "Predicate describing whether BUFFER floats or not.  Invisible buffers are
considered to float."
  (let ((bufname (buffer-name buffer))
	(float-list fa-floating-buffers)
	floats)
    (if (fa-buffer-invisible-p buffer)
	t
      (while float-list
	(save-match-data
	  (if (string-match (car float-list) bufname)
	      (setq float-list nil
		    floats t)))
	(setq float-list (cdr float-list)))
      floats)))


;; IDENTIFY PREVIOUSLY UNKNOWN NON-FLOATING BUFFERS

(defun fa-new-bufs ()
  "Return a list of new visible and non-floating buffers."
  (let ((buflist (buffer-list))
	cur-buf
	new-bufs)
    (while buflist
      (setq cur-buf (car buflist))
      (if (or (assoc cur-buf fa-buffer-alist)
	      (fa-buffer-floats-p cur-buf))
	  nil
	;; Must be a new non-floating buffer
	(setq new-bufs (cons cur-buf new-bufs)))
      (setq buflist (cdr buflist)))
    new-bufs))

(defun fa-new-displayed-bufs (frame)
  "Return a list of buffers displayed in FRAME which were previously not
associated with it."
  (let ((known-bufs (cdr (assoc frame fa-frame-alist)))
	first-win
	cur-win
	cur-buf
	new-bufs)
    (if (frame-live-p frame)
	(progn
	  (setq cur-win (setq first-win (frame-selected-window frame)))
	  (while (progn
		   (setq cur-buf (window-buffer cur-win))
		   (if (or (memq cur-buf known-bufs)
			   (fa-buffer-floats-p cur-buf))
		       nil
		     (setq new-bufs (cons cur-buf new-bufs)))
		   (setq cur-win (next-window cur-win 1 nil))
		   (not (eq cur-win first-win))))
	  new-bufs)
      nil)))
    

(defun fa-windows-for-buffer (buf)
  "Return a list of windows showing BUF."
  (let ((frames (cdr (assoc buf fa-buffer-alist)))
	windows
	cur-frame
	first-win
	cur-win)
    (if (not (member (selected-frame) frames))
	(setq frames (cons (selected-frame) frames)))
    (while frames
      (setq cur-frame (car frames))
      (setq frames (cdr frames))
      (if (not (frame-live-p cur-frame))
	  (fa-delete-frame cur-frame t nil)
	(setq cur-win (setq first-win (frame-selected-window cur-frame)))
	(while (progn
		 (if (eq buf (window-buffer cur-win))
		     (setq windows (cons cur-win windows)))
		 (setq cur-win (next-window cur-win 1 nil))
		 (not (eq cur-win first-win))))))
    windows))

;; FUNCTIONS FOR ASSOCIATING and DISSOCIATING BUFFERS and FRAMES, FOR
;; IDENTIFYING THE BEST BUFFERS FOR DISPLAY ON A FRAME, THE BEST FRAMES
;; FOR DISPLAYING A BUFFER, AND FOR IDENTIFYING WHEN FRAMES NO LONGER 
;; HAVE BUFFERS ASSOCIATED WITH THEM.

(defun fa-associate (buf frame)
  "Add BUF to the alist entry for FRAME, and vice versa iff they are not 
already associated."
  (fa-frame-trace (concat "Associating "
			  (prin1-to-string buf)
			  " with "
			  (prin1-to-string frame)))
  (let ((cur-buf (assoc buf fa-buffer-alist))
	(cur-frame (assoc frame fa-frame-alist)))
    (if cur-buf
	(if (memq frame (cdr cur-buf))
	    nil
	  (setcdr cur-buf (cons frame (cdr cur-buf))))
      (setq fa-buffer-alist 
	    (cons (cons buf (list frame))
		  fa-buffer-alist)))
    (if cur-frame
	(if (memq buf (cdr cur-frame))
	    nil
	  (setcdr cur-frame (cons buf (cdr cur-frame))))
      (setq fa-frame-alist 
	    (cons (cons frame (list buf))
		  fa-frame-alist)))))

(defun fa-dissociate (buf frame)
  "Remove references to BUF from the frame alist entry for FRAME and vice 
versa."
  (let ((cur-buf (assoc buf fa-buffer-alist))
	(cur-frame (assoc frame fa-frame-alist)))
    (if cur-buf
	(if (memq frame (cdr cur-buf))
	    (setcdr cur-buf (delq frame (cdr cur-buf)))))
    (if cur-frame
	(if (memq buf (cdr cur-frame))
	    (setcdr cur-frame (delq buf (cdr cur-frame)))))))

(defun fa-frame-empty-p (frame)
  "Predicate returning t if frame is not known to have buffers associated
with it."
  (not (cdr (assoc frame fa-frame-alist))))

(defun fa-best-buffer (frame)
  "Return the most appropriate buffer to display on the given FRAME.  This 
will be called when a buffer is deleted, so that only buffers associated with
FRAME ever get displayed on it.  By choice it will return a buffer that is
not being displayed elsewhere."
  (let* ((buflist (cdr (assoc frame fa-frame-alist)))
	 (first (car buflist))
	 (next first)
	 (windows)
	 (result))
    (while next
      (if (and (setq windows (fa-windows-for-buffer next))
	       (not (equal (list (selected-window)) windows)))
	  (progn  ;; the buffer is displayed in a window other than 
	          ;; the selected one
	    (setq buflist (cdr buflist))
	    (setq next (car buflist)))
	(setq result next)
	(setq next nil)))
    (or result
	first
	(get-buffer "*scratch*")
	(switch-to-buffer "*scratch"))))

(defun fa-best-frame (buf)
  "Return the best frame for displaying buffer BUF.  This will be called when
switch-to-buffer has selected a buffer which may be in a different frame.
If BUF floats or is invisible, the current frame will be returned."
  (if (fa-buffer-floats-p buf)
      (selected-frame)
    (let ((frames (cdr (assoc buf fa-buffer-alist)))
	  cur-frame
	  is-possible
	  is-active-frame
	  is-displayed-on)
      (if (null frames)
	  (selected-frame)
	(setq is-possible (car frames))
	(while frames
	  (setq cur-frame (car frames))
	  (setq frames (cdr frames))
	  (if (not (frame-live-p cur-frame))
	      (progn
		(fa-dissociate buf cur-frame)
		(fa-check-frame cur-frame)
		(setq is-displayed-on (fa-best-frame buf))
		(setq frames nil))
	    (if (eq t (frame-visible-p cur-frame))
		(setq is-active-frame cur-frame))
	    (if (memq buf (fa-frame-displayed-buffers cur-frame))
		(progn
		  (setq is-displayed-on cur-frame)
		  (if (eq t (frame-visible-p cur-frame))
		      (setq frames nil)))))) ;; get out of the loop early
	(or is-displayed-on
	    is-active-frame
	    is-possible)))))


;; FUNCTIONS FOR MAKING ASSOCIATIONS HAPPEN, AND FOR FRAME NAVIGATION AT THE
;; FA DATA STRUCTURES LEVEL

(defun fa-leave-frame (frame)
  "Update the list of buffers associated with FRAME as we leave it.  Also,
if we are not in a windows system, record the current screen configuration
and height and width."
  (let ((new-buffers (append (fa-new-bufs)
			     (fa-new-displayed-bufs frame))))
    (mapcar (lambda (buf)
	      (fa-associate buf frame))
	    new-buffers))
  (if window-system
      nil
    (setcdr (assoc frame fa-win-configs) (current-window-configuration))
    (setq fa-cur-height (setcdr (assoc frame fa-window-height) (frame-height)))
    (setq fa-cur-width (setcdr (assoc frame fa-window-width) (frame-width)))))

(defun fa-internal-goto (frame)
  "Make FRAME the first element of fa-frame-alist."
  (if (eq frame (car (car fa-frame-alist)))
      nil
    (let* ((prev fa-frame-alist)
	   (next (cdr prev)))
      (while (and next
		  (not (eq frame (car (car next)))))
	(setq prev next)
	(setq next (cdr next)))
      ;; Get here with next containing frame, prev containing previous frame.

      (if next
	  (progn
	    (setcdr prev (cdr next))
	    (setcdr next fa-frame-alist)
	    (setq fa-frame-alist next))))))

(defun fa-internal-next ()
  "Rotate fa-frame-list to make element 2 the first element and element 1
the last."
  (let ((the-rest (cdr fa-frame-alist))
	last)
    (if the-rest
	(progn
	  (setq last the-rest)
	  (while (cdr last) (setq last (cdr last))) ;; find last element
	  (setcdr fa-frame-alist nil)
	  (setcdr last fa-frame-alist)
	  (setq fa-frame-alist the-rest)))))


;; FUNCTIONS FOR DEALING WITH NAMED FRAMES AT THE FA DATA STRUCTRUES LEVEL

(defun fa-name-frame (name &optional frame)
  "Give the name NAME to FRAME."
  (interactive "sFrame name: ")
  (or frame (selected-frame))
  (let ((entry (assoc frame fa-frame-names)))
    (if entry
	(setcdr entry name)
      (setq fa-frame-names
	    (cons (cons frame name)
		  fa-frame-names))))
  (if window-system
      (modify-frame-parameters frame (list (cons 'name name)))))

(defun fa-frame-name (frame)
  "Return the name of FRAME."
  (or (cdr (assoc frame fa-frame-names))
      ""))

(defun fa-internal-next-unnamed ()
  "Goto the next unnamed screen."
  (let ((first-frame-entry (car fa-frame-alist))
	cur-frame-entry)
    (if first-frame-entry
	(progn
	  (fa-internal-next)
	  (while (and (not (eq first-frame-entry 
			       (setq cur-frame-entry (car fa-frame-alist))))
		      (assoc (car cur-frame-entry) fa-frame-names))
	    (fa-internal-next))))))

(defun fa-internal-goto-named (name)
  "Make the frame called NAME the first element of fa-frame-alist."
  (let ((name-list fa-frame-names))
    (while name-list
      (if (string= name (cdr (car name-list)))
	  (progn
	    (fa-internal-goto (car (car name-list)))
	    (setq name-list nil))
	(setq name-list (cdr name-list))))))


;; ADVICE FOR FRAME FUNCTIONS USED IN FA INTERNALS.  THIS ADVICE ALLOWS
;; THIS PACKAGE TO BE USED IN A NON-WINDOWS ENVIRONMENT.

(defadvice frame-selected-window (around frame-assoc first)
  "If in a non-windows environment, just return the currently selected window."
  (if window-system
      ad-do-it
    (setq ad-return-value (car (car fa-frame-alist)))))

(defadvice frame-visible-p (around frame-assoc first)
  "If in a non-windows environment, determine whether the supplied screen
number exists and is at the head of fa-frame-alist."
  (if window-system
      ad-do-it
    (setq ad-return-value 
	  (eq (ad-get-arg 0)
	      (car (car fa-frame-alist))))))
	    
(defadvice frame-live-p (around frame-assoc first)
  "If in a non-windows environment, determine whether the supplied screen
number exists."
  (if window-system
      ad-do-it
    (setq ad-return-value 
	  (assoc (ad-get-arg 0) fa-frame-alist))))

(defadvice window-frame (around frame-assoc first)
  "If in a non-windows environment, determine whether the supplied screen
number exists."
  (if window-system
      ad-do-it
    (setq ad-return-value 
	  (car (car fa-frame-alist)))))

(defadvice select-frame (around frame-assoc first)
  "If in a non-windows environment, just return the parameter."
  (if window-system
      ad-do-it
    (setq ad-return-value (ad-get-arg 0))))

;; INTERNAL UTILITIES THAT DO THINGS TO BUFFERS

(defun fa-delete-dis-buffers (buflist kill-quietly)
  "Function to delete dissociated buffers.  BUFLIST is a list of buffers which
may or may not be associated with frames.  If KILL-QUIETLY is t then buffers
will be deleted without asking."
  (let (cur-buf framelist alist)
    (while buflist
      (setq cur-buf (car buflist))
      (setq buflist (cdr buflist))
      (setq framelist (cdr (setq alist (assoc cur-buf fa-buffer-alist))))
		    
      ;; Kill the buffer if there are no more frames that use this buffer
		    
      (and (null framelist)
	   (or kill-quietly
	       (y-or-n-p (concat "Delete buffer " (buffer-name cur-buf) " ")))
	   (kill-buffer cur-buf)))))

(defconst fa-deleting-frame nil
  "Whether or not we are in the process of deleting a frame.  The advised
kill-buffer needs to know this.")

(defun fa-delete-frame (frame kill-quietly call-delete-frame)
  "Delete the records associated with FRAME, if KILL-QUIETLY is t then buffers
will be deleted without asking, if CALL-DELETE-FRAME is t then delete the 
frame."
  (let* ((fa-deleting-frame t)
	 (buflist (cdr (assoc frame fa-frame-alist)))
	 (buflist-again buflist))
    (while buflist
      (fa-dissociate (car buflist) frame)
      (setq buflist (cdr buflist)))
    (fa-delete-dis-buffers buflist-again kill-quietly)
    (if window-system  ;; delete window configuration stuff if needed.
	nil
      (setq fa-win-configs (delq (assoc frame fa-win-configs) fa-win-configs))
      (setq fa-window-height 
	    (delq (assoc frame fa-window-height) fa-window-height))
      (setq fa-window-width 
	    (delq (assoc frame fa-window-width) fa-window-width)))
    (setq fa-frame-names (delq (assoc frame fa-frame-names) fa-frame-names))
    (setq fa-frame-alist (delq (assoc frame fa-frame-alist) fa-frame-alist))
    (if call-delete-frame
	(delete-frame frame))
    (if (and fa-frame-alist window-system)    ;; Must not record the current 
	(fa-leave-frame (selected-frame)))))  ;; screen config if we are not
                                              ;; in a windows system.
;; EXTERNALLY CALLABLE FUNCTIONS.

(defun fa-perform-deferred ()
  "Delete any frames in fa-deferred-list that do not have buffers associated
with them, and then use fa-leave-frame to ensure that frame associations
are up to date."
  (let (frame)
    (while fa-deferred-list
      (progn
	(setq frame (car fa-deferred-list))
	(setq fa-deferred-list (cdr fa-deferred-list))
	(if (null (cdr (assoc frame fa-frame-alist)))
	    (fa-delete-frame frame nil t)))))
  (fa-leave-frame (selected-frame))
  (setq fa-deferred-new-frames 0))


(defun fa-frame-exists (name)
  "Does the frame named NAME exist?  Return the frame."
  (let ((name-list fa-frame-names)
	result)
    (while name-list
      (if (string= name (cdr (car name-list)))
	  (progn
	    (setq result (car (car name-list)))
	    (setq name-list nil))
	(setq name-list (cdr name-list))))
    (if result
	(if (frame-live-p result)
	    result
	  ;; If the frame no longer lives, delete it
	  (fa-delete-frame result nil nil)
	  (fa-frame-exists name))
      nil)))

(defun fa-create-named-frame (name &optional frame-params)
  "Create a new frame with the name NAME, and the optional parameters
FRAME-PARAMS.  Any dead frames are dealt with as they are found." 
  (interactive "sFrame name: ")
  (fa-leave-frame (selected-frame))
  (fa-name-frame name (select-frame (new-frame frame-params)))
  (fa-internal-goto-named name))

(defun fa-goto-named (name &optional create frame-params)
  "Raise the frame given by NAME, where NAME is the name associated with
the frame.  The optional FRAME-PARAMS will be passed to any eventual call to 
make-frame."
  (let ((frame (fa-frame-exists name)))
    (if frame
	(fa-select frame)
      (if create
	  (fa-create-named-frame name frame-params)))))

(defun fa-check-frame (frame)
  "Determine whther FRAME is live, and if not delete it.  Return t if frame
is live."
  (if (frame-live-p frame)
      t
    (fa-delete-frame frame (yes-or-no-p 
		 "Dead frame detected.  Delete its buffers quietly? ") nil)
    nil))

(defun fa-next-frame ()
  "Raise the next frame in the frame list."
  (interactive)
  (let ((old-frame (selected-frame))
	new-frame)
    (fa-leave-frame old-frame)
    (fa-internal-next)
    (if (fa-check-frame (setq new-frame (car (car fa-frame-alist))))
	(progn
	  (raise-frame new-frame)
	  (select-frame new-frame))
	(fa-next-frame))))

(defun fa-select (frame)
  "Raise the frame given by FRAME.  If FRAME is dead stay with the current 
frame"
  (let ((old-frame (selected-frame)))
    (fa-leave-frame old-frame)
    (fa-internal-goto frame)
    (if (fa-check-frame (car (car fa-frame-alist)))
	(progn
	  (raise-frame frame)
	  (select-frame frame))
      (fa-internal-goto old-frame))))

(defun fa-next-unnamed ()
  "Goto the next unnamed screen."
  (let ((framelist (cdr fa-frame-alist))
	(frame))
    (while framelist
      (setq frame (car (car framelist)))
      (setq framelist (cdr framelist))
      (if (assoc frame fa-frame-names)
	  nil
	(setq framelist nil)
	(fa-select frame)))))


;; SUPPORT STUFF FOR FA-FRAME-SELECT-BUF

(defvar fa-frame-select-map nil
  "A copy of the minibuffer completion keymap, with all keys that were
bound to next and previous line now bound to fa-frame-select-next and prev.
This setting-up is performed below.")

(defun fa-frame-setup-keys ()
  "Function to define keybindings at first use of fa-frame-select-buf 
function."
  (if (not fa-frame-select-map)
      (progn
	(setq fa-frame-select-map
	      (copy-keymap minibuffer-local-completion-map))
	(mapcar
	 (function
	  (lambda (keyseq)
	    (define-key fa-frame-select-map keyseq 'fa-frame-select-prev)))
	 (where-is-internal 'previous-line nil nil))
	
	(mapcar
	 (function
	  (lambda (keyseq)
	    (define-key fa-frame-select-map keyseq 'fa-frame-select-next)))
	 (where-is-internal 'next-line nil nil)))))

(defun fa-frame-select-next ()
  "Find the next frame in frame-buf-list."
  (interactive)
  (erase-buffer)
  (setq fa-frame-select-index (+ 1 fa-frame-select-index))
  (if (>= fa-frame-select-index (length frame-buf-list))
      (setq fa-frame-select-index 0))
  (insert (nth fa-frame-select-index frame-buf-list)))

(defun fa-frame-select-prev ()
  "Find the previous frame in the fa-frame-list."
  (interactive)
  (erase-buffer)
  (setq fa-frame-select-index (- fa-frame-select-index 1))
  (if (< fa-frame-select-index 0)
      (setq fa-frame-select-index (- (length frame-buf-list) 1)))
  (insert (nth fa-frame-select-index frame-buf-list)))

(defun fa-frame-make-buflist ()
  "Function to create a list of buffers and frames.
This is ordered so that buffers which belong to the current frame appear 
first, then named frames, then buffers belonging to unnamed frames and 
finally floating buffers (those that are ignored for the purposes of
association with frames)."
  (let ((framelist (cdr fa-frame-alist))
	(buflist (buffer-list))
	(cur-buffer (current-buffer))
	cur-frame-buffers
	buffers-accounted-for
	frame-names
	floating-buffers
	other-buffers
	cur-frame
	elem
	frame-buffers
	name)

    ;; First we have buffers associated with the current frame.
    ;; After making sure that the current buffer is at the tail
    ;; of cur-frame-buffers.

    (setq cur-frame-buffers (cdr (car fa-frame-alist)))
    (setq cur-frame (append (delq cur-buffer cur-frame-buffers)
			    (cons cur-buffer nil)))
    (setq buffers-accounted-for cur-frame-buffers)

    (setq cur-frame-buffers
	  (mapcar (lambda (buf)
		    (buffer-name buf))
		  cur-frame-buffers))
    
    ;; then buffers from unnamed frames

    (while framelist
      (setq cur-frame (car framelist))
      (setq framelist (cdr framelist))
      (setq name (cdr (assoc (car cur-frame) fa-frame-names)))
      (if (cdr cur-frame)
	  (progn
	    (setq frame-buffers (cdr cur-frame))
	    (while frame-buffers
	      (setq elem (car frame-buffers))
	      (setq frame-buffers (cdr frame-buffers))
	      (if (memq elem buffers-accounted-for)
		  nil
		(setq buffers-accounted-for 
		      (append buffers-accounted-for (list elem)))
		(if (null name)
		    (setq other-buffers 
			  (append other-buffers 
				  (list (buffer-name elem))))))))))

    ;; then named frames (but not the current one)

    (setq framelist (cdr fa-frame-alist))
    (while framelist
      (setq cur-frame (car framelist))
      (setq framelist (cdr framelist))
      (if (setq name (cdr (assoc (car cur-frame) fa-frame-names)))
	  (setq frame-names (append frame-names (list name)))))

    ;; Now visible unassociated buffers

    (while buflist
      (setq elem (car buflist))
      (setq buflist (cdr buflist))
      (if (or (fa-buffer-invisible-p elem)
	      (memq elem buffers-accounted-for))
	  nil
	(if (fa-buffer-floats-p elem)
	    (setq floating-buffers (cons (buffer-name elem) floating-buffers))
	  (setq other-buffers (cons (buffer-name elem) other-buffers)))))

    ;; Now put 'em together

    (append cur-frame-buffers frame-names other-buffers floating-buffers)))


(defun fa-frame-select-buf (&optional cur-frame)
  "Function to fully integrate buffer and frame selection in one command.
This is so cool.  The optional cur-frame parameter causes a buffer 
selection to be made in the current-frame.  This can be done interactively
by setting a prefix argument."
  (interactive "P")
  (if (not fa-frame-assoc-active)
      (call-interactively fa-select-buffer-fn)
    (fa-leave-frame (car (car fa-frame-alist)))
    (fa-frame-setup-keys)
    (let* ((save-minibuf-map minibuffer-local-completion-map)
	   (fa-frame-select-index 0)
	   default
	   selection)
      (setq frame-buf-list (fa-frame-make-buflist))
      (setq default (car frame-buf-list))
      (setq selection
	    (unwind-protect
		(progn
		  (setq minibuffer-local-completion-map fa-frame-select-map)
		  (setq fa-frame-select-index 0)
		  (completing-read (concat "Switch to: (default "
					   default
					   ") ")
				   (mapcar '(lambda (str)
					      (list str))
					   frame-buf-list)))
	      (setq minibuffer-local-completion-map save-minibuf-map)))
      (if (string= selection "")
	  (setq selection default))
      (if (fa-frame-exists selection)
	  (fa-goto-named selection)
	(switch-to-buffer selection)))))


;; NOW ADVICE FOR FRAME, FILE AND BUFFER NAVIGATION ETC.

(defconst fa-making-new-frame nil
  "Boolean describing whether we are in code that creates a new frame or not.
DON'T TOUCH!")

(defun fa-make-screen ()
  "Return a new screen number, and set up an entry in fa-win-configs for
the new screen."
  (setq fa-cur-screen (1+ fa-cur-screen))
  (setq fa-win-configs 
	(cons (cons fa-cur-screen (current-window-configuration))
	      fa-win-configs))
  (setq fa-window-height
	(cons (cons fa-cur-screen fa-cur-height)
	      fa-window-height))
  (setq fa-window-width
	(cons (cons fa-cur-screen fa-cur-width)
	      fa-window-width))
  fa-cur-screen)

(defvar fa-deferred-new-frames 0
  "Count of frames that have been created while fa-defer-operations has
been set.  This is used to show a warning message.")

(defadvice make-frame (around frame-assoc first)
  "Inform frame-utils of the existence of the new frame."
  (fa-frame-trace "before make-frame")
  (if (null fa-making-new-frame)
      (if fa-defer-operations
	  ;; Make a not of the deferral, so we know it's not going
	  ;; to get out of hand
	  (if (> (setq fa-deferred-new-frames (1+ fa-deferred-new-frames)) 1)
	      (message (concat "Frame-assoc: confusion level "
			       (int-to-string fa-deferred-new-frames))))
	;; we are not deferring the operation, so do some associating.
	(fa-leave-frame (selected-frame))))
  (if window-system
      (let ((fa-making-new-frame t))
	ad-do-it
	(select-frame ad-return-value)
	;; If the current buffer is already associated with a frame, we
	;; do not show it in the new frame, but select *scratch* instead.
	(if (assoc (current-buffer) fa-buffer-alist)
	    (switch-to-buffer "*scratch*"))
	(setq fa-frame-alist (cons (cons ad-return-value nil) fa-frame-alist)))
    (delete-other-windows (selected-window))
    (switch-to-buffer "*scratch*")
    (setq ad-return-value (fa-make-screen))
    (setq fa-frame-alist
	  (cons (cons ad-return-value nil) fa-frame-alist)))
  (fa-frame-trace "after make-frame"))

(defadvice special-display-popup-frame (around frame-assoc first)
  "Inform frame-utils of the existence of the new frame."
  (fa-leave-frame (selected-frame))
  (let ((fa-making-new-frame t))
    ad-do-it))

(defadvice find-file-other-frame (around frame-assoc first)
  "Advice to prevent make-frame from updating the buffer list until after
the new frame has been created."
  (let ((fa-making-new-frame t))
    (fa-frame-trace "before find-file-other-frame")
    (fa-leave-frame (selected-frame))
    (if window-system
	ad-do-it
      (make-frame)
      (find-file (ad-get-arg 0)))
    (fa-associate (current-buffer) (selected-frame))
    (fa-frame-trace "before find-file-other-frame")))

(defadvice find-file (around frame-assoc first)
  "Make find-file create a frame unless the file is already open, or we 
have  been called with an argument, or the current frame owns no buffers."
  (fa-frame-trace "before find-file")
  (let ((desire-new-frame
	 (cond 
	  ((eq fa-ff-use-new-frame t)
	   (not current-prefix-arg))
	  ((eq fa-ff-use-new-frame nil)
	   nil)
	  (t
	   current-prefix-arg))))
    (fa-leave-frame (selected-frame))
    (if (or (null (cdr (assoc (selected-frame) fa-frame-alist)))
	    (get-file-buffer (ad-get-arg 0))
	    (not desire-new-frame))
	(progn
	  ad-do-it
	  (fa-associate (current-buffer) (selected-frame)))
      (find-file-other-frame (ad-get-arg 0)))
    (fa-frame-trace "after find-file")))

(defadvice delete-frame (around frame-assoc first)
  "If we are not in a windows environment do nothing, else delete the 
frame-details and any wholly owned buffers before deleting the frame.  
If arg is supplied, delete buffers interactively."
  (if fa-defer-operations
      (setq fa-deferred-list (cons frame fa-deferred-list))
    (let ((frame (or (ad-get-arg 0)
		     (selected-frame)))
	  (arg current-prefix-arg))
      (fa-leave-frame frame)
      (fa-delete-frame frame t nil)
      (if window-system
	  ad-do-it
	(raise-frame (car (car fa-frame-alist)))))))

(defadvice switch-to-buffer (around frame-assoc first)
  "Make switch-to-buffer switch to the frame that is currently displaying
the buffer, if such a one exists unless called with an argument."
  (interactive "bSwitch to buffer: ") 
  (if current-prefix-arg
      ad-do-it
    (let* ((buf (get-buffer (ad-get-arg 0)))
	   (target-frame (fa-best-frame buf)))
      (if (eq (selected-frame) target-frame)
	  nil
	(fa-leave-frame (selected-frame))
	(fa-internal-goto target-frame)
	(raise-frame target-frame)
	(select-frame target-frame))
      ad-do-it)))

(defadvice bury-buffer (around frame-assoc last)
  "If no argument is provided, constrain the bury-buffer operation to the
 selected-frame.  If an argument is provided, follow more closely the original 
 functionality (but taking into account frame associations), and select  
 other-buffer in its natural frame."
   (let* ((buf (get-buffer (or (ad-get-arg 0)
 			      (current-buffer))))
 	 (newbuf (other-buffer buf t))
 	 (buflist (buffer-list))
 	 (frame-disp-bufs (fa-frame-displayed-buffers (selected-frame)))
 	 (frame-entry (assoc (selected-frame) fa-frame-alist))
 	 done)
     (ad-set-arg 0 buf)
     ad-do-it
     (while (and (not done)
 		(setq buf (car buflist)))
       (cond ((fa-buffer-invisible-p buf) 
 	     ;; ignore invisible buffers
 	     nil)   
 	    ((memq buf frame-disp-bufs)
 	     ;; ignore buffers already appearing in this frame.
 	     nil)    
 	    ((fa-buffer-floats-p buf)
 	     ;; switch to the floating buffer in this frame
 	     (setq done t))
 	    (current-prefix-arg
 	     ;; this is a non-floating buffer and we are using all frames
 	     ;; (as prefix arg is provided) so select it in its natural frame.
 	     (setq done t))
 	    (t
 	     ;; We are only looking for buffers associated with this frame.
 	     (if (memq buf (cdr frame-entry))
 		 (setq done t))))
       (setq buflist (cdr buflist)))
     (if done
 	(let ((current-prefix-arg (not current-prefix-arg)))
 	  (switch-to-buffer buf))
       (switch-to-buffer "*scratch*"))))
 
(defadvice kill-buffer (around frame-assoc first)
  "If buffer is killed remove all references to this buffer from 
fa-buffer-alist and fa-frame-alist.  If this leaves a frame with no
buffers, then delete the frame unless called with an argument.  Also, 
ensure that any windows displaying this buffer, now display a buffer
appropriate to the frame in which they appear."
  (interactive "bKill buffer: ")
  (if fa-deleting-frame
      ad-do-it
    (let* ((param (ad-get-arg 0))
	   (buf (if (stringp param)
		    (get-buffer param)
		  param))
	   (bufname (buffer-name buf))
	   (windows (fa-windows-for-buffer buf))
	   frames
	   cur-win
	   cur-frame)
      (fa-leave-frame (selected-frame))
      (setq frames (cdr (assoc buf fa-buffer-alist)))
      (fa-frame-trace (concat "before kill-buffer " 
			      (prin1-to-string buf) " " 
			      (prin1-to-string windows)))
      (if (get-buffer bufname)
	  (progn
	    ad-do-it
	    (if ad-return-value   ;; ie, if kill was successful
		(progn
		  ;; Start by dissociating killed buffer from frames
		  (while frames
		    (setq cur-frame (car frames))
		    (setq frames (cdr frames))
		    (fa-dissociate buf cur-frame)
		    (if (not (frame-live-p cur-frame))
			(fa-delete-frame cur-frame t nil)))
		  ;; Now remove the buffer entry from fa-buffer-alist
		  (setq fa-buffer-alist (delq (assoc buf fa-buffer-alist) 
					      fa-buffer-alist))
		  ;; Now select other buffers for windows in frames that 
		  ;; should persist and delete frames that should not
		  (while windows
		    (setq cur-win (car windows))
		    (setq windows (cdr windows))
		    (if (window-live-p cur-win)
			(progn
			  (setq cur-frame (window-frame cur-win))
			  (if (and (null current-prefix-arg)
				   (cdr fa-frame-alist) ;; must be more than 1
				   (null (cdr (assoc cur-frame 
						     fa-frame-alist))))
			      (delete-frame cur-frame)
			    (set-window-buffer 
			     cur-win (fa-best-buffer cur-frame))))))))
	    (fa-frame-trace (concat "at end of kill-buffer " 
				    (buffer-name buf))))))))

(defadvice handle-switch-frame (after frame-assoc first)
  "Identify for frame_utils that a frame switch has occurred."
  (fa-leave-frame (car (car fa-frame-alist)))
  (fa-internal-goto (selected-frame)))

(defadvice raise-frame (around frame-assoc first)
  "If we are not in a windows environment, we take a screen definition as
the parameter and restore the window configuration from it."
  (if window-system
      ad-do-it
    (let* ((frame (ad-get-arg 0))
	   (config (cdr (assoc (ad-get-arg 0) fa-win-configs)))
	   (height (cdr (assoc frame fa-window-height)))
	   (width (cdr (assoc frame fa-window-width))))
      (if (window-configuration-p config)
	  (set-window-configuration config))
      (if (not (eq fa-cur-height height))
	  (set-frame-height frame height))
      (if (not (eq fa-cur-width width))
	  (set-frame-height frame width)))))

(defadvice frame-selected-window (around frame-assoc first)
  "If we are not in a windows environment, we return the current window."
  (if window-system
      ad-do-it
    (setq ad-return-value (selected-window))))

(defadvice selected-frame (around frame-assoc first)
  "If in a non-windows environment return the window configuration."
  (if window-system
      ad-do-it
    (setq ad-return-value (car (car fa-frame-alist)))))

(defadvice frame-live-p (around frame-assoc first)
  "If we are not in a windows environment just reurn t."
  (if window-system
      ad-do-it
    (setq ad-return-value t)))

(defadvice modify-frame-parameters (around frame-assoc first)
  "If we are not in a windows environment do nothing."
  (if window-system
      ad-do-it
    nil))

(defadvice set-frame-width (around frame-assoc first)
  "If we are not in a windows environment use fa-first-frame."
  (if (not window-system)
      (ad-set-arg 0 fa-first-frame))
  ad-do-it)

(defadvice set-frame-height (around frame-assoc first)
  "If we are not in a windows environment use fa-first-frame."
  (if (not window-system)
      (ad-set-arg 0 fa-first-frame))
  ad-do-it)

(defadvice select-frame (around frame-assoc first)
  "If we are not in a windows environment restore the screen configuration as
supplied by the parameter."
  (if window-system
      ad-do-it
    (setq ad-return-value (ad-get-arg 0))
    (let ((config (cdr (assoc ad-return-value fa-win-configs)))
	  (height (cdr (assoc ad-return-value fa-window-height)))
	  (width (cdr (assoc ad-return-value fa-window-width))))
      (if (window-configuration-p config)
	  (set-window-configuration config))
      (if (not (eq fa-cur-height height))
	  (set-frame-height ad-return-value height))
      (if (not (eq fa-cur-width width))
	  (set-frame-height ad-return-value width)))))

(defadvice frame-parameters (around frame-assoc first)
  "If we are not in a windows environment return nil."
  (if (not window-system)
      (ad-set-arg 0 fa-first-frame))
  ad-do-it)

;;; ADVICE AND FUNCTIONS FOR DEALING WITH SERVER EDITS

(defun fa-server-visit-fn ()
  "Associate the selected-buffer with the selected-frame."
  (fa-associate (current-buffer) (selected-frame)))

(setq server-visit-hook (cons 'fa-server-visit-fn 
			      (and (boundp 'server-visit-hook)
				   server-visit-hook)))

(defun fa-server-done-fn ()
  "Dissociate the server-edit frame from the current buffer.  If the buffer
is then floating, offer to delete it.  If the frame has no buffers delete it
without asking."
  (let ((buf (current-buffer))
	(frame (or (fa-frame-exists "SERVER EDIT")
		   (selected-frame))))
    (switch-to-buffer (fa-best-buffer frame))
    (fa-dissociate buf frame)
    (if (cdr (assoc frame fa-frame-alist))
	nil
      (delete-frame frame))))

(setq server-done-hook (cons 'fa-server-done-fn 
			     (and (boundp 'server-done-hook)
				  server-done-hook)))

(defadvice server-visit-files (before frame-assoc first)
  "Make the server edit happen in a new frame."
  (fa-goto-named "SERVER EDIT" t))

(defconst fa-inhibit-bury nil
  "Boolean used to inhibit the action of bury-buffer from server-buffer-done")

(defadvice bury-buffer (around frame-assoc first)
  "Only bury the buffer if we are not inhibited by fa-inhibit-bury.  This
inhibition is necessary when frame-assoc is in use"
  (if fa-inhibit-bury
      nil
    ad-do-it))

(defadvice server-buffer-done (around frame-assoc first)
  "Prevent bury-buffer from happening since we deal with buffers in a 
different way as we are running frame-assoc."
  (let ((fa-inhibit-bury t))
    ad-do-it))

;; Now we must perform any standard frame initialisation.

(if window-system
    (let ((params (frame-parameters (selected-frame))))
      (if (or (null fa-use-default)
	      (and (string= "black"
			    (cdr (assoc 'foreground-color params)))
		   (string= "black"
			    (cdr (assoc 'cursor-color params)))
		   (string= "white"
			    (cdr (assoc 'background-color params)))))
	  (progn
	    (modify-frame-parameters (selected-frame)
				     fa-default-frame-parameters)
	    (setq default-frame-alist  fa-default-frame-parameters)))
      (setq fa-frame-alist (list (list (selected-frame)))))
  (setq frame-initial-frame-alist nil)
  (setq fa-frame-alist (list (list (fa-make-screen)))))


;;; Activation and deactivation code for frame-assoc

(defvar fa-frame-assoc-active nil
  "Boolean describing whether the frame-association code is active.")

(defvar fa-activate-hook nil
  "Hook run when frame-assoc is activated by fa-frame-assoc.")

(defvar fa-deactivate-hook nil
  "Hook run when frame-assoc is deactivated by fa-frame-assoc.")

(defun fa-remove-deaduns ()
  "Remove any dead buffers and frames from the fa-xxx-alist variables."
  (let ((framelist fa-frame-alist)
	(buflist fa-buffer-alist)
	cur-entry frame buf)
    (while framelist
      (progn
	(setq frame (car (car framelist)))
	(setq framelist (cdr framelist))
	(if (not (frame-live-p frame))
	    (setq fa-frame-alist (delq (assoc frame fa-frame-alist)
				       fa-frame-alist)))))
    (while buflist
      (progn
	(setq buf (car (car buflist)))
	(setq buflist (cdr buflist))
	(if (not (bufferp buf))
	    (setq fa-buffer-alist (delq (assoc buf fa-buffer-alist)
					fa-buffer-alist)))))))

(defun fa-identify-new-frames ()
  "Generate a list of frames not currently known to fa-frame-alist."
  (let ((all-frames (frame-list))
	(new-frames))
    (while all-frames
      (if (not (assoc (car all-frames) fa-frame-alist))
	  (setq new-frames (cons (car all-frames) new-frames)))
      (setq all-frames (cdr all-frames)))
    new-frames))

(defun fa-associate-displayed-buffers (framelist)
  "Associate buffers displayed in frames given by FRAMELIST parameter with
those frames."
  (let (buflist frame buf)
    (while framelist
      (setq frame (car framelist))
      (setq framelist (cdr framelist))
      (setq buflist (fa-new-displayed-bufs frame))
      (while buflist
	(setq buf (car buflist))
	(setq buflist (cdr buflist))
	(fa-associate buf frame)))))

(defun fa-associate-lists (buflist framelist)
  "Associate all buffers in BUFLIST with all frames in FRAMELIST."
  (let (buf frame flist)
    (while buflist
      (setq buf (car buflist))
      (setq buflist (cdr buflist))
      (setq flist framelist)
      (while flist
	(setq frame (car flist))
	(setq flist (cdr flist))
	(fa-associate buf frame)))))
   
(defun fa-frame-assoc (&optional arg)
  "Function to activate or disable frame association.
With argument ARG, activate frame-assoc if ARG is positive.  With no arg,
toggle activation and disabling.  Runs fa-activate-hook when activating
frame association and fa-deactivate-hook when disabling it."
  (interactive "P")               
  (let ((already-active fa-frame-assoc-active))
    (setq fa-frame-assoc-active
	  (if (null arg) (not fa-frame-assoc-active)
	    (> (prefix-numeric-value arg) 0)))
    (if fa-frame-assoc-active
	(progn
	  ;; We are activating frame-assoc.
	  (if (not already-active)
	    (let (framelist)
	      ;; Remove any dead buffers and frames from the fa-xxx-alist
	      ;; variables.
	      (fa-remove-deaduns)

	      ;; Identify any new frames.  If there are no new frames
	      ;; use the current frame for association with new buffers.

	      (setq framelist (or (fa-identify-new-frames)
				  (list (car (car fa-frame-alist)))))
	      
	      ;; Associate new buffers that are being displayed only 
	      ;; with the frame in which they appear.

	      (fa-associate-displayed-buffers (frame-list))

	      ;; Associate all other new buffers with all new frames.

	      (fa-associate-lists (fa-new-bufs) framelist)

	      ;; Activate all frame-assoc advice.
	      (ad-activate-regexp "frame-assoc")

	      ;; Run hooks
	      (run-hooks 'fa-activate-hook)))
	  (message "Frame association is now active."))
      ;; else we are deactivating frame association.
      (if already-active
	  (progn
	    ;; Deactivate all frame-assoc advice.
	    (ad-deactivate-regexp "frame-assoc")

	    ;; Run hooks
	    (run-hooks 'fa-deactivate-hook)))
      (message "Frame association is now inactive."))))


(provide 'frame-assoc)


;; DEBUG CODE FOLLOWS 

(defvar fa-ft-count 0)
(defvar fa-trace nil
  "Whether fa-frame-trace is enabled or not.")

(defun fa-frame-trace (string)
  "Function to record the current frame etc, in a buffer called frame-trace."
  (if fa-trace
      (save-excursion
	(set-buffer "frame-trace")
	(goto-char (point-max))
	(forward-line -1)
	(insert "\n" (int-to-string (setq fa-ft-count (1+ fa-ft-count)))
		"\n Called from: " string
		"\nFRAMES:  " (prin1-to-string fa-frame-alist)
		"\nBUFFERS: "  (prin1-to-string fa-buffer-alist)
		"\n"))))

(defun debug-stuff ()
  (progn
    (setq fa-frame-names nil)
    (setq fa-frame-alist nil)
    (setq fa-buffer-alist nil))

  (setq debug-on-error t)
  (setq fa-trace t)
  (describe-variable 'fa-frame-alist)
  (describe-variable 'fa-frame-names)
  (describe-variable 'fa-buffer-alist))



