;; Marc's font scaling stuff.

(defvar my-default-font
  "-misc-fixed-medium-r-normal--14-130-75-75-c-70-iso8859-1"
  "User-defined customisation variable for defining the default font.")

(defvar my-secondary-font
  "Liberation Mono-9"
  "A secondary font that scales better than my-default-font.")

(defvar fs::font-scaling-factor 1.2
  "The amount by which font scale will be incremented or decremented
by fs::font-scale-up or fs::font-scale-down.")

(defvar fs::scaled-font-preferences
  '((-1 . my-secondary-font)
    (0 . my-default-font)
    (1 . my-secondary-font))
  "Alist of font scaling indices to prefered fonts.  The nearest scaling
index to the car of an alist identfies the font to be used in
association with that scaling factor.  The alist is sorted in ascending
car order.  Use the function fs::font-preference-for-scale to get the
preferred font for a given scale-factor.")

(defvar fs::set-frame-defaults-when-scaling
  "If t, set default frame parameters when resizing a frame, so that new frames
have the same size and font"
t)

(defun fs::preferred-entry-for-index(index prev &optional next)
  "Identify whether INDEX better matches PREV or NEXT from an ordered
alist of font preferences.  This is a utility function for
fs::font-preference-for-index."
  (if next
      (if (<= index (car prev))
	  (cdr prev)
	(if (> index (car next))
	    nil
	  ;; index is between prev and next
	  (if (< (- index (car prev))
		 (- (car next) index))
	      (cdr prev)
	    (cdr next))))
    (cdr prev)))


(defun fs::font-preference-for-index(index)
  "Identify the font from fs::scaled-font-preferences that best matches
the supplied scale INDEX."
  (if fs::scaled-font-preferences
      (let* ((next fs::scaled-font-preferences)
	     (prev)
	     (result))
	(while (and (not result)
		    (setq prev (car next)))
	  (setq next (cdr next))
	  (setq result (fs::preferred-entry-for-index index
						      prev (car next))))
	result)
    ;; No fs::scaled-font-preferences set
    (frame-parameter nil 'font)))


(defun fs::cur-font-height()
 (face-attribute 'default :height (selected-frame)))


(defun fs::check-frame-dimensions()
  "Check the current frame dimensions, recording them as variables in the
  curent frame's frame-alist.  This is used to identify whether a frame's
  dimensions have been manually modified and if so to make that modified
  dimension the new target size.  Changes made to the frame dimensions
  resulting from resizing the font do not change the target size."
  (let* ((auto-height (frame-parameter nil 'fs::auto-height))
	 (frame-height (frame-height))
	 (auto-width (frame-parameter nil 'fs::auto-width))
	 (frame-width (frame-width)))

    (if (not (frame-parameter nil 'fs::target-height))
	;; No target height recorded.  Record current height as target
	(set-frame-parameter nil 'fs::target-height frame-height)
      (if auto-height
	  (if (not (= auto-height frame-height))
	      ;; height has changed since being recorded.  This must
	      ;; have been a manual change, so make this changed
	      ;; height the new target
	      (set-frame-parameter nil 'fs::target-height frame-height))))

    (if (not (frame-parameter nil 'fs::target-width))
	;; No target width recorded.  Record current width as target
	(set-frame-parameter nil 'fs::target-width frame-width)
      (if auto-width
	  (if (not (= auto-width frame-width))
	      ;; width has changed since being recorded.  This must
	      ;; have been a manual change, so make this changed
	      ;; width the new target
	      (set-frame-parameter nil 'fs::target-width frame-width))))))


(defun fs::update-frame-dimensions()
  "Attempt to set frame height and width to their target settings and
record what they actually get set to."
  (let* ((target-height (frame-parameter nil 'fs::target-height))
	 (target-width (frame-parameter nil 'fs::target-width)))
    
    (set-frame-height nil target-height)
    (set-frame-width nil target-width)
    (set-frame-parameter nil 'fs::auto-height (frame-height))
    (set-frame-parameter nil 'fs::auto-width (frame-width))))


(defun fs::frame-scale-index()
  "Identify the current scale index for the current frame."
  (or (frame-parameter nil 'fs::font-scale-index) 0))


(defun fs::frame-scaling-factor(index)
  "Identify the scaling factor for a given scale INDEX."
  (let ((result 1))
    (while (> index 0)
      (setq result (* result fs::font-scaling-factor))
      (setq index (1- index)))
    (while (< index 0)
      (setq result (/ result fs::font-scaling-factor))
      (setq index (1+ index)))
    result))


(defun fs::cur-scaling-factor()
  "Return the scaling factor for the current frame."
  (fs::frame-scaling-factor
   (or (frame-parameter nil 'fs::font-scale-index) 0)))


(defun fs::record-font-base-height(font height)
  "Record the base height of the given font in an alist in frame-parameters."
  (let ((font-heights (frame-parameter nil 'fs::base-height))
	(this))
    (if font-heights
	(setq font-heights (assq-delete-all font font-heights)))
    (set-frame-parameter nil 'fs::base-height 
			 (cons (cons font height) font-heights))))


(defun fs::font-base-height(font)
  "Get the base hight for the given FONT (which must be the name of the 
  current frame font), recording it if it is not known."
  (let* ((font-heights (frame-parameter nil 'fs::base-height))
	 (height (cdr (assoc font font-heights))))
    (if (not height)
	(progn
	  (setq height
		(/ (face-attribute 'default :height (selected-frame))
		   (fs::cur-scaling-factor)))
	  (fs::record-font-base-height font height)))
	  
    height))


(defun fs::set-default-frame-scale(index)
  "Set default frame parameters so that new frames will have the same size
as the most recently resized one."
  (let ((entry (assoc 'fs::font-scale-index default-frame-alist)))
    (if entry
	(setcdr entry index)
      (add-to-list 'default-frame-alist 
		   (cons 'fs::font-scale-index index)))))


(defun fs::after-make-frame-hook-fn(frame)
  "Hook function to be run after creating a new frame, FRAME."
  (let ((scale (cdr (assoc 'fs::font-scale-index default-frame-alist))))
    (fs::set-scaled-font (or scale 0) frame)))

(setq after-make-frame-functions
      (append after-make-frame-functions 
	      (cons 'fs::after-make-frame-hook-fn nil)))


(defun fs::set-scaled-font(index &optional frame)
  "Rescale the font scale index to INDEX for FRAME, (default is
current frame.  This may cause the font to change, and may change
the frame's height and width, though reverting the index will
also revert the height and width unless you manually change it in
the meantime."
  (let ((target-font (fs::font-preference-for-index index))
	(base-height)
	(cur-frame (selected-frame)))

    (and frame (select-frame frame))

    (fs::check-frame-dimensions)
    (set-frame-font (eval target-font))
    (setq base-height (fs::font-base-height target-font))
    (set-face-attribute 'default (selected-frame) :height 
			(round (* base-height 
				  (fs::frame-scaling-factor index))))
    (set-frame-parameter nil 'fs::font-scale-index index)
    (fs::update-frame-dimensions)
    (and frame (select-frame cur-frame))
    (if fs::set-frame-defaults-when-scaling
	(fs::set-default-frame-scale index)))
  (redisplay))


(defun fs::font-scale-up()
  "Scale up the font size."
  (interactive)
  (fs::set-scaled-font 
   (+ (frame-parameter nil 'fs::font-scale-index) 1)))


(defun fs::font-scale-down()
  "Scale down the font size."
  (interactive)
  (fs::set-scaled-font 
   (- (frame-parameter nil 'fs::font-scale-index) 1)))

(fs::set-scaled-font 0)

(provide 'font-scale)