;; mykeys.el
;; Provide mykeys minor mode.  In this mode, all keys can have
;; non-default keybindings, while allowing the standard emacs
;; keybindings to be made available by simply exitting the mode.
;;
;; Also ESC enters command mode in which a number of keys have special
;; meanings:
;; n - repeat last search forwards
;; p - repeat last search backwards
;; / - enter a new search term
;; e - execute the last defined keyboard macro
;; . - repeat last edit sequence
;; : - enter a vi-like command
;; ESC - exit command mode
;;
;; As a further complication/bonus, any key bound to a function from the
;; ESC-prefix keymap will still work as normal, regardless of whether we
;; have just entered (or exitted) command mode.
;;
;; Any other key sequence will do what it does normally and cause command
;; mode to be exited.


(defconst mykeys-esc-key (kbd "<f99>")
  "The key to which ESC will be translated in mykeys mode.")

(defvar mykeys-in-command-mode nil
  "Whether we are in command mode.")

(defvar mykeys-command-mode-map (make-sparse-keymap)
  "The keymap for command mode in mykeys-mode.")

(defconst mykeys-base-map (make-sparse-keymap)
  "An empty keymap for non-ESC mode in mykeys-mode.")

(defconst mykeys-base-lighter " MK"
  "The lighter string for the mykeys minor mode in non-command mode.")

(defconst mykeys-command-mode-suffix "+"
  "The lighter string for the mykeys minor mode in non-command mode.")

(defun mykeys-enter-command-mode()
  "Activate command-mode"
  (interactive)
  (setcar (cdr (assq 'mykeys-mode minor-mode-alist)) 
	  (concat mykeys-base-lighter mykeys-command-mode-suffix))
  (setcdr (assq 'mykeys-mode minor-mode-map-alist) mykeys-command-mode-map)
  (setq mykeys-in-command-mode t)
  (force-mode-line-update))

(defun mykeys-exit-command-mode ()
  "Deactivate command mode"
  (interactive)
  (setcar (cdr (assq 'mykeys-mode minor-mode-alist)) mykeys-base-lighter)
  (setcdr (assq 'mykeys-mode minor-mode-map-alist) mykeys-base-map)
  (setq mykeys-in-command-mode nil)
  (force-mode-line-update))

(define-key mykeys-base-map mykeys-esc-key 'mykeys-enter-command-mode)
(define-key mykeys-command-mode-map mykeys-esc-key 'mykeys-exit-command-mode)

(define-minor-mode mykeys-mode
  "Modified keymap mode."
  :lighter " mykeys"
  :global t
  :keymap mykeys-base-map
  (if mykeys-mode
      (progn
	(define-key key-translation-map (kbd "<ESC>") mykeys-esc-key)
	(add-hook 'pre-command-hook 'mykeys-pre-command)
	(add-hook 'post-command-hook 'mykeys-post-command)
	)	
    (define-key key-translation-map (kbd "<ESC>") nil)
    (remove-hook 'pre-command-hook 'mykeys-pre-command)
    (remove-hook 'post-command-hook 'mykeys-post-command)
    )
  (mykeys-exit-command-mode))


(defun mykeys-last-key-was-escape-p ()
  (or (eq last-command 'mykeys-enter-command-mode)
      (eq last-command 'mykeys-exit-command-mode)))

(defun mykeys-is-command-mode-command-p ()
  "Whether COMMAND is being invoked from mykeys command-mode."
  (and mykeys-in-command-mode
       (lookup-key mykeys-command-mode-map (this-single-command-raw-keys))))

(defun mykeys-keys-yield-esc-command ()
  (key-binding 
   (vconcat [27] (this-single-command-raw-keys))))


(defun mykeys-pre-command ()
  "If the last key typed was ESC and we are not about to execute one of
  the mykeys command-mode functions, we may have meant to run a standard
  emacs command.
  Also, if we are in command-mode and no command is bound to the current
  key, then we exit command-mode."
  (if (and (mykeys-last-key-was-escape-p)
	   (not (mykeys-is-command-mode-command-p)))
      ;; Maybe the user wants to type ESC-x or something.
      (let ((fn (mykeys-keys-yield-esc-command)))
	(and fn	(setq this-command fn))))
  (if (and mykeys-in-command-mode
	   (not (mykeys-is-command-mode-command-p)))
      (mykeys-exit-command-mode)))


(defun mykeys-post-command ()
  "If we just ran the quit command, exit command mode."
  (condition-case nil
      (progn
	(if (eq this-command 'keyboard-quit)
	    (mykeys-exit-command-mode))
	)
      (error nil)))
  
(mykeys-mode -1)
(provide 'mykeys-mode)
