x;; redo.el
;; Provide an automated redo capture system.  This is a bit like having 
;; keyboard macros being recorded at all times and having immediate access
;; to them.  

;; Redo is recorded in redo groups.  A redo group begins with an
;; interactive edit (changes to the buffer by minibuffer-invoked
;; commands do not begin a redo group).
;; The current redo group is buffer-local (like undo)
;; Redo groups are completed when a redo stop action occurs.  When this
;; happens, the redo-group is added to the global redo-ring.
;; Redo stop actions include:
;; - switching buffers;
;; - moving lines by a non-edit, non-search operation (unless the mark
;;   is active);
;; - writing the buffer
;; - keyboard-quit
;; - An edit of the buffer from another frame
;; - undo past the buffer state at the start of the redo group
;; - A change of the window configuration
;; - an interactive command (M-x)

(require 'ring)

;; Each redo-group is stored as a cons cell containing a vector of key
;; sequences and optionally a copy of the redo-ring.  The redo-ring
;; copy is only provided when the key sequence invokes execute-redo.
;; This allows the executed redo to have access to the redo-ring as it
;; was at the time the redo was recorded.

(defvar redo-current-group (cons [] nil)
  "A record of the so far saved keystrokes for the current redo-group.")

(make-variable-buffer-local 'redo-current-group)

(defvar redo-ring (make-ring 10)
  "Ring of the last N distinct redo entries")

(defvar redo-ring-idx 0
  "Index into redo-ring, used for redo-pop, etc.")

(defvar redo-state
  '()
  "Alist of redo-state.")

(make-variable-buffer-local 'redo-state)


;; redo state consists of a number of pieces of data.  These are stored
;; in an alist.  The following functions provide primitives for managing
;; that state data, 
;;
(defun redo-set-state (key value)
  "In redo-state set entry given by KEY to VALUE."
  (let ((entry (assq key redo-state)))
    (if entry
	(setcdr entry value)
      (setq redo-state (append redo-state (cons (cons key value) nil)))))
  value)


(defun redo-get-state (key)
  "Get the value associated with redo-state entry KEY."
  (cdr (assq key redo-state)))

(defvar redo-pre-cmd-buffer nil
  "The buffer that was current when redo-pre-command was last called.")

(defun redo-record-state ()
  "Record the current undo state in redo-state."
  (redo-set-state 'buffer (setq redo-pre-cmd-buffer (current-buffer)))
  (redo-set-state 'tick (buffer-modified-tick))
  (redo-set-state 'point (point))
  (redo-set-state 'frame (selected-frame))
  (redo-set-state 'winconf (current-window-configuration)))


(defun redo-record-keys (keys)
  "Record the current set of keys into the redo-group vector."
  (setcar redo-current-group
	  (vconcat (car redo-current-group) keys)))


(defun redo-commit-redo ()
  "Record the latest captured redo-group into the redo-ring."
  (ring-insert redo-ring redo-current-group)
  (setq redo-ring-idx 0))


(defun redo-reset-redo ()
  "Reset the current redo-group."
  (setq redo-current-group (cons [] nil))
  (when (redo-recording-p)
    (redo-set-state 'recording nil)
    (message (concat "FINISHED RECORDING in " 
		     (prin1-to-string (current-buffer))))))


(defun redo-finish-redo ()
  "Save the current redo-group and reset."
  (save-excursion
    (if (redo-in-minibuffer-p)
	(set-buffer redo-pre-cmd-buffer))
    (redo-commit-redo)
    (redo-reset-redo)))


(defun redo-start-recording (keys)
  "Start recording a new redo-group."
  (redo-reset-redo)
  (redo-set-state 'recording t)
  (redo-record-keys keys)
  (message (concat "STARTING TO RECORD in " 
		   (prin1-to-string (current-buffer)))))


(defun redo-recording-p ()
  "Are we currently recording redo?"
  (redo-get-state 'recording))


(defun redo-quit-p ()
  "Did we just hit a quit operation?."
  (eq last-command 'keyboard-quit))


(defun redo-buffer-exempt-p (buffer)
  "Is BUFFER exempt from redo-collection."
  (or (string-match-p "^ \\*.*\\*$" (buffer-name buffer))
      (string-equal "*Messages*" (buffer-name buffer))
      (save-excursion
	(set-buffer buffer)
	buffer-read-only)))


(defun redo-terminate-recording-p ()
  "Should we finish recording the current redo-group?"
  nil)


(defun redo-start-redo-event-p ()
  "Was the last thing we did something that could start a redo record?."
  (message (concat "XX: " (prin1-to-string (redo-get-state 'tick))
		   ", " (prin1-to-string (buffer-modified-tick))))
  (if (redo-buffer-exempt-p (current-buffer))
      nil
    (or 
     (not (redo-get-state 'tick))
     (not (= (redo-get-state 'tick) (buffer-modified-tick))))))


(defun redo-pre-command ()
  "Record enough state information that we can figure out whether to 
start, stop or continue recording redo-state in the post-command hook."
  (redo-record-state))


(defun redo-post-command ()
  "Start, stop or continue recording redo depending on state and what is
going on."
  (if (redo-recording-p)
      (if (redo-quit-p)
	  (redo-reset-redo)
	(if (redo-terminate-recording-p)
	    (redo-finish-redo)
	  (redo-record-keys (this-command-keys))))
    ;; Not recording.  Work out if we should start.
    (if (redo-start-redo-event-p)
	(redo-start-recording (this-command-keys)))))


(defun redo-execute-entry (entry)
  "Execute ????"
  (let ((keys (car entry))
	(redo-current-group (or (cdr entry) redo-current-group)))
    (execute-kbd-macro keys)))


(defun redo-rm-undo-boundary (undo)
  "Remove the undo boundary in cdr of undo."
  (or (cadr undo)
      (setcdr undo (cddr undo)))
  undo)





(defun redo-close-up-undo (undo-prev undo)
  "Remove undo-boundaries from the section of undo preceding undo-prev."
  (let ((this undo))
    (while this
      (if (eq (cdr this) undo-prev)
	  (setq this nil)
	(message (concat "this: " (prin1-to-string this)))
	(redo-rm-undo-boundary this)
	(message (concat "this2: " (prin1-to-string this)))
	(setq this (cdr this)))))
  undo)


(add-hook 'pre-command-hook 'redo-pre-command)
(add-hook 'post-command-hook 'redo-post-command)


(remove-hook 'pre-command-hook 'redo-pre-command)
(remove-hook 'post-command-hook 'redo-post-command)



redo-state

(redo-record-state)

redo-current-group


(defun redo-fetch-from-ring ()
  "Return the currently selected redo"
  (ring-ref redo-ring redo-ring-idx))



this-command-keys

(setq undo-before buffer-undo-list)
(setq undo-after buffer-undo-list)

(redo-execute-entry '([97 98 99 up 100 101 102 down] . nil))



;; hook fns:
Before command
record current state


after command
if we are recording
  if  we should be no longer recording
    commit this record
  else
    append the last keys to redo-current-group

else
  

(insert-string (prin1-to-string undo-before))
(nil (#("a" 0 1 (fontified t)) . -4810) (#<marker at 4881 in redo.el> . -1) nil (#("a" 0 1 (fontified t)) . -4811) nil (4810 . 4812) (t 21474 . 35328))

(insert-string (prin1-to-string undo-after))
(nil (4882 . 4885) 4946 nil (4943 . 4946) nil (#("a" 0 1 (fontified t)) . -4810) (#<marker at 5361 in redo.el> . -1) nil (#("a" 0 1 (fontified t)) . -4811) nil (4810 . 4812) (t 21474 . 35328))

(defun redo-remove-undo-boundaries (start-undo end-undo)
  ""
  (while (and end-undo
	      (not (eq start-undo end-undo)))
    (setq end-undo (cdr end-undo))
    (if (not (car end-undo))
)
  end-undo)

(redo-remove-undo-boundaries undo-before undo-after)


(cdr undo-before)

(setq x '(nil (a b c) 2 nil 3 nil 4))




### WORKING ON THE FUNCTION BELOW.  THE STUFF ABOVE LOOKS LIKE IT WORKS
	      






(close-one '(a nil nil (2 3 )))

(defun closeup (list)
  (let ((this (cdr list)))

(closeup x)


(cdr (cdr (cdr x)))

    
(defun redo-find-in-undo (undo-start undo)
  "Check whether we can find undo-start in undo."
  (while (and undo
	      (not (eq undo-start undo)))
    (setq undo (cdr undo)))
  undo)
